from __future__ import absolute_import, division, print_function
import numpy as np
from subprocess import call
import argparse
import pickle, os, timeit, time
import matplotlib.pyplot as plt
from datetime import datetime
import sys
parent_path = os.path.dirname(os.path.realpath(__file__))
grandparent_path = os.path.dirname(parent_path)
sys.path.append(grandparent_path)
sys.path.append(parent_path)
from Utils.common import load_run_data

os.environ['CUDA_VISIBLE_DEVICES'] = '0'

parser = argparse.ArgumentParser()

parser.add_argument('--n_pixels_shuffles', type=int, help='In case of "Shuffled_Pixels": how many pixels swaps',
                    default=200)

args = parser.parse_args()
n_pixels_shuffles = args.n_pixels_shuffles

base_run_name = 'ShuffledPixels' + str(n_pixels_shuffles) + '_SamplesM_kl'

# -------------------------------------------------------------------------------------------
# Run experiments for a grid of 'number of training samples in each training task'
# -------------------------------------------------------------------------------------------


n_train_tasks = 10
run_experiments = True # If false, just analyze the previously saved experiments

root_saved_dir = 'saved/'
basic_list = range(5, 61, 5)
# basic_list = [1,2]
m_train_samples_vec = [ele * 1000 for ele in basic_list] # max value 60000
sub_runs_names = [base_run_name + '/' + str(m_train_samples) for m_train_samples in m_train_samples_vec]

if run_experiments:
    start_time = timeit.default_timer()
    for i_run, m_train_samples in enumerate(m_train_samples_vec):
        call(['python', 'PriorMetaLearning/main_Meta_Bayes.py',
              '--run-name', sub_runs_names[i_run],
              '--data-source', 'MNIST',
              '--data-transform', 'Shuffled_Pixels',
              '--n_pixels_shuffles', str(n_pixels_shuffles),
              '--n_train_tasks', str(n_train_tasks),
              '--limit_train_samples_in_train_tasks', str(m_train_samples),
              '--limit_train_samples_in_test_tasks', '2000',
              '--model-name',   'FcNet3',
              '--complexity_type',  'PAC_Bayes_Abel_kl', # 'NewBoundMcAllaster' 'NewBoundSeeger'
              # 'PAC_Bayes_lambda' 'PAC_Bayes_quad' 'PAC_Bayes_variational_role'
              '--n_test_tasks', '20',  # 100 or 20
              '--n_meta_train_epochs', '50',  # catoni_bound: 150 for kl_bound:40
              '--n_meta_test_epochs', '200',  # 200
              '--meta_batch_size', '16',
              '--mode', 'MetaTrain',  # 'MetaTrain'  \ 'LoadMetaModel'
              ])
    stop_time = timeit.default_timer()
    # Save log text
    message = ['Run finished at ' +  datetime.now().strftime(' %Y-%m-%d %H:%M:%S'),
               'Tasks number grid: ' + str(m_train_samples_vec),
                'Total runtime: ' + time.strftime("%H hours, %M minutes and %S seconds", time.gmtime(stop_time - start_time)),
               '-'*30]
    log_file_path = os.path.join(root_saved_dir, base_run_name, 'log') + '.out'
    with open(log_file_path, 'a') as f:
        for string in message:
            print(string, file=f)
            print(string)

# -------------------------------------------------------------------------------------------
# Analyze the experiments
# -------------------------------------------------------------------------------------------

# test error analysis
mean_error_per_tasks_n = np.zeros(len(m_train_samples_vec))
std_error_per_tasks_n = np.zeros(len(m_train_samples_vec))
# test loss analysis
mean_loss_per_tasks_n = np.zeros(len(m_train_samples_vec))
std_loss_per_tasks_n = np.zeros(len(m_train_samples_vec))
# test bound analysis
mean_bound_per_tasks_n = np.zeros(len(m_train_samples_vec))
std_bound_per_tasks_n = np.zeros(len(m_train_samples_vec))

# test_err_vec
# test_err_loss
# test_err_bound
for i_run, n_train_tasks in enumerate(m_train_samples_vec):
    run_result_path = os.path.join(root_saved_dir, sub_runs_names[i_run])
    prm, info_dict = load_run_data(run_result_path)
    test_err_vec = info_dict['test_err_vec']
    mean_error_per_tasks_n[i_run] = test_err_vec.mean()
    std_error_per_tasks_n[i_run] = test_err_vec.std()
    # test loss
    test_loss_vec = info_dict['test_loss_vec']
    mean_loss_per_tasks_n[i_run] = test_loss_vec.mean()
    std_loss_per_tasks_n[i_run] = test_loss_vec.std()
    # test bound
    test_bound_vec = info_dict['test_bound_vec']
    mean_bound_per_tasks_n[i_run] = test_bound_vec.mean()
    std_bound_per_tasks_n[i_run] = test_bound_vec.std()

# Saving the analysis:
with open(os.path.join(root_saved_dir, base_run_name, 'runs_analysis.pkl'), 'wb') as f:
    pickle.dump([mean_error_per_tasks_n, std_error_per_tasks_n, m_train_samples_vec], f)

with open(os.path.join(root_saved_dir, base_run_name, 'runs_analysis.pkl'), 'wb') as f:
        pickle.dump([mean_loss_per_tasks_n, std_loss_per_tasks_n, m_train_samples_vec], f)

with open(os.path.join(root_saved_dir, base_run_name, 'runs_analysis.pkl'), 'wb') as f:
    pickle.dump([mean_bound_per_tasks_n, std_bound_per_tasks_n, m_train_samples_vec], f)

# Plot the analysis:
plt.figure()
plt.errorbar(m_train_samples_vec, 100*mean_error_per_tasks_n, yerr=100*std_error_per_tasks_n)
plt.xticks(m_train_samples_vec)
plt.xlabel('Number of training-tasks')
plt.ylabel('Error on new task [%]')


plt.figure()
plt.errorbar(m_train_samples_vec, mean_loss_per_tasks_n, yerr=std_loss_per_tasks_n)
plt.xticks(m_train_samples_vec)
plt.xlabel('Number of training-tasks')
plt.ylabel('Loss on new task')


plt.figure()
plt.errorbar(m_train_samples_vec, mean_bound_per_tasks_n, yerr=std_bound_per_tasks_n)
plt.xticks(m_train_samples_vec)
plt.xlabel('Number of training-tasks')
plt.ylabel('Bound on new task')
plt.show()

