import ray
import copy
import numpy as np
import pandas as pd
import os
import sys
import math
import os
import pacoh.experiments.hyperparam_search.util as util
import pacoh.experiments.hyperparam_search.tune_setup as ts

from ray import tune

from ray.tune.suggest.hyperopt import HyperOptSearch
from ray.tune import Analysis
from datetime import datetime

import argparse

BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
DATA_DIR = os.path.join(BASE_DIR, 'data')
HPARAM_EXP_DIR = os.path.join(DATA_DIR, 'tune-hparam-pacoh-final')

SEED = 28
# N_THREADS_PER_RUN = 1
TEST_SEEDS = [28, 29, 30, 31, 32]


def main(args):
    # if args['no_cuda']:
    #     os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
    # else:
    #     os.environ["CUDA_VISIBLE_DEVICES"] = "0"
    import config
    # import tensorflow as tf

    # ray.init(num_cpus=args['num_cpus'], memory=500 * 1024 ** 3, object_store_memory=200 * 1024 ** 3)
    ray.init(num_cpus=args['num_cpus'], memory=(args['mem_gb']-30) * 1024 ** 3, object_store_memory=20 * 1024 ** 3)

    def train_reg(config, reporter):
        args = config['args']
        print(args)
        del config['args']
        # tf.config.threading.set_inter_op_parallelism_threads(N_THREADS_PER_RUN)
        sys.path.append(BASE_DIR)

        print('Init config:', config)
        # 1) Update config for current run
        util.process_config(config)
        print('Processed config: ', config)

        # 2) setup model
        import pacoh.experiments.hyperparam_search.tune_setup as ts
        modelTuner = ts.get_tuner_class(args)
        modelTuner.init_run(config, random_seed=SEED)

        # 3) train and evaluate model
        for i in range(modelTuner.n_optimization_steps):
            results = modelTuner.run_iters(modelTuner.n_iterations_per_step)
            reporter(**results)

    @ray.remote
    def train_test(config):
        args = config['args']
        del config['args']

        # tf.config.threading.set_inter_op_parallelism_threads(N_THREADS_PER_RUN)

        print('Init config:', config)
        # 1) Update config for current run
        util.process_config(config)
        print('Processed config: ', config)

        # 2) setup model
        import pacoh.experiments.hyperparam_search.tune_setup as ts
        modelTuner = ts.get_tuner_class(args)
        modelTuner.init_run(config, random_seed=config['randomness_config']['random_seed'])

        # 3) train and evaluate model
        results = modelTuner.run_iters(modelTuner.n_total_iterations,
                                       config['randomness_config']['random_seed'])
        config.update(results)

        return config

    modelTuner = ts.get_tuner_class(args)
    if args['load_analysis_from'] != '':
        assert os.path.isdir(args['load_analysis_from']), 'load_analysis_from must be a valid directory'
        print('Loading existing tune analysis results from %s' % args['load_analysis_from'])
        analysis = Analysis(args['load_analysis_from'])
        exp_name = os.path.basename(args['load_analysis_from'])
    else:
        space = modelTuner.get_search_space()
        base_config = modelTuner.get_base_config()
        base_config['args'] = args


        config = {
            "num_samples": modelTuner.num_samples,
            "config": base_config,
            "stop": {
                "timesteps_total": 300000
            },
        }
        print(config)

        # Run hyper-parameter search

        algo = HyperOptSearch(
            space,
            max_concurrent=args['num_cpus'],
            metric=modelTuner.metric,
            mode=modelTuner.mode)

        exp_name = '%s_%s_%s_%s' % (EXP_PREFIX, args['dataset'], args['meta_algo_type'], args['inference_algo_type'])

        analysis = tune.run(train_reg, name=exp_name, search_alg=algo, verbose=1,
                            local_dir=HPARAM_EXP_DIR, resume=args['resume'], **config)#checkpoint_freq=1,

    from pacoh.experiments.hyperparam_search.util import select_best_configs
    best_configs = select_best_configs(analysis, metric=modelTuner.metric, mode=modelTuner.mode, N=args['n_test_runs'])

    test_configs = []
    for config in best_configs:
        for seed in TEST_SEEDS:
            test_config = copy.deepcopy(config)
            if 'randomness_config' in test_config:
                test_config['randomness_config'].update({'random_seed': seed})
            else:
                test_config['randomness_config'] = {'random_seed': seed}
            test_configs.append(test_config)

    result_dicts = ray.get([train_test.remote(config) for config in test_configs])

    result_df = pd.DataFrame(result_dicts)
    print(result_df.to_string())

    csv_file_name = os.path.join(HPARAM_EXP_DIR, '%s_%s.csv' % (exp_name, datetime.now().strftime("%b_%d_%Y_%H:%M:%S")))
    result_df.to_csv(csv_file_name)
    print("\nSaved result csv to %s" % csv_file_name)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run meta mll hyper-parameter search.')
    parser.add_argument('--dataset', type=str, default='sin_20', help='dataset')
    parser.add_argument("--no_cuda", action="store_true", help="Whether to use CUDA when available")
    parser.add_argument('--num_cpus', type=int, default=1, help='Number of workers')
    parser.add_argument('--mem_gb', type=int, default=100, help='Number of workers')
    parser.add_argument('--meta_algo_type', type=str, default='pacoh_svgd')
    parser.add_argument('--inference_algo_type', type=str, default='svgd')
    parser.add_argument('--n_test_runs', type=int, default=5, help='number of test runs')
    parser.add_argument("--resume", action="store_true", help="Whether to resume the experiments from a checkpoint")
    parser.add_argument("--load_analysis_from", type=str, default='')

    EXP_PREFIX = 'b_cal_err'

    args = parser.parse_args()
    args.no_cuda = True

    print('Running', os.path.abspath(__file__), '\n')
    print("--- Experiment Settings ---")
    print("Dataset:", args.dataset)
    print("No CUDA:", args.no_cuda)
    print("Num CPUS:", args.num_cpus)
    print("Available memory (in gigabytes):", args.mem_gb)
    print("Meta algorithm type:", args.meta_algo_type)
    print("Meta-test algorithm type:", args.inference_algo_type)
    print("Num test runs:", args.n_test_runs)
    print("Resume:", args.resume)
    print('\n')

    args_dict = vars(args)
    print(args_dict)
    main(args_dict)
