import math

import numpy as np
import tensorflow as tf

from pacoh import util as util

from .meta_algo import MetaLearner
import pacoh.modules as modules

class BMAML(MetaLearner):
    def __init__(self, meta_train_sampler, meta_test_sampler, model_config, likelihood_config, prior_config, meta_train_config,
                 meta_test_config, random_seed=None):

        super().__init__(meta_train_sampler, meta_test_sampler, model_config, random_seed=random_seed)

        # set variables
        self.meta_train_sampler = meta_train_sampler
        self.meta_test_sampler = meta_test_sampler

        self.likelihood_config = likelihood_config
        self.meta_train_config = meta_train_config
        self.meta_test_config = meta_test_config

        self.n_particles = self.meta_train_config['n_particles']
        self.split_batch = self.meta_train_config.get('split_batch', True)
        self.n_train_samples = tf.convert_to_tensor(self.meta_train_sampler.n_train_samples, dtype=tf.float32)
        self.prior_factor = self.meta_train_config['prior_factor']
        self.chaser_loss = self.meta_train_config['chaser_loss']

        # initialize relevant components / modules of the algorithm
        self.model = self._initialize_model(self.model_config, self.meta_train_sampler)
        self.likelihood = util.get_likelihood(config=likelihood_config, output_dim=self.model_config['output_dim'],
                                              n_batched_models=self.n_particles)
        self.prior = modules.BatchedGaussianPrior(batched_model=self.model, likelihood=self.likelihood,
                                                 config=prior_config, name='gaussian_prior')
        self.kernel = modules.RBF_Kernel_TF(self.meta_train_config['bandwidth'])
        self.affine_transform = modules.AffineTransform(normalization_mean=self.meta_train_sampler.y_mean,
                                                        normalization_std=self.meta_train_sampler.y_std)

        # check whether likelihood has trainable variables (e.g. the noise variance)
        if len(self.likelihood.trainable_variables) > 0:
            self.model_variables_only = False
        else:
            self.model_variables_only = True

        # initialize particles
        if self.model_variables_only:
            self.particles_init = tf.Variable(self.model.get_variables_stacked_per_model())
        else:
            model_params = self.model.get_variables_stacked_per_model()
            likelihood_params = tf.split(self.prior.sample(self.n_particles)[0], self.prior.split_sizes, axis=-1)[-1]
            self.particles_init = tf.Variable(tf.concat([model_params, likelihood_params], axis=-1))

        self._setup_optimizer(meta_train_config)

        print("\nTraining optimization parameters:")
        print(f"Meta batch size of: {self.meta_train_sampler.meta_batch_size}")
        self.meta_train_config = util.process_training_config(self.meta_train_config, meta_train_sampler)

    def predict(self, x, sample_functions=False, sample_from_prior=False):
        """
        TODO
        """
        x = self._process_eval_batch(x)
        y_pred = self.model(x)
        y_pred = self._unstandardize(y_pred)

        y_pred = tf.expand_dims(y_pred, axis=0)
        return y_pred, None

    def plot_predict(self, n_tasks=2):
        from matplotlib import pyplot as plt

        tasks = self.meta_test_sampler.meta_train_tasks[:n_tasks]
        fig, axes = plt.subplots(nrows=1, ncols=n_tasks)
        x_plot = tf.expand_dims(tf.range(tasks[0].x_min - 0.5, tasks[0].x_max + 0.5, delta=0.1), axis=-1)
        x_plot_tiled, _ = self.meta_test_sampler.meta_train_tasks[0]._tile_batch(self._standardize(x_plot))

        for i, task in enumerate(tasks):
            axes[i].scatter(*map(lambda x: x.numpy(), task.val_data), color='blue', alpha=0.1)
            axes[i].scatter(*map(lambda x: x.numpy(), task.train_data), color='orange')

            # pre-adaptation
            y_hat_pre = self._unstandardize(self.model(x_plot_tiled))
            for j in range(y_hat_pre.shape[0]):
                axes[i].plot(x_plot.numpy(), y_hat_pre[j].numpy(), label='pre', c='green')

            # post-adaptation
            x_context, y_context = task.get_batch()
            y_hat_post, _ = self.meta_predict(x_context, y_context, x_plot, standardize_context=False)
            for j in range(y_hat_post.shape[0]):
                axes[i].plot(x_plot.numpy(), y_hat_post[j].numpy(), label='post', c='red')
        fig.legend()
        fig.show()

    def meta_predict(self, x_context, y_context, x_test, standardize_context=True):
        # data handling
        if standardize_context:
            x_context, y_context = self._standardize(x_context, y_context)
        x_test = self._process_eval_batch(x_test)

        lr = self.meta_train_config['inner_learning_rate']

        # perform adaptation step with context data
        particles = self.particles_init
        for _ in range(self.meta_train_config['n_inner_steps'] + self.meta_test_config['n_steps_extra']):
            particles = self._svgd_step(particles, x_context, y_context, lr)

        # predict with adapted model
        model_params, likelihood_params = self._split_particles(particles)
        y_hat = self.model.forward_parametrized(x_test, tf.reshape(model_params, (1, -1)))

        if self.model_variables_only:
            pred_dist = self.likelihood.get_pred_dist(y_hat)
        else:
            pred_dist = self.likelihood.get_pred_dist(y_hat, param=likelihood_params)

        # unstandardize predictions
        y_pred = self._unstandardize(y_hat)
        pred_dist = self.affine_transform.apply(pred_dist)
        return y_pred, pred_dist

    def meta_eval(self, print_results=False, max_tasks_parallel=5):
        import time
        eval_metrics_dict_list = []

        t = time.time()
        for eval_task in self.meta_test_sampler.meta_train_tasks:
            x_train_tiled, y_train_tiled = eval_task.process_batch(*eval_task.train_data)
            x_val, y_val = eval_task.val_data

            y_pred, pred_dist = self.meta_predict(x_train_tiled, y_train_tiled, x_val, standardize_context=False)
            eval_metrics_dict = self.likelihood.calculate_eval_metrics(y_pred, pred_dist, y_val)
            eval_metrics_dict_list.append(eval_metrics_dict)

        eval_metrics_mean, eval_metrics_std = self._aggregate_eval_metrics_across_tasks(eval_metrics_dict_list)
        if print_results:
            message = ""
            for key in eval_metrics_mean:
                message += '- Val-%s: %.3f +- %.3f' % (key, eval_metrics_mean[key], eval_metrics_std[key])
            print('\n\tMeta-Test - Time %.2f sec %s' % (time.time() - t, message))

        return eval_metrics_mean, eval_metrics_std

    def step(self):
        # get meta batches
        meta_batch_x, meta_batch_y, _, _, _ = self.meta_train_sampler.get_meta_batch()
        if self.chaser_loss:
            loss = self._step_tf_chaser(meta_batch_x, meta_batch_y)
        else:
            loss = self._step_tf(meta_batch_x, meta_batch_y)
        return loss

    @tf.function
    def _step_tf(self, meta_batch_x, meta_batch_y):
        # setup various things
        meta_batch_size = self.meta_train_sampler.meta_batch_size
        inner_lr = self.meta_train_config['inner_learning_rate']
        n_inner_steps = self.meta_train_config['n_inner_steps']
        meta_grads_per_task = []

        loss = 0.0

        for i in range(meta_batch_size):
            x_train, y_train, x_val, y_val, n_points_val, n_points_train = self._split_batch(meta_batch_x[i], meta_batch_y[i])

            with tf.GradientTape() as outer_tape:
                outer_tape.watch(self.particles_init)
                particles = self.particles_init

                for _ in range(n_inner_steps):
                    # inner gradient update
                    particles = self._svgd_step(particles, x_train, y_train, inner_lr)

                # compute outer loss & grads
                outer_loss = - self._log_prob(particles, x_val, y_val, n_points_val)

            meta_grad = outer_tape.gradient(outer_loss, self.particles_init)
            meta_grads_per_task.append(meta_grad)
            loss += outer_loss

        loss = loss / meta_batch_size
        # average outer grads across tasks and perform meta-gradient steps
        meta_grads = tf.reduce_mean(tf.stack(meta_grads_per_task, axis=0), axis=0)
        self.optimizer.apply_gradients([(meta_grads, self.particles_init)])
        return loss

    @tf.function
    def _step_tf_chaser(self, meta_batch_x, meta_batch_y):
        # setup various things
        meta_batch_size = self.meta_train_sampler.meta_batch_size
        inner_lr = self.meta_train_config['inner_learning_rate']
        n_inner_steps = self.meta_train_config['n_inner_steps']
        n_leader_steps_extra = self.meta_train_config['n_leader_steps_extra']
        leader_lr_prefactor = self.meta_train_config.get('leader_lr_prefactor', 1.0)
        meta_grads_per_task = []

        assert self.split_batch, "split batch must be enabled for chaser loss"

        loss = 0.0

        for i in range(meta_batch_size):
            x_train, y_train, x_val, y_val, n_points_val, n_points_train = self._split_batch(meta_batch_x[i], meta_batch_y[i])
            x_full, y_full = meta_batch_x[i], meta_batch_y[i]

            # compute leader particles
            particles = self.particles_init
            for _ in range(n_inner_steps + n_leader_steps_extra):
                particles = self._svgd_step(particles, x_full, y_full, lr=leader_lr_prefactor * inner_lr)
            leader_particles = tf.stop_gradient(particles)

            # compute chaser particles & loss
            with tf.GradientTape() as outer_tape:
                outer_tape.watch(self.particles_init)

                # compute chaser particles
                particles = self.particles_init
                for _ in range(n_inner_steps):
                    particles = self._svgd_step(particles, x_train, y_train, lr=inner_lr)
                chaser_particles = particles

                # chaser loss
                outer_loss = tf.reduce_sum((leader_particles - chaser_particles)**2)

            meta_grad = outer_tape.gradient(outer_loss, self.particles_init)
            meta_grads_per_task.append(meta_grad)
            loss += outer_loss

        loss = loss / meta_batch_size
        # average outer grads across tasks and perform meta-gradient steps
        meta_grads = tf.reduce_mean(tf.stack(meta_grads_per_task, axis=0), axis=0)
        self.optimizer.apply_gradients([(meta_grads, self.particles_init)])
        return loss

    def _svgd_step(self, particles, x, y, lr):
        n_points_train = x.shape[-2]
        # compute logprob_grad
        with tf.GradientTape() as inner_tape:
            inner_tape.watch(particles)
            inner_loss = - self._log_prob(particles, x, y, n_points_train)
        grads_log_prob = inner_tape.gradient(inner_loss, particles)
        score = tf.reshape(grads_log_prob, shape=(self.n_particles, -1))

        # compute kernel matrix and grads
        K_XX, K_grad = self._get_kernel_matrix_and_grad(particles)

        # SVGD grads
        svgd_grads_stacked = tf.matmul(K_XX, score) - K_grad / self.n_particles
        grads = tf.reshape(svgd_grads_stacked, shape=particles.shape)
        particles = particles - lr * grads
        return particles

    def _get_kernel_matrix_and_grad(self, stacked_particles):
        p2 = tf.identity(stacked_particles)
        with tf.GradientTape() as tape:
            tape.watch(stacked_particles)
            K_XX = self.kernel(stacked_particles, p2)
        K_grad = tape.gradient(K_XX, stacked_particles)
        return K_XX, K_grad

    def _log_prob(self, particles, x, y, n_train_samples):
        # prior
        log_prior_prob = tf.reshape(self.prior.log_prob(tf.expand_dims(particles, axis=0)), shape=(self.n_particles,))

        # likelihood
        model_params, likelihood_params = self._split_particles(particles)
        y_hat = self.model.forward_parametrized(x, tf.reshape(model_params, (1, -1)))
        if self.model_variables_only:
            log_likelihood = self.likelihood.log_prob(y_hat, y)
        else:
            log_likelihood = self.likelihood.log_prob(y_hat, y, likelihood_params)
        log_likelihood = tf.reshape(log_likelihood, shape=(self.n_particles,))
        return tf.reduce_mean(log_likelihood + (self.prior_factor / n_train_samples) * log_prior_prob)

    def _process_eval_batch(self, x):
        return self.meta_test_sampler.meta_train_tasks[0].process_eval_batch(x, y=None)[0]

    def _split_batch(self, x, y):
        # split data into part for inner and outer update if desired
        if self.split_batch:
            n_points_train = math.ceil(x.shape[1] / 2.0)
            n_points_val = x.shape[1] - n_points_train
            x_train, y_train = x[:, :n_points_train, :], y[:, :n_points_train, :]  # data for inner update
            x_val, y_val = x[:, n_points_train:, :], y[:, n_points_train:, :]  # data for computing meta loss
        else:
            n_points_val = n_points_train = x.shape[1]
            x_train = x_val = x
            y_train = y_val = y
        return x_train, y_train, x_val, y_val, n_points_val, n_points_train

    def _initialize_model(self, model_config, dataset_sampler):
        # setup model
        model = modules.BatchedNNsModel(model_config=model_config, n_batched_models=self.n_particles)
        # init variables
        if hasattr(dataset_sampler, 'get_meta_batch'):
            x, y, _, _, _ = dataset_sampler.get_meta_batch()
            x, y = x[0], y[0]
        elif hasattr(dataset_sampler, 'get_batch'):
            x, y = dataset_sampler.get_batch()
        else:
            raise Exception("Unkown dataset sampler")
        model(x)

        return model

    def _split_particles(self, particles):
        if self.model_variables_only:
            model_params, likelihood_params = particles, None
        else:
            model_params, likelihood_params = tf.split(particles, self.prior.split_sizes, axis=-1)
        return model_params, likelihood_params

