import time

import tensorflow as tf
import numpy as np

import pacoh.modules as modules
from pacoh import util


class BayesianNeuralNetworkMLAP:
    """
    Bayesian Neural Network with approximate inference as suggested in Meta Learning by Adjusting Priors (MLAP)
    """
    def __init__(self, dataset_sampler, model_config, likelihood_config, prior_config, posterior_config,
                 train_config, meta_learned_prior_module, hyper_posterior_config, n_batched_models, random_seed=None):

        # TODO: Construct a base Learner model, that does some initialization and setting
        # TODO: and implements that standard functions marked functions below (already took first stab with algo.py)
        self._set_seed(random_seed)
        self.model_config = self._setup_model_config(model_config, dataset_sampler)

        self.dataset_sampler = dataset_sampler
        self.affine_transform = modules.AffineTransform(normalization_mean=dataset_sampler.y_mean,
                                                        normalization_std=dataset_sampler.y_std)

        # set variables
        self.likelihood_config = likelihood_config
        self.hyper_posterior_config = hyper_posterior_config

        # Instantiate randomized predictor
        batch_sample = dataset_sampler.get_batch()
        prior_config['n_batched_priors'] = 1
        posterior_config['n_batched_priors'] = 1
        self.n_batched_models = n_batched_models
        self.rp = modules.MLAP_RandomisedPredictor(self.model_config, likelihood_config, prior_config, posterior_config,
                                           self.n_batched_models, hyper_posterior_config=hyper_posterior_config,
                                           batch_sample=batch_sample, meta_learned_prior_module=meta_learned_prior_module)

        # Setup basic training config (required for the general fit)
        self.train_config = util.process_training_config(train_config, dataset_sampler)

        # Setup problem specific training_config
        self.n_tasks = 1
        self.n_train_samples = tf.convert_to_tensor(dataset_sampler.n_train_samples, tf.float32)
        self.bound_type = self.train_config['bound_type']
        self.task_complexity_weight = tf.convert_to_tensor(self.train_config['task_complexity_weight'], tf.float32)
        self.delta = tf.convert_to_tensor(self.train_config['delta'], tf.float32)
        self.use_mse_as_loss = self.likelihood_config.get('use_mse_as_loss', False)
        if self.use_mse_as_loss:
            print("NOTE: Using MSE during training")

        self.optimizer = self._setup_optimizer(self.train_config)

        # Setup inference step
        self.step = self._setup_inference_step()
        self.process_eval_batch = self.dataset_sampler.process_eval_batch

    # Custom optimizer for MLAP
    def _setup_optimizer(self, training_config):
        self.trainable_parameters = self.rp.posterior_module.variables
        config = training_config
        optimizers = []

        for n_p in self.trainable_parameters:
            lr = config['learning_rate']

            if 'prior' in n_p.name:
                lr *= config['prior_lr_scale']

            if 'log_var' in n_p.name:
                lr *= config['variance_lr_scale']

            # if config['lr_scheduler'] is 'Polynomial':
            #     lr = tf.optimizers.schedules.PolynomialDecay(decay_steps=self.n_batches * self.epochs,
            #                                                  initial_learning_rate=lr,
            #                                                  end_learning_rate=config['end_learning_rate'])

            if config['optimizer'] == 'adam':
                optimizer = tf.optimizers.Adam(learning_rate=lr)
            elif config['optimizer'] == 'adagrad':
                optimizer = tf.optimizers.Adagrad(learning_rate=lr)

            optimizers.append(optimizer)

        return optimizers

    # Custom for MLAP
    def _setup_inference_step(self):
        assert self.rp.model_variables_only
        assert self.n_tasks == 1
        assert self.bound_type in {'seeger', 'mcallester', 'variational'}

        """ A) Logprob and score function """

        @tf.function
        def mcallester_objective(empirical_risk_per_task, n_samples_per_task):
            """

            Args:
                empirical_risk_per_task (): (n_tasks, )
                n_samples_per_task (): (n_tasks, )
            Returns:

            """
            ##### ===== #####
            # Required artifacts
            ##### ===== #####

            # Fixed
            task_complexity_weight = self.task_complexity_weight
            n_tasks = self.n_tasks # ()
            delta = self.delta #  ()
            hyper_kl = 0.0

            # Requiring computation
            expected_kl_per_task = self.rp.expected_KLs() # (n_tasks, )
            ##### ===== #####

            task_complexity_numerator = (hyper_kl + expected_kl_per_task) + tf.math.log(2 * n_samples_per_task * n_tasks / delta)
            task_complexity_denominator = 2 * (n_samples_per_task - 1)
            task_complexity_per_task = tf.math.sqrt(task_complexity_numerator / task_complexity_denominator) # (n_tasks, )

            empirical_risk_averaged_over_tasks = tf.reduce_mean(empirical_risk_per_task)
            task_complexity_averaged_over_tasks = tf.reduce_mean(task_complexity_per_task)

            loss = tf.reduce_sum([empirical_risk_averaged_over_tasks,
                   task_complexity_weight * task_complexity_averaged_over_tasks])

            return loss

        @tf.function
        def seeger_objective(empirical_risk_per_task, n_samples_per_task):
            """

            Args:
                empirical_risk_per_task (): (n_tasks, )
                n_samples_per_task (): (n_tasks, )
            Returns:

            """
            ##### ===== #####
            # Required artifacts
            ##### ===== #####

            # Fixed
            task_complexity_weight = self.task_complexity_weight
            n_tasks = self.n_tasks  # ()
            delta = self.delta  # ()

            # Requiring computation
            hyper_kl = 0.0  # ()
            expected_kl_per_task = self.rp.expected_KLs()  # (n_tasks, )
            ##### ===== #####

            epsilon_numerator = hyper_kl + expected_kl_per_task + tf.math.log(
                4 * tf.math.sqrt(n_samples_per_task) * n_tasks / delta)
            epsilon = epsilon_numerator / n_samples_per_task
            task_complexity_per_task = 2. * epsilon + tf.sqrt(2. * epsilon * empirical_risk_per_task)


            empirical_risk_averaged_over_tasks = tf.reduce_mean(empirical_risk_per_task)
            task_complexity_averaged_over_tasks = tf.reduce_mean(task_complexity_per_task)

            loss = tf.reduce_sum([empirical_risk_averaged_over_tasks,
                                  task_complexity_weight * task_complexity_averaged_over_tasks])

            return loss

        @tf.function
        def variational_objective(empirical_risk_per_task, n_samples_per_task):
            """

            Args:
                empirical_risk_per_task (): (n_tasks, )
                n_samples_per_task (): (n_tasks, )
            Returns:

            """
            ##### ===== #####
            # Required artifacts
            ##### ===== #####

            # Fixed
            task_complexity_weight = self.task_complexity_weight
            n_tasks = self.n_tasks  # ()

            # Requiring computation
            expected_kl_per_task = self.rp.expected_KLs()  # (n_tasks, )
            ##### ===== #####

            unnormalized_empirical_risk_per_task = empirical_risk_per_task * n_samples_per_task
            task_complexity_per_task = expected_kl_per_task # (n_tasks, )

            total_empirical_risk = tf.reduce_sum(unnormalized_empirical_risk_per_task)
            total_task_complexity = tf.reduce_sum(task_complexity_per_task)

            # Normalize to get on the same scale as with the other bounds, to facilitate hyper-param search
            const = tf.reduce_mean(n_samples_per_task) * n_tasks

            empirical_risk_averaged_over_tasks = total_empirical_risk / const
            task_complexity_averaged_over_tasks = total_task_complexity / const

            loss = tf.reduce_sum([empirical_risk_averaged_over_tasks,
                   task_complexity_weight * task_complexity_averaged_over_tasks])

            return loss

        def get_objective_function():
            if self.bound_type == 'mcallester':
                return mcallester_objective
            elif self.bound_type == 'seeger':
                return seeger_objective
            elif self.bound_type == 'variational':
                return variational_objective
            else:
                raise ValueError('Unknown bound type: {}.'.format(self.bound_type))

        objective_function = get_objective_function()

        @tf.function
        def get_log_prob_and_grad(x, y, n_samples_per_task):
            with tf.GradientTape() as tape:
                tape.watch(self.trainable_parameters)
                mp = self.rp.sample_functions(sample_from_prior=False) # [meta_batch_size * n_samples_used_for_gradient_estimate, n_parameters]

                if self.use_mse_as_loss:
                    empirical_risk_per_task = self.rp.custom_mse_model(mp, x, y)  # [n_tasks, 1]
                else:
                    log_likelihood = self.rp.custom_log_likelihood_model(mp, x, y)  # [meta_batch_size, 1]
                    empirical_risk_per_task = -log_likelihood

                loss = objective_function(empirical_risk_per_task, n_samples_per_task)

            grads = tape.gradient(loss, self.trainable_parameters)
            return empirical_risk_per_task, grads

        self.get_log_prob_and_grad = get_log_prob_and_grad

        def _perform_update(grads):
            for opt, g, tp in zip(self.optimizer, grads, self.trainable_parameters):
                opt.apply_gradients([(g, tp)])

        @tf.function
        def step(x, y):
            empirical_risk_per_task, grads = self.get_log_prob_and_grad(x, y, self.n_train_samples)

            _perform_update(grads)

            return -empirical_risk_per_task

        return step

    # TODO: Abstract all the functions below
    def fit(self):
        n_steps = self.train_config['n_optimization_steps']
        n_iter_per_step = self.train_config['n_iterations_per_step']
        n_total_iterations = self.train_config['n_total_iterations']
        log_period = self.train_config['log_period']

        n_iter = 0

        tf.print("Start training")

        for s in range(n_steps):
            t = time.time()
            for _ in range(n_iter_per_step):
                n_iter += 1
                x, y = self.dataset_sampler.get_batch()
                log_prob = self.step(x, y)
                avg_log_prob = tf.reduce_mean(log_prob).numpy()
                if n_iter % log_period == 0:
                    message = '\n\tIter %d/%d - Train-Time %.2f sec' % (n_iter, n_total_iterations, time.time() - t)
                    message += ' - Train-Log-Prob: %.3f' % avg_log_prob
                    print(message)
                    t = time.time()

            if s + 1 < n_steps:
                # If it is not the last step
                t = time.time()
                message = ""
                eval_metrics_mean, eval_metrics_std = self.eval()
                for key in eval_metrics_mean:
                    if eval_metrics_std is not None:
                        message += ' - Val-%s: %.3f +- %.3f' % (key, eval_metrics_mean[key], eval_metrics_std[key])
                    else:
                        message += ' - Val-%s: %.3f' % (key, eval_metrics_mean[key])

                print('\n\tIter %d/%d - Eval-Time %.2f sec %s' % (n_iter, n_total_iterations, time.time() - t, message))

    def predict(self, x, sample_functions=True, sample_from_prior=False):
        if x.ndim == 2:
            x, _ = self.process_eval_batch(x, None)

        assert x.shape[0] == self.rp.n_batched_models
        y_pred, pred_dist = self.rp.predict(x, sample_functions, sample_from_prior)

        # unnormalize predictions
        y_pred = tf.multiply(y_pred, self.dataset_sampler.y_std) + self.dataset_sampler.y_mean
        pred_dist = self.affine_transform.apply(pred_dist)

        return y_pred, pred_dist

    def eval(self):
        x, y_true = self.dataset_sampler.val_data
        y_pred, pred_dist = self.predict(x)

        o = self.rp.likelihood.calculate_eval_metrics(y_pred, pred_dist, y_true)
        if self.train_config['plot_functions']:
            import matplotlib.pyplot as plt
            fig, (ax) = plt.subplots(1, 1, figsize=(4, 4))
            # self.dataset.plot_data(train=True, val=False, ax=ax)
            self.dataset_sampler.plot_prediction_functions(model=self, ax=ax,
                                                           plot_pred_std=self.train_config['plot_pred_std'],
                                                           plot_pred_lines=self.train_config['plot_pred_lines'],
                                                           plot_data=True)
            plt.legend()
            plt.show()

        return o, None

    @staticmethod
    def _setup_model_config(model_config, dataset_sampler):
        if 'type' in model_config:
            # determines the NN class and, if necessary, adds image parameters (height, width, depth) to the config
            if model_config['type'] == 'cnn':
                model_config = model_config[model_config['type']]
                # for key in ['img_height', 'img_width', 'img_depth']:
                #     model_config[key] = getattr(dataset_sampler, key)
                model_config['model_class'] = modules.CNN
            elif model_config['fc']:
                model_config = model_config[model_config['type']]
                model_config['model_class'] = modules.FullyConnectedNN
            else:
                raise AssertionError('model type must be either \'cnn\' or \'fc\'')
        else:
            # make sure that the NN class is already specified in the config
            assert 'model_class' in model_config

        return model_config


    def _calib_error(self, pred_dist, y_true):
        # cdf returns a tensor of dimensions (n_batched_models, 1)
        cdf_vals = pred_dist.cdf(y_true)

        if y_true.ndim == 3:
            y_true = y_true[0]

        num_points = y_true.shape[1]
        conf_levels = tf.linspace(0.05, 0.95, 20)
        emp_freq_per_conf_level = tf.reduce_sum(tf.cast(cdf_vals[:, None] <= conf_levels, tf.float32),
                                                axis=0) / num_points

        calib_rmse = tf.sqrt(tf.reduce_mean((emp_freq_per_conf_level - conf_levels) ** 2))
        return calib_rmse

    # Data related methods
    def _prepare_batch(self, task):
        sampler = task.train_batch_sampler
        batch = tf.data.experimental.get_next_as_optional(sampler)
        x, y = batch.get_value()

        x.set_shape(x.shape)
        y.set_shape(y.shape)
        x, y = self._standardize(x, y)

        return x, y, tf.cast(task.n_train_samples, tf.float32), tf.cast(task.batch_size, tf.float32)

    # Randomness
    def _set_seed(self, random_seed):
        if random_seed is not None:
            np.random.seed(random_seed + 1)
            tf.random.set_seed(random_seed + 2)