import time

import tensorflow as tf
import tensorflow_addons as tfa
import numpy as np

import pacoh.modules as modules
from pacoh import util


class BayesianNeuralNetworkSVGD:
    """
    Bayesian Neural Network with approximate inference via Stein Variational Gradient Descent (SVGD)
    """
    def __init__(self, dataset_sampler, model_config, likelihood_config, prior_config, posterior_config,
                 train_config, meta_learned_prior_module=None, random_seed=None):

        self._set_seed(random_seed)
        self.model_config = self._setup_model_config(model_config, dataset_sampler)

        self.dataset_sampler = dataset_sampler
        self.affine_transform = modules.AffineTransform(normalization_mean=dataset_sampler.y_mean,
                                                        normalization_std=dataset_sampler.y_std)

        # batch_sample = tf.data.experimental.get_next_as_optional(self.dataset_sampler.train_batch_sampler).get_value()
        batch_sample = dataset_sampler.get_batch()
        self.rp = modules.RandomisedPredictor(self.model_config, likelihood_config, prior_config, posterior_config,
                                              batch_sample, meta_learned_prior_module=meta_learned_prior_module)


        # Setup basic training config
        self.train_config = util.process_training_config(train_config, dataset_sampler)

        self.prior_prob_weight = self.train_config['prior_prob_weight']
        self.optimizer = self._setup_optimizer(self.train_config)
        self.n_train_samples = tf.convert_to_tensor(self.dataset_sampler.n_train_samples, dtype=tf.float32)

        self.kernel = modules.RBF_Kernel_TF(self.train_config['bandwidth'])
        self.process_eval_batch = self.dataset_sampler.process_eval_batch

    def fit(self):
        n_steps = self.train_config['n_optimization_steps']
        n_iter_per_step = self.train_config['n_iterations_per_step']
        n_total_iterations = self.train_config['n_total_iterations']
        log_period = self.train_config['log_period']
        n_iter_fixed_likelihood = self.train_config.get('n_iter_fixed_likelihood', 0)

        n_iter = 0

        tf.print("Start training")

        for s in range(n_steps):
            t = time.time()
            for _ in range(n_iter_per_step):
                n_iter += 1
                stop_likelihood_grads = n_iter < n_iter_fixed_likelihood
                x, y = self.dataset_sampler.get_batch()
                log_prob = self.step(x, y, stop_likelihood_grads)
                avg_log_prob = tf.reduce_mean(log_prob).numpy()
                if n_iter % log_period == 0:
                    message = '\n\tIter %d/%d - Train-Time %.2f sec' % (n_iter, n_total_iterations, time.time() - t)
                    message += ' - Train-Log-Prob: %.3f' % avg_log_prob
                    print(message)
                    t = time.time()

            if s + 1 < n_steps:
                # If it is not the last step
                t = time.time()
                message = ""
                eval_metrics_mean, eval_metrics_std = self.eval()
                for key in eval_metrics_mean:
                    if eval_metrics_std is not None:
                        message += ' - Val-%s: %.3f +- %.3f' % (key, eval_metrics_mean[key], eval_metrics_std[key])
                    else:
                        message += ' - Val-%s: %.3f' % (key, eval_metrics_mean[key])

                print('\n\tIter %d/%d - Eval-Time %.2f sec %s' % (n_iter, n_total_iterations, time.time() - t, message))

    def predict(self, x, sample_functions=True, sample_from_prior=False):
        if x.ndim == 2:
            x, _ = self.process_eval_batch(x, None)

        # standardize
        # x = tf.divide((x - self.dataset_sampler.x_mean), self.dataset_sampler.x_std)
        assert x.shape[0] == self.rp.n_batched_models
        y_pred, pred_dist = self.rp.predict(x, n_samples=None, sample_functions=sample_functions,
                                            sample_from_prior=sample_from_prior)

        # unnormalize predictions
        y_pred = tf.multiply(y_pred, self.dataset_sampler.y_std) + self.dataset_sampler.y_mean
        pred_dist = self.affine_transform.apply(pred_dist)

        return y_pred, pred_dist

    def eval(self):
        x, y_true = self.dataset_sampler.val_data
        y_pred, pred_dist = self.predict(x)

        o = self.rp.likelihood.calculate_eval_metrics(y_pred, pred_dist, y_true)
        if self.train_config['plot_functions']:
            import matplotlib.pyplot as plt
            fig, (ax) = plt.subplots(1, 1, figsize=(4, 4))
            # self.dataset.plot_data(train=True, val=False, ax=ax)
            self.dataset_sampler.plot_prediction_functions(model=self, ax=ax,
                                                           plot_pred_std=self.train_config['plot_pred_std'],
                                                           plot_pred_lines=self.train_config['plot_pred_lines'],
                                                           plot_data=True)
            plt.legend()
            plt.show()

        return o, None

    def _calib_error(self, pred_dist, y_true):
        # cdf returns a tensor of dimensions (n_batched_models, 1)
        cdf_vals = pred_dist.cdf(y_true)

        if y_true.ndim == 3:
            y_true = y_true[0]

        num_points = y_true.shape[1]
        conf_levels = tf.linspace(0.05, 0.95, 20)
        emp_freq_per_conf_level = tf.reduce_sum(tf.cast(cdf_vals[:, None] <= conf_levels, tf.float32),
                                                axis=0) / num_points

        calib_rmse = tf.sqrt(tf.reduce_mean((emp_freq_per_conf_level - conf_levels) ** 2))
        return calib_rmse

    def _setup_optimizer(self, training_config):
        if training_config['optimizer'] == 'adam':
            self.optimizer = tf.keras.optimizers.Adam(learning_rate=training_config['learning_rate'])
        elif training_config['optimizer'] == 'adamw':
            self.optimizer = tfa.optimizers.AdamW(learning_rate=training_config['learning_rate'],
                                                  weight_decay=training_config['weight_decay'])
        else:
            raise Exception("Unexpected optimizer choice")

        return self.optimizer

    @tf.function
    def step(self, x, y, stop_likelihood_grads=False):
        log_prob, grads_log_prob = self._get_log_prob_and_grad(x, y)
        score = grads_log_prob[0]

        K_XX, grad_K_XX = self._get_kernel_matrix_and_grad()
        inter = tf.matmul(K_XX, score)

        phi = (inter - grad_K_XX) / self.rp.posterior_config['n_samples']
        # / log_prob.shape[0]
        grads = -phi

        if stop_likelihood_grads and not self.rp.model_variables_only:
            ones_shape = grads.shape[:-1] + (self.rp.posterior_module.split_sizes[0],)
            zeros_shape = grads.shape[:-1] + (self.rp.posterior_module.split_sizes[1],)
            mask = tf.concat([tf.ones(ones_shape), tf.zeros(zeros_shape)], axis=-1)
            tf.assert_equal(mask.shape, grads.shape)
            grads *= mask

        self.optimizer.apply_gradients([(grads, self.rp.posterior_module.variables[0])])
        return log_prob

    @tf.function
    def _get_log_prob_and_grad(self, x, y):
        # There is only, one dataset_sampler so the number of samples doesn't need to come from the batch as we know it
        with tf.GradientTape() as tape:
            tape.watch(self.rp.trainable_variables())

            # sample params
            param_sample = self.rp.sample_params()

            # compute unnormalized posterior
            log_likelihood = self.rp.compute_likelihood(x, y, param_sample, prefactor=self.n_train_samples)
            prior_log_prob = self.rp.prior_log_prob(param_sample)
            log_prob = log_likelihood + self.prior_prob_weight * prior_log_prob

        grads_log_prob = tape.gradient(log_prob, self.rp.trainable_variables())
        return log_prob, grads_log_prob

    @tf.function
    def _get_kernel_matrix_and_grad(self):
        p2 = tf.identity(self.rp.posterior_module.particles)
        with tf.GradientTape() as tape:
            tape.watch(self.rp.posterior_module.particles)
            K_XX = self.kernel(self.rp.posterior_module.particles, p2)
        K_grad = tape.gradient(K_XX, self.rp.posterior_module.particles)
        return K_XX, K_grad


    def _setup_model_config(self, model_config, dataset_sampler):
        if 'type' in model_config:
            # determines the NN class and, if necessary, adds image parameters (height, width, depth) to the config
            if model_config['type'] == 'cnn':
                model_config = model_config[model_config['type']]
                # for key in ['img_height', 'img_width', 'img_depth']:
                #     model_config[key] = getattr(dataset_sampler, key)
                model_config['model_class'] = modules.CNN
            elif model_config['fc']:
                model_config = model_config[model_config['type']]
                model_config['model_class'] = modules.FullyConnectedNN
            else:
                raise AssertionError('model type must be either \'cnn\' or \'fc\'')
        else:
            # make sure that the NN class is already specified in the config
            assert 'model_class' in model_config

        return model_config

    def _set_seed(self, random_seed):
        if random_seed is not None:
            np.random.seed(random_seed + 1)
            tf.random.set_seed(random_seed + 2)