import tensorflow as tf
import numpy as np
import argparse
import functools

import pacoh.algorithms as algorithms
from pacoh.datasets.data_sampler import DatasetSampler
from pacoh.algorithms.BNN_SVGD import BayesianNeuralNetworkSVGD

""" Config """

LAYER_SIZE = 64
N_LAYERS = 4
ACTIVATION = 'relu'
OUTPUT_DIM = 1

config = {
'likelihood_config': {'class': {'learn_likelihood_variables': False},
                     'reg': {'log_var': 2 * np.log(0.01),
                             'learn_likelihood_variables': True},
                     'type': 'reg'
                     },

'model_config': {'fc': {'input_dim': None,
                       'output_dim': OUTPUT_DIM,
                       'layer_units': [LAYER_SIZE for _ in range(N_LAYERS)] + [OUTPUT_DIM],
                       'activations': [ACTIVATION for _ in range(N_LAYERS)] + [None]
                       },
                'type': 'fc'},

'hyper_prior_config': {'mean_mean': 0.0,  # mean of the hyper-prior for the means
                      'bias_mean_std': 3,  # std of the hyper-prior for the mean of the biases
                      'kernel_mean_std': 1,  # std of the hyper-prior for the mean of the kernels
                      'log_var_mean': -0.5,
                      'bias_log_var_std': 1,  # std of the hyper-prior for the log-variances of the prior
                      'kernel_log_var_std': 1,  # std of the hyper-prior for the log-variances of the prior
                      'likelihood_log_var_mean_mean': -4,
                      'likelihood_log_var_mean_std': 1,
                      'likelihood_log_var_log_var_mean': -2,
                      'likelihood_log_var_log_var_std': 0.1
                      },

'prior_config': {'gaussian': {'model': {'mean_mu': 0.0,
                                        'bias_log_var_mu': 2 * np.log(1.0),
                                        'weight_log_var_mu': 2 * np.log(0.5)  # 2 * np.log(0.5)
                                        },
                             'likelihood': {'log_var_mu': 2 * np.log(0.01),
                                            'log_var_log_var': 2 * np.log(2.0)
                                           },
                             'init_std': 0.2,
                             'add_noise': True,
                             'init_from_models': True,
                             'n_samples': 5,  # number of samples from prior for estimating the mll
                             },
                'type': 'gaussian'
                },

'hyper_posterior_config': {'particle': {'n_samples': 3,
                                       },
                              'type': 'particle'
                              },

'posterior_config': {'n_samples': 20,
                    'use_init_std': True,
                    'type': 'particle'
                    },

'meta_train_config': {'n_total_iterations': 4000,
                   'n_train_loss_logs': 4,
                   'n_eval_runs': 0,
                   'optimizer': 'adam',
                   'learning_rate': 0.002,
                   'plot_functions': False,
                   'plot_pred_lines': True,
                   'plot_pred_std': True,
                   'prior_prob_weight': 1.0,
                   'bandwidth': 1.0
                   },

'meta_test_config': {'n_total_iterations': 2000,
                     'n_train_loss_logs': 5,
                     'n_eval_runs': 0,
                     'optimizer': 'adam',
                     'learning_rate': 0.01,
                     'plot_functions': True,
                     'plot_pred_lines': True,
                     'plot_pred_std': True,
                     'prior_prob_weight': 1.0,
                     'bandwidth': 1000.
                     }

}

prior_config = config['prior_config']
config['prior_config'] = prior_config = prior_config[prior_config['type']]

posterior_config = config['posterior_config']

hyper_posterior_config = config['hyper_posterior_config']
config['hyper_posterior_config'] = hyper_posterior_config = hyper_posterior_config[
    hyper_posterior_config['type']]


def setup_meta_learner(meta_train_data, meta_test_data, config, random_seed=22):
    model_config = config['model_config']
    likelihood_config = config['likelihood_config']
    hyper_prior_config = config['hyper_prior_config']
    training_config = config['meta_train_config']
    inference_training_config = config['meta_test_config']

    #tf.config.experimental_run_functions_eagerly(True)

    # setup data set and corresponding samplers
    from pacoh.datasets.data_sampler import MetaDatasetSampler, DatasetSampler

    n_batched_models_train = hyper_posterior_config['n_samples'] * prior_config['n_samples']
    n_batched_models_test = hyper_posterior_config['n_samples'] * posterior_config['n_samples']

    # Setup samplers
    meta_train_sampler = MetaDatasetSampler(meta_train_data, per_task_batch_size=5, meta_batch_size=5,
                                            n_batched_models=n_batched_models_train, tiled=True,
                                            standardization=True, flatten_data=False)

    meta_test_sampler = MetaDatasetSampler(meta_test_data, per_task_batch_size=5,
                                           n_batched_models=n_batched_models_test, tiled=True,
                                           standardization=True, flatten_data=False)

    meta_test_sampler.copy_standardization_stats(meta_train_sampler)
    meta_test_sampler._update_task_standardization_stats()


    meta_reg_model = algorithms.MetaBayesianNeuralNetwork_SVGD(meta_train_sampler, meta_test_sampler, model_config,
                                                               likelihood_config,
                                                               prior_config, posterior_config,
                                                               hyper_prior_config, hyper_posterior_config,
                                                               training_config, inference_training_config,
                                                               plot_prior_during_training=False, random_seed=random_seed)


    return meta_reg_model


def fit_bnn(x_train, y_train, config, meta_reg_model):
    assert x_train.ndim == y_train.ndim == 2

    if meta_reg_model is None:
        n_batched_models = config['posterior_config']['n_samples']
        config['prior_config']['n_batched_models'] = n_batched_models
        config['prior_config']['n_batched_priors'] = 1
    else:
        n_batched_models = config['posterior_config']['n_samples'] * config['hyper_posterior_config']['n_samples']

    dataset_sampler = DatasetSampler(train_data=(x_train, y_train), val_data=(x_train, y_train),
                                     n_batched_models=n_batched_models,
                                     batch_size=5, tiled=True)


    if meta_reg_model is None:
        reg_model = BayesianNeuralNetworkSVGD(dataset_sampler,
                                              config['model_config'],
                                              config['likelihood_config'],
                                              config['prior_config'],
                                              config['posterior_config'],
                                              config['meta_test_config'],
                                              meta_learned_prior_module=None)
    else:
        dataset_sampler.copy_standardization_stats(meta_reg_model.meta_train_sampler)
        reg_model = meta_reg_model.meta_test_inference_class(dataset_sampler,
                                                             meta_reg_model.model_config,
                                                             meta_reg_model.likelihood_config,
                                                             meta_reg_model.prior_config,
                                                             meta_reg_model.posterior_config,
                                                             meta_reg_model.meta_test_config,
                                                             meta_learned_prior_module=meta_reg_model.rp.prior_module)

    reg_model.fit()

    return reg_model


def select_action(reg_model, bandit_task, mode='thomson_sampling'):
    y_pred, pred_dist = reg_model.predict(bandit_task.x_domain)
    if mode == 'thomson_sampling':
        # 1) sample from posterior
        posterior_particle = y_pred[np.random.choice(y_pred.shape[0])]
        # 2) take the argmax as action
        action = np.argmax(posterior_particle)
    elif mode == 'ucb':
        # 1) compute ucb
        ucb = pred_dist.mean() + 2 * pred_dist.stddev()
        # 2) take the argmax as action
        action = np.argmax(ucb)
    else:
        raise NotImplementedError

    return action


def bandit_loop(bandit_task, config, meta_reg_model, n_iter=20, mode='thomson_sampling', plot_pred_dist=False, random_seed=None):
    np.random.seed(random_seed)
    tf.random.set_seed(random_seed + 1)

    action_list = []
    reward_list = []
    avg_regret_list = []
    max_regret_list = []


    for iter in range(n_iter):
        if iter < 1:
            # random action
            action = np.random.choice(bandit_task.n_arms)
        else:
            x_train = bandit_task._action_to_x(action_list)
            y_train = np.expand_dims(np.array(reward_list), axis=-1)

            reg_model = fit_bnn(x_train, y_train, config, meta_reg_model)

            action = select_action(reg_model, bandit_task, mode=mode)

            if plot_pred_dist:
                y_pred, pred_dist = reg_model.predict(bandit_task.x_domain)
                assert bandit_task.x_domain.shape[-1] == 1
                from matplotlib import pyplot as plt

                fig, (ax) = plt.subplots(1, 1, figsize=(4, 4))
                ax.plot(bandit_task.x_domain, pred_dist.mean())
                x_plot = tf.squeeze(bandit_task.x_domain)
                ucb = tf.squeeze(pred_dist.mean() + pred_dist.stddev())
                lcb = tf.squeeze(pred_dist.mean() - pred_dist.stddev())
                ax.fill_between(x_plot, lcb, ucb, alpha=0.2)

                for i in range(y_pred.shape[0]):
                    ax.plot(x_plot, tf.reshape(y_pred[i], (1, -1))[0], color='green', alpha=0.3, linewidth=1)

                plt.scatter(x_train, y_train)
                plt.show()


        action_list.append(action)
        reward_list.append(bandit_task.noisy_reward(action))

        max_regret_list.append(bandit_task.maximum_regret(reward_list, action_list))
        avg_regret_list.append(bandit_task.average_regret(action_list))
        print('Iter %i - Action: %.2f - Avg-Regret %.4f - Max-Regret %.4f' % (iter, float(bandit_task._action_to_x(action)[0]),
                                                                              float(avg_regret_list[-1]), float(max_regret_list[-1])))

    return action_list, reward_list, max_regret_list, avg_regret_list


def main(args, config):
    """ Bandit definition """
    from matplotlib import pyplot as plt
    import numpy as np

    """ Setup Bandit """
    from pacoh.datasets.bandits import MetaCauchyBandit, MetaMHCBandit
    bandit_task = MetaCauchyBandit()

    meta_train_data = bandit_task.generate_meta_train_data()
    meta_test_data = bandit_task.generate_meta_test_data()

    """ Meta-Learn prior """
    if args.meta_learn:
        meta_reg_model, meta_train_sampler, meta_test_sampler = setup_meta_learner(meta_train_data, meta_test_data, config)
        meta_reg_model.meta_fit()
    else:
        meta_reg_model = None

    """ Bandit loop """

    action_list, reward_list, max_regret_list, avg_regret_list = bandit_loop(bandit_task, config, meta_reg_model,
                                                                             plot_pred_dist=True,
                                                                             n_iter=10, random_seed=65)

    plt.plot(range(len(max_regret_list)), max_regret_list)
    plt.title('max regret')
    plt.show()

    plt.plot(range(len(avg_regret_list)), avg_regret_list)
    plt.title('avg regret')
    plt.show()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Bayesian optimization with meta-learned prior')
    parser.add_argument("--meta_learn", type=bool, default=True)
    args = parser.parse_args()
    main(args, config)



