import tensorflow as tf
import numpy as np
import argparse
import functools

import pacoh.algorithms as algorithms
from pacoh.datasets.data_sampler import DatasetSampler
from pacoh.algorithms.BNN_SVGD import BayesianNeuralNetworkSVGD

from pacoh_gp.GPR_meta_mll import GPRegressionMetaLearned
""" Config """


config = {
    'learning_mode': 'both',
    'covar_module': 'SE',
    'mean_module': 'constant',
    'num_iter_fit': 10,
    'lr_params': 1e-2,
}
kernel = 'linear'

def setup_meta_learner(meta_train_data, meta_test_data, config, random_seed=22):
    meta_reg_model = GPRegressionMetaLearned(meta_train_data, **config, random_seed=random_seed)
    return meta_reg_model


def select_action(reg_model, bandit_task, x_train, y_train, mode='thomson_sampling'):
    if x_train.shape[0] == 1:
        # gpytorch can't deal with 1 training point --> use 2x the same point
        x_train = np.concatenate([x_train, x_train], axis=0)
        y_train = np.concatenate([y_train, y_train], axis=0)

    if mode == 'thomson_sampling':
        pred_dist = reg_model.predict(x_train, y_train, bandit_task.x_domain, return_density=True,
                                      return_posterior_f=True)
        # 1) sample from posterior
        posterior_particle = pred_dist.sample()
        # 2) take the argmax as action
        action = np.argmax(posterior_particle)
    elif mode == 'ucb':
        pred_dist = reg_model.predict(x_train, y_train, bandit_task.x_domain, return_density=True,
                                      return_posterior_f=False)
        # 1) compute ucb
        ucb = (pred_dist.mean + 2 * pred_dist.stddev).detach().numpy()
        # 2) take the argmax as action
        action = np.argmax(ucb)
    else:
        raise NotImplementedError

    return action


def bandit_loop(bandit_task, reg_model, n_iter=20, mode='thomson_sampling', plot_pred_dist=False, random_seed=None):
    np.random.seed(random_seed)
    tf.random.set_seed(random_seed + 1)

    action_list = []
    reward_list = []
    avg_regret_list = []
    max_regret_list = []

    for iter in range(n_iter):
        if iter < 1:
            # random action
            action = np.random.choice(bandit_task.n_arms)
        else:
            x_train = bandit_task._action_to_x(action_list)
            y_train = np.array(reward_list)

            action = select_action(reg_model, bandit_task, x_train, y_train, mode=mode)

            if plot_pred_dist and iter > 1:
                pred_mean, pred_std = reg_model.predict(x_train, y_train, bandit_task.x_domain)
                assert bandit_task.x_domain.shape[-1] == 1
                from matplotlib import pyplot as plt

                fig, (ax) = plt.subplots(1, 1, figsize=(4, 4))
                ax.plot(bandit_task.x_domain, pred_mean)
                x_plot = tf.squeeze(bandit_task.x_domain)
                ucb = tf.squeeze(pred_mean + 2 * pred_std)
                lcb = tf.squeeze(pred_mean - 2 * pred_std)
                ax.fill_between(x_plot, lcb, ucb, alpha=0.2)
                plt.scatter(x_train, y_train)
                plt.show()


        action_list.append(int(action))
        reward_list.append(float(bandit_task.noisy_reward(action)))

        max_regret_list.append(float(bandit_task.maximum_regret(reward_list, action_list)))
        avg_regret_list.append(float(bandit_task.average_regret(action_list)))
        print('Iter %i - Action: %.2f - Avg-Regret %.4f - Max-Regret %.4f' % (iter, float(bandit_task._action_to_x(action)[0]),
                                                                              float(avg_regret_list[-1]), float(max_regret_list[-1])))

    return action_list, reward_list, max_regret_list, avg_regret_list


def main(args, config):
    """ Bandit definition """
    from matplotlib import pyplot as plt
    import numpy as np

    """ Setup Bandit """
    from pacoh.datasets.bandits import MetaCauchyBandit, MetaMHCBandit
    bandit_task = MetaMHCBandit()
    #bandit_task = MetaCauchyBandit()

    meta_train_data = bandit_task.generate_meta_train_data()
    meta_test_data = bandit_task.generate_meta_test_data()

    meta_reg_model = setup_meta_learner(meta_train_data, meta_test_data, config)
    """ Meta-Learn prior """
    if args.meta_learn:
        meta_reg_model.meta_fit(log_period=200)
    else:
        assert config['learning_mode'] == 'vanilla'

    """ Bandit loop """

    action_list, reward_list, max_regret_list, avg_regret_list = bandit_loop(bandit_task, meta_reg_model,
                                                                             mode='thomson_sampling',
                                                                             plot_pred_dist=False,
                                                                             n_iter=50, random_seed=65)

    plt.plot(range(len(max_regret_list)), max_regret_list)
    plt.title('max regret')
    plt.show()

    plt.plot(range(len(avg_regret_list)), avg_regret_list)
    plt.title('avg regret')
    plt.show()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Bayesian optimization with meta-learned prior')
    parser.add_argument("--meta_learn", type=bool, default=True)
    args = parser.parse_args()
    main(args, config)



