import math

import numpy as np
import tensorflow as tf

from pacoh import util as util

from .meta_algo import MetaLearner

class MAML(MetaLearner):
    def __init__(self, meta_train_sampler, meta_test_sampler, model_config, likelihood_config, meta_train_config,
                 meta_test_config, random_seed=None):

        super().__init__(meta_train_sampler, meta_test_sampler, model_config, random_seed=random_seed)

        # set variables
        self.meta_train_sampler = meta_train_sampler
        self.meta_test_sampler = meta_test_sampler

        self.likelihood_config = likelihood_config
        self.meta_train_config = meta_train_config
        self.meta_test_config = meta_test_config

        # initialize relevant components / modules of the algorithm
        self.model = self._initialize_model(self.model_config, self.meta_train_sampler)
        self.likelihood = util.get_likelihood(config=likelihood_config, output_dim=self.model_config['output_dim'],
                                              n_batched_models=1)

        self._setup_optimizer(meta_train_config)

        print("\nTraining optimization parameters:")
        print(f"Meta batch size of: {self.meta_train_sampler.meta_batch_size}")
        self.meta_train_config = util.process_training_config(self.meta_train_config, meta_train_sampler)

    def predict(self, x, sample_functions=False, sample_from_prior=False):
        """
        TODO
        """
        x = self._process_eval_batch(x)
        y_pred = self.model(x)
        y_pred = self._unstandardize(y_pred)

        y_pred = tf.expand_dims(y_pred, axis=0)
        return y_pred, None

    def plot_predict(self, n_tasks=3):
        from matplotlib import pyplot as plt

        tasks = self.meta_test_sampler.tasks[:n_tasks]
        fig, axes = plt.subplots(nrows=1, ncols=n_tasks)
        x_plot = tf.expand_dims(tf.range(tasks[0].x_min - 0.5, tasks[0].x_max + 0.5, delta=0.1), axis=-1)
        for i, task in enumerate(tasks):
            y_hat_pre = self._unstandardize(self.model(self._standardize(x_plot)))
            axes[i].scatter(*map(lambda x: x.numpy(), task.val_data), color='blue')
            axes[i].scatter(*map(lambda x: x.numpy(), task.train_data), color='orange')
            axes[i].plot(x_plot.numpy(), y_hat_pre.numpy(), label='pre')

            x_context, y_context = task.train_data
            y_hat_post = self.meta_predict(x_context, y_context, x_plot)
            axes[i].plot(x_plot.numpy(), y_hat_post.numpy(), label='post', color='orange')
        fig.legend()
        fig.show()

    def meta_predict(self, x_context, y_context, x_test):
        test_model = self._copy_model(self.model, x_context, name='test_model')
        lr = self.meta_train_config['inner_learning_rate']

        x_context, y_context = self._standardize(x_context, y_context)

        # perform adaptation step with context data
        for _ in range(self.meta_test_config['n_optimization_steps']):
            test_model, _ = self._grad_step(test_model, x_context, y_context, lr)

        # predict with adapted model
        x_test = self._standardize(x_test)
        y_hat = test_model(x_test)
        return self._unstandardize(y_hat)

    def step(self):
        # get meta batches
        meta_batch_x, meta_batch_y, _, _, _ = self.meta_train_sampler.get_meta_batch()
        loss = self._step_tf(meta_batch_x, meta_batch_y)
        return loss

    @tf.function
    def _step_tf(self, meta_batch_x, meta_batch_y):
        # setup various things
        meta_batch_size = self.meta_train_sampler.meta_batch_size
        inner_lr = self.meta_train_config['inner_learning_rate']
        n_inner_steps = self.meta_train_config['n_inner_steps']
        first_order = self.meta_train_config['first_order']
        split_batch = self.meta_train_config.get('split_batch', False)
        meta_grads_per_task = []

        loss = 0.0

        for i in range(meta_batch_size):
            x, y = meta_batch_x[i], meta_batch_y[i]

            # split data into part for inner and outer update if desired
            if split_batch:
                idx_split = math.ceil(x.shape[0] / 2.0)
                x_inner, y_inner = x[:idx_split], y[:idx_split]  # data for inner update
                x_outer, y_outer = x[idx_split:], y[idx_split:]  # data for computing meta loss
            else:
                x_inner = x_outer = x
                y_inner = y_outer = y

            task_model = self._copy_model(self.model, x, name='nn_copy_%i' % i)
            model_params_init = self.model.get_variables_vectorized()

            with tf.GradientTape() as outer_tape:
                if not first_order:
                    outer_tape.watch(model_params_init)
                model_params = model_params_init

                for _ in range(n_inner_steps):
                    # inner gradient update
                    with tf.GradientTape() as inner_tape:
                        inner_tape.watch(model_params)
                        y_hat = task_model.forward_parametrized(x_inner, model_params)
                        inner_loss = self._loss_fn(y_hat, y_inner)
                    inner_grads = inner_tape.gradient(inner_loss, model_params)
                    model_params = model_params - inner_lr * inner_grads

                # compute outer loss & grads
                if first_order:
                    outer_tape.watch(model_params)
                y_hat = task_model.forward_parametrized(x_outer, model_params)
                outer_loss = self._loss_fn(y_hat, y_outer)

            if first_order:
                meta_grad = outer_tape.gradient(outer_loss, model_params)
            else:
                meta_grad = outer_tape.gradient(outer_loss, model_params_init)
            meta_grads_per_task.append(meta_grad)
            loss += outer_loss


        loss = loss / meta_batch_size
        # average outer grads across tasks and perform meta-gradient steps
        meta_grads_vectorized = tf.reduce_mean(tf.concat(meta_grads_per_task, axis=0), axis=0, keepdims=True)
        meta_grads = self.model.split_and_unvectorize_grads(meta_grads_vectorized)
        self.optimizer.apply_gradients(zip(meta_grads, self.model.trainable_variables))
        return loss

    def _grad_step(self, model, x, y, lr):
        model_params = model.get_variables_vectorized()

        with tf.GradientTape() as tape:
            tape.watch(model_params)
            y_hat = model.forward_parametrized(x, model_params)
            loss = self._loss_fn(y_hat, y)

        model_params = model_params - lr * tape.gradient(loss, model_params)
        model.set_variables_vectorized(model_params)
        return model, loss

    def _loss_fn(self, y_hat, y):
        y_hat, y = tf.expand_dims(y_hat, axis=0), tf.expand_dims(y, axis=0)
        loss = - tf.reduce_mean(self.likelihood.log_prob(y_hat, y))
        return loss

    def _setup_meta_test_models_and_step(self, meta_test_tasks):
        # make copies of self.model
        eval_models = []
        for i, task in enumerate(meta_test_tasks):
            x, y = task.get_batch()
            eval_model = self._copy_model(self.model, x, name='nn_eval_%i'%i)
            eval_models.append(eval_model)

        lr = self.meta_train_config['inner_learning_rate']

        def eval_models_step():
            losses = []
            for task, eval_model in zip(meta_test_tasks, eval_models):
                x, y = task.get_batch()
                eval_model, loss = self._grad_step(eval_model, x, y, lr)
                losses.append(loss)
            return losses
        return eval_models, eval_models_step

    def _meta_test_models_eval(self, test_tasks, eval_models):
        if hasattr(self.likelihood, 'rmse'): #regression
            eval_metric_name = 'avg_rmse'
            eval_metric_per_task = []
            for task, eval_model in zip(test_tasks, eval_models):
                # get validation data
                x, y = task.val_data

                # standardize predict and un-standardize prediction
                x = self._process_eval_batch(x)
                y_hat = eval_model(x)
                y_hat = self._unstandardize(y_hat)

                # compute eval metric
                eval_metric_per_task.append(self.likelihood.rmse(y_hat, y))

            eval_metrics_mean = {eval_metric_name: np.mean(eval_metric_per_task)}
            eval_metrics_std = {eval_metric_name: np.std(eval_metric_per_task)}
            eval_metrics_grouped = {eval_metric_name: eval_metric_per_task}

        elif hasattr(self.likelihood, 'accuracy'): #classification
            from pacoh.experiments.hyperparam_search.util import _aggregate_eval_metrics_across_tasks
            eval_metrics_grouped = {}
            for task, eval_model in zip(test_tasks, eval_models):
                # get validation data
                x, y = task.val_data

                # standardize predict and un-standardize prediction
                x = self._process_eval_batch(x)
                y_hat = eval_model(x)
                y_hat = self._unstandardize(y_hat)
                eval_metrics_dict = self.likelihood.calculate_eval_metrics(tf.expand_dims(y_hat, axis=0), None, y)
                for key in eval_metrics_dict:
                    l = eval_metrics_grouped.get(key, [])
                    l.append(eval_metrics_dict[key].numpy())
                    eval_metrics_grouped[key] = l

            eval_metrics_mean = {}
            eval_metrics_std = {}
            for key in eval_metrics_grouped:
                eval_metrics_mean[key] = np.mean(eval_metrics_grouped[key])
                eval_metrics_std[key] = np.std(eval_metrics_grouped[key])
        else:
            raise NotImplementedError('meta test eval only supports rmse or accuracy')

        return eval_metrics_mean, eval_metrics_std, eval_metrics_grouped

    def _meta_test_training_loop(self, test_tasks, eval_models, eval_models_step):
        n_grad_steps = self.meta_test_config['n_optimization_steps']
        for i in range(n_grad_steps):
            eval_models_step()

    def _process_eval_batch(self, x):
        return self.meta_test_sampler.tasks[0].process_eval_batch(x, y=None)[0]

    @staticmethod
    def _initialize_model(model_config, dataset_sampler):
        # setup model
        model = model_config['model_class'](model_config, name='nn_model')

        # init variables
        if hasattr(dataset_sampler, 'get_meta_batch'):
            x, y, _, _, _ = dataset_sampler.get_meta_batch()
            x, y = x[0], y[0]
        elif hasattr(dataset_sampler, 'get_batch'):
            x, y = dataset_sampler.get_batch()
        else:
            raise Exception("Unkown dataset sampler")
        model(x)

        return model

    def _copy_model(self, model, x, name):
        copied_model = self.model_config['model_class'](self.model_config, name=name)
        # initialize model by running one forward pass
        copied_model(x)
        # copy variables
        copied_model.set_variables_vectorized(model.get_variables_vectorized())
        return copied_model

""" helper functions """