import pacoh.modules as modules
import pacoh.algorithms as algorithms
from pacoh import util as util

from .meta_algo import MetaLearner

import tensorflow as tf
import numpy as np

class MetaBayesianNeuralNetwork_MLAP(MetaLearner):
    def __init__(self, meta_train_sampler, meta_test_sampler, model_config, likelihood_config, prior_config, posterior_config,
                 hyper_prior_config, hyper_posterior_config, meta_train_config, meta_test_config, n_batched_models, random_seed=None):

        super().__init__(meta_train_sampler, meta_test_sampler, model_config, random_seed=random_seed)
        self.plot_functions_during_training = meta_train_config.get('plot_functions', False)

        assert meta_train_sampler.meta_batch_size == meta_train_sampler.n_tasks, "MLAP doesn't support meta batching"

        # set variables
        self.likelihood_config = likelihood_config
        self.hyper_prior_config = hyper_prior_config
        self.hyper_posterior_config = hyper_posterior_config

        # Setup the randomised predictor
        self.prior_config = prior_config
        self.posterior_config = posterior_config
        x, y, _, _, task_ids = self.meta_train_sampler.get_meta_batch(shuffle=False)
        x, y = self.postprocess_meta_batch(x, y)

        # Check assumptions
        assert self.prior_config['n_batched_priors'] == 1
        assert posterior_config['n_batched_priors'] == self.meta_train_sampler.n_tasks
        assert n_batched_models == (posterior_config['n_samples'] * self.meta_train_sampler.n_tasks)
        assert np.allclose(tf.range(meta_train_sampler.n_tasks), task_ids)
        
        self.rp = modules.MLAP_RandomisedPredictor(self.model_config, likelihood_config, prior_config, posterior_config,
                                           n_batched_models, hyper_prior_config, hyper_posterior_config,
                                           batch_sample=(x, y))

        # initialize relevant components / modules of the algorithm
        self.affine_transform = modules.AffineTransform(normalization_mean=self.meta_train_sampler.y_mean,
                                                        normalization_std=self.meta_train_sampler.y_std)

        # Training setup
        # Process meta_train configs
        print("\nTraining optimization parameters:")
        print(f"Meta batch size of: {self.meta_train_sampler.meta_batch_size}")
        self.meta_train_config = util.process_training_config(meta_train_config, meta_train_sampler)

        # Setup problem specific training_config
        self.n_tasks = self.meta_train_sampler.n_tasks
        self.bound_type = self.meta_train_config['bound_type']
        self.task_complexity_weight = tf.convert_to_tensor(self.meta_train_config['task_complexity_weight'], tf.float32)
        self.meta_complexity_weight = tf.convert_to_tensor(self.meta_train_config['meta_complexity_weight'], tf.float32)
        self.delta = tf.convert_to_tensor(self.meta_train_config['delta'], tf.float32)
        self.use_mse_as_loss = self.likelihood_config.get('use_mse_as_loss', False)
        if self.use_mse_as_loss:
            print("NOTE: Using MSE during training")

        self.optimizer = self._setup_optimizer(meta_train_config)

        # Setup inference step
        self.step = self._setup_inference_step()

        # Process meta_test configs
        print("\nMeta-test inference optimization parameters:")
        self.meta_test_inference_class = algorithms.BayesianNeuralNetworkMLAP
        self.meta_test_config = util.process_training_config(meta_test_config, meta_test_sampler)
        self.meta_test_n_batched_models = self.posterior_config['n_samples']

        self._setup_eval_batch_processing()

    def postprocess_meta_batch(self, x, y=None):
        new_dims_x = tf.concat([tf.expand_dims(-1, 0), x.shape[2:]], axis=0)
        x = tf.reshape(x, new_dims_x)

        if y is not None:
            new_dims_y = tf.concat([tf.expand_dims(-1, 0), y.shape[2:]], axis=0)
            y = tf.reshape(y, new_dims_y)

        return x, y

    def predict(self, x, sample_functions=False, sample_from_prior=False):
        if x.ndim == 2:
            x, _ = self.process_eval_batch(x, None)

        assert x.shape[0] == self.rp.n_batched_models
        y_pred, pred_dist = self.rp.predict(x, sample_functions, sample_from_prior)

        y_pred = tf.multiply(y_pred, self.meta_train_sampler.y_std) + self.meta_train_sampler.y_mean
        pred_dist = self.affine_transform.apply(pred_dist)

        return y_pred, pred_dist

    def _setup_inference_step(self):
        assert self.rp.model_variables_only
        assert self.bound_type in {'seeger', 'mcallester', 'variational'}

        """ A) Logprob and score function """

        @tf.function
        def mcallester_meta_objective(empirical_risk_per_task, n_samples_per_task):
            """

            Args:
                empirical_risk_per_task (): (n_tasks, )
                n_samples_per_task (): (n_tasks, )
            Returns:

            """
            ##### ===== #####
            # Required artifacts
            ##### ===== #####

            # Fixed
            task_complexity_weight = self.task_complexity_weight
            meta_complexity_weight = self.meta_complexity_weight
            n_tasks = self.n_tasks # ()
            delta = self.delta #  ()

            # Requiring computation
            hyper_kl = self.rp.hyper_kl() # ()
            expected_kl_per_task = self.rp.expected_KLs() # (n_tasks, )
            ##### ===== #####

            task_complexity_numerator = (hyper_kl + expected_kl_per_task) + tf.math.log(2 * n_samples_per_task * n_tasks / delta)
            task_complexity_denominator = 2 * (n_samples_per_task - 1)
            task_complexity_per_task = tf.math.sqrt(task_complexity_numerator / task_complexity_denominator) # (n_tasks, )

            meta_complexity_numerator = hyper_kl + tf.math.log(2 * n_tasks / delta)
            meta_complexity_denominator = 2 * (n_tasks - 1)
            meta_complexity = tf.math.sqrt(meta_complexity_numerator / meta_complexity_denominator)

            empirical_risk_averaged_over_tasks = tf.reduce_mean(empirical_risk_per_task)
            task_complexity_averaged_over_tasks = tf.reduce_mean(task_complexity_per_task)

            loss = tf.reduce_sum([empirical_risk_averaged_over_tasks,
                   task_complexity_weight * task_complexity_averaged_over_tasks,
                   meta_complexity_weight * meta_complexity])

            return loss

        @tf.function
        def variational_meta_objective(empirical_risk_per_task, n_samples_per_task):
            """

            Args:
                empirical_risk_per_task (): (n_tasks, )
                n_samples_per_task (): (n_tasks, )
            Returns:

            """
            ##### ===== #####
            # Required artifacts
            ##### ===== #####

            # Fixed
            task_complexity_weight = self.task_complexity_weight
            meta_complexity_weight = self.meta_complexity_weight
            n_tasks = self.n_tasks  # ()

            # Requiring computation
            hyper_kl = self.rp.hyper_kl()  # ()
            expected_kl_per_task = self.rp.expected_KLs()  # (n_tasks, )
            ##### ===== #####

            unnormalized_empirical_risk_per_task = empirical_risk_per_task * n_samples_per_task
            task_complexity_per_task = expected_kl_per_task # (n_tasks, )
            meta_complexity = hyper_kl

            total_empirical_risk = tf.reduce_sum(unnormalized_empirical_risk_per_task)
            total_task_complexity = tf.reduce_sum(task_complexity_per_task)

            # Normalize to get on the same scale as with the other bounds, to facilitate hyper-param search
            const = tf.reduce_mean(n_samples_per_task) * n_tasks

            empirical_risk_averaged_over_tasks = total_empirical_risk / const
            task_complexity_averaged_over_tasks = total_task_complexity / const
            meta_complexity = meta_complexity / const

            loss = tf.reduce_sum([empirical_risk_averaged_over_tasks,
                   task_complexity_weight * task_complexity_averaged_over_tasks,
                   meta_complexity_weight * meta_complexity])

            return loss

        @tf.function
        def seeger_meta_objective(empirical_risk_per_task, n_samples_per_task):
            """

            Args:
                empirical_risk_per_task (): (n_tasks, )
                n_samples_per_task (): (n_tasks, )
            Returns:

            """
            ##### ===== #####
            # Required artifacts
            ##### ===== #####

            # Fixed
            task_complexity_weight = self.task_complexity_weight
            meta_complexity_weight = self.meta_complexity_weight
            n_tasks = self.n_tasks  # ()
            delta = self.delta  # ()

            # Requiring computation
            hyper_kl = self.rp.hyper_kl()  # ()
            expected_kl_per_task = self.rp.expected_KLs()  # (n_tasks, )
            ##### ===== #####

            epsilon_numerator = hyper_kl + expected_kl_per_task + tf.math.log(
                4 * tf.math.sqrt(n_samples_per_task) * n_tasks / delta)
            epsilon = epsilon_numerator / n_samples_per_task
            task_complexity_per_task = 2. * epsilon + tf.sqrt(2. * epsilon * empirical_risk_per_task)

            meta_complexity_numerator = hyper_kl + tf.math.log(2 * n_tasks / delta)
            meta_complexity_denominator = 2 * (n_tasks - 1)
            meta_complexity = tf.math.sqrt(meta_complexity_numerator / meta_complexity_denominator)

            empirical_risk_averaged_over_tasks = tf.reduce_mean(empirical_risk_per_task)
            task_complexity_averaged_over_tasks = tf.reduce_mean(task_complexity_per_task)

            loss = tf.reduce_sum([empirical_risk_averaged_over_tasks, task_complexity_weight * task_complexity_averaged_over_tasks, meta_complexity_weight * meta_complexity])

            return loss

        def get_objective_function():
            if self.bound_type == 'mcallester':
                return mcallester_meta_objective
            elif self.bound_type == 'seeger':
                return seeger_meta_objective
            elif self.bound_type == 'variational':
                return variational_meta_objective
            else:
                raise ValueError('Unknown bound type: {}.'.format(self.bound_type))

        objective_function = get_objective_function()

        @tf.function
        def get_log_prob_and_grad(x, y, n_samples_per_task):
            with tf.GradientTape() as tape:
                tape.watch(self.trainable_parameters)
                mp = self.rp.sample_functions(sample_from_prior=False) # [meta_batch_size * n_samples_used_for_gradient_estimate, n_parameters]

                if self.use_mse_as_loss:
                    empirical_risk_per_task = self.rp.custom_mse_model(mp, x, y)  # [n_tasks, 1]
                else:
                    log_likelihood = self.rp.custom_log_likelihood_model(mp, x, y)  # [meta_batch_size, 1]
                    empirical_risk_per_task = -log_likelihood

                loss = objective_function(empirical_risk_per_task, n_samples_per_task)

            grads = tape.gradient(loss, self.trainable_parameters)
            return empirical_risk_per_task, grads

        self.get_log_prob_and_grad = get_log_prob_and_grad

        def _perform_update(grads):
            for opt, g, tp in zip(self.optimizer, grads, self.trainable_parameters):
                opt.apply_gradients([(g, tp)])

        def step():
            meta_batch_x, meta_batch_y, n_samples_per_task, batch_size, task_ids = self.meta_train_sampler.get_meta_batch(shuffle=False)
            meta_batch_x, meta_batch_y = self.postprocess_meta_batch(meta_batch_x, meta_batch_y)

            emprical_risk, grads = self.get_log_prob_and_grad(meta_batch_x, meta_batch_y, n_samples_per_task)

            _perform_update(grads)

            return -emprical_risk

        return step

    def _setup_optimizer(self, training_config):
        self.trainable_parameters = list(self.rp.prior_module.variables) + list(self.rp.posterior_module.variables)
        config = training_config
        optimizers = []

        for n_p in self.trainable_parameters:
            lr = config['learning_rate']

            if 'prior' in n_p.name:
                lr *= config['prior_lr_scale']

            if 'log_var' in n_p.name:
                lr *= config['variance_lr_scale']

            if config['optimizer'] == 'adam':
                optimizer = tf.optimizers.Adam(learning_rate=lr)
            elif config['optimizer'] == 'adagrad':
                optimizer = tf.optimizers.Adagrad(learning_rate=lr)

            optimizers.append(optimizer)

        return optimizers

    def _setup_meta_test_models_and_step(self, meta_test_tasks):
        # initialize BNN models with the meta-learned prior for each meta-test task
        meta_learned_prior = self.rp.prior_module
        eval_models = []

        for meta_test_task in meta_test_tasks:
            # each meta_test_task corresponds to a dataset sampler
            eval_model = self.meta_test_inference_class(meta_test_task, self.model_config, self.likelihood_config,
                                                        self.prior_config, self.posterior_config, self.meta_test_config,
                                                        meta_learned_prior_module=meta_learned_prior,
                                                        hyper_posterior_config=self.hyper_posterior_config,
                                                        n_batched_models=self.meta_test_n_batched_models)
            eval_models.append(eval_model)

        # define parallel training step on all the models
        @tf.function
        def eval_models_step_tf(batch_per_model):
            log_likelihood_per_eval_model = tf.stack([tf.reduce_mean(eval_model.step(batch[0], batch[1]))
                                                      for eval_model, batch in zip(eval_models, batch_per_model)])

            return log_likelihood_per_eval_model

        def eval_models_step():
            # returns a tensor with dimensions ( #num_eval_models, )
            # prepare batches
            batch_per_eval_model = [model.dataset_sampler.get_batch() for model in eval_models]

            # run step in parallel
            log_likelihood_per_eval_model = eval_models_step_tf(batch_per_eval_model)

            return log_likelihood_per_eval_model

        assert len(eval_models) == len(meta_test_tasks)
        return eval_models, eval_models_step

    def _setup_eval_batch_processing(self):
        self._process_eval_batch = self.process_eval_batch

        def process_eval_batch(x, y=None):
            x, y = self._process_eval_batch(x, y)

            tile_multiplies_x = tf.concat(
                [tf.expand_dims(self.rp.posterior_module._n_batched_models, 0), tf.ones_like(x.shape)], axis=0)
            x = tf.expand_dims(x, 0)
            x = tf.tile(x, tile_multiplies_x)

            if y is not None:
                tile_multiplies_y = tf.concat(
                    [tf.expand_dims(self.rp.posterior_module._n_batched_models, 0), tf.ones_like(y.shape)], axis=0)
                y = tf.expand_dims(y, 0)
                y = tf.tile(y, tile_multiplies_y)

            x, y = self.postprocess_meta_batch(x, y)

            return x, y

        self.process_eval_batch = process_eval_batch