import time
import numpy as np
import tensorflow as tf

import pacoh.modules as modules
import pacoh.algorithms as algorithms
from pacoh import util as util

from .meta_algo import MetaLearner


class MetaBayesianNeuralNetwork_SVGD(MetaLearner):

    def __init__(self, meta_train_sampler, meta_test_sampler, model_config, likelihood_config, prior_config, posterior_config,
                 hyper_prior_config, hyper_posterior_config, meta_train_config, meta_test_config,
                 plot_prior_during_training=False, random_seed=None, catoni_constant = 10):

        super().__init__(meta_train_sampler, meta_test_sampler, model_config, random_seed=random_seed)

        self.catoni_constant = catoni_constant
        # set variables
        self.likelihood_config = likelihood_config
        self.prior_config = prior_config
        self.posterior_config = posterior_config
        self.meta_train_config = meta_train_config
        self.meta_test_config = meta_test_config
        self.hyper_prior_config = hyper_prior_config

        self.meta_test_inference_class = algorithms.BayesianNeuralNetworkSVGD
        self.prior_prob_weight = tf.convert_to_tensor(self.meta_train_config['prior_prob_weight'], tf.float32)
        self.plot_functions_during_training = plot_prior_during_training

        # process training configs
        print("\nTraining optimization parameters:")
        print(f"Meta batch size of: {self.meta_train_sampler.meta_batch_size}")
        self.logsumexp_estimate = self.meta_train_config.get('logsumexp_estimate', False)
        self.meta_train_config = util.process_training_config(meta_train_config, meta_train_sampler)

        # process testing configs
        print("\nMeta-test inference optimization parameters:")
        self.meta_test_config = util.process_training_config(meta_test_config, meta_test_sampler)


        # initialize relevant components / modules of the algorithm
        x, y, _, _, _ = self.meta_train_sampler.get_meta_batch()
        self.rp = modules.MetaRandomisedPredictor(self.model_config, likelihood_config, prior_config,
                                                  hyper_prior_config, hyper_posterior_config, self.meta_train_sampler,
                                                  batch=(x[0],y[0]), logsumexp_estimate=self.logsumexp_estimate, catoni_constant = self.catoni_constant)
        self.reshape_input = getattr(self.rp.model, 'reshape_input', False)

        self.affine_transform = modules.AffineTransform(normalization_mean=self.meta_train_sampler.y_mean,
                                                        normalization_std=self.meta_train_sampler.y_std)
        self.kernel = modules.RBF_Kernel_TF(self.meta_train_config['bandwidth'])

        self._setup_optimizer(meta_train_config)

    def predict(self, x, sample_functions=False, sample_from_prior=False):
        if x.ndim == 2:
            x, _ = self.process_eval_batch(x, None)

        # standardize
        # x = tf.divide((x - self.meta_train_sampler.x_mean), self.meta_train_sampler.x_std)

        assert x.shape[0] == self.rp.n_batched_models
        y_pred, pred_dist = self.rp.predict(x, sample_functions, sample_from_prior)
        y_pred = tf.multiply(y_pred, self.meta_train_sampler.y_std) + self.meta_train_sampler.y_mean
        pred_dist = self.affine_transform.apply(pred_dist)

        return y_pred, pred_dist

    def plot_meta_data(self, train_tasks_to_include, val_tasks_to_include, ax=None):
        import matplotlib.pyplot as plt

        new_axis = False
        if ax is None:
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 4), sharex=True, sharey=True)
            new_axis = True

        self.meta_train_sampler.plot_data(train_tasks_to_include, True, ax=ax1)
        self.meta_test_sampler.plot_data(val_tasks_to_include, False, ax=ax2)

        if new_axis:
            ax1.legend()
            ax2.legend()
            plt.show()

    def step(self):
        # compute score (grads of log_prob)
        meta_batch_x, meta_batch_y, n_train_samples, _, _ = self.meta_train_sampler.get_meta_batch()
        log_prob, grads_log_prob = self._get_log_prob_and_grad(meta_batch_x, meta_batch_y, n_train_samples)
        score = grads_log_prob[0]

        # SVGD
        particles = self.rp.hyper_posterior_module.particles
        K_XX, grad_K_XX = self._get_kernel_matrix_and_grad(particles)
        inter = tf.matmul(K_XX, score)
        phi = (inter - grad_K_XX) / particles.shape[1]
        grads = -phi

        self.optimizer.apply_gradients([(grads, self.rp.hyper_posterior_module.variables[0])])
        return log_prob

    @tf.function
    def _get_log_prob_and_grad(self, meta_batch_x, meta_batch_y, n_train_samples):
        hyper_posterior_module = self.rp.hyper_posterior_module
        with tf.GradientTape() as tape:
            tape.watch(hyper_posterior_module.variables)
            sample = hyper_posterior_module.sample()
            log_likelihood = self.rp.compute_mll_estimate(sample, meta_batch_x, meta_batch_y, n_train_samples)
            log_prior_prob = self.rp.log_prob_hyper_prior(sample, self.prior_prob_weight)
            log_prob = log_likelihood + log_prior_prob

        grads = tape.gradient(log_prob, hyper_posterior_module.variables)
        return log_prob, grads

    @tf.function
    def _get_kernel_matrix_and_grad(self, X):
        X2 = tf.identity(X)
        with tf.GradientTape() as tape:
            tape.watch(X)
            K_XX = self.kernel(X, X2)
        K_grad = tape.gradient(K_XX, X)
        return K_XX, K_grad

    def _setup_meta_test_models_and_step(self, meta_test_tasks):
        # initialize BNN models with the meta-learned prior for each meta-test task
        meta_learned_prior = self.rp.prior_module
        eval_models = []
        for meta_test_task in meta_test_tasks:
            # each meta_test_task corresponds to a dataset sampler
            eval_model = self.meta_test_inference_class(meta_test_task, self.model_config, self.likelihood_config,
                                                        self.prior_config, self.posterior_config,
                                                        self.meta_test_config,
                                                        meta_learned_prior_module=meta_learned_prior)
            eval_models.append(eval_model)

        # define parallel training step on all the models
        @tf.function
        def eval_models_step_tf(batch_per_model):
            log_likelihood_per_eval_model = tf.stack([tf.reduce_mean(eval_model.step(batch[0], batch[1]))
                                                      for eval_model, batch in zip(eval_models, batch_per_model)])

            return log_likelihood_per_eval_model

        def eval_models_step():
            # returns a tensor with dimensions ( #num_eval_models, )

            # prepare data
            batch_per_eval_model = [model.dataset_sampler.get_batch() for model in eval_models]

            # run step in parallel
            log_likelihood_per_eval_model = eval_models_step_tf(batch_per_eval_model)

            return log_likelihood_per_eval_model

        assert len(eval_models) == len(meta_test_tasks)
        return eval_models, eval_models_step

