import time

import tensorflow as tf
import tensorflow_addons as tfa
import numpy as np

import pacoh.modules as modules
from pacoh import util

# TODO: Finish
class Learner:
    plot_functions_during_eval = False
    train_config = None
    data_sampler = None

    def __init__(self, dataset_sampler, model_config, random_seed=None):

        self._set_seed(random_seed)
        self.model_config = self._setup_model_config(model_config, dataset_sampler)

        self.dataset_sampler = dataset_sampler
        self.affine_transform = modules.AffineTransform(normalization_mean=dataset_sampler.y_mean,
                                                        normalization_std=dataset_sampler.y_std)

        self.process_eval_batch = self.dataset_sampler.process_eval_batch

    def fit(self):
        n_steps = self.train_config['n_optimization_steps']
        n_iter_per_step = self.train_config['n_iterations_per_step']
        n_total_iterations = self.train_config['n_total_iterations']
        log_period = self.train_config['log_period']

        n_iter = 0

        tf.print("Start training")

        for s in range(n_steps):
            t = time.time()
            for _ in range(n_iter_per_step):
                n_iter += 1
                x, y = self.dataset_sampler.get_batch()
                log_prob = self.step(x, y)
                avg_log_prob = tf.reduce_mean(log_prob).numpy()
                if n_iter % log_period == 0:
                    message = '\n\tIter %d/%d - Train-Time %.2f sec' % (n_iter, n_total_iterations, time.time() - t)
                    message += ' - Train-Log-Prob: %.3f' % avg_log_prob
                    print(message)
                    t = time.time()

            if s + 1 < n_steps:
                # If it is not the last step
                t = time.time()
                message = ""
                eval_metrics_mean, eval_metrics_std = self.eval()
                for key in eval_metrics_mean:
                    if eval_metrics_std is not None:
                        message += ' - Val-%s: %.3f +- %.3f' % (key, eval_metrics_mean[key], eval_metrics_std[key])
                    else:
                        message += ' - Val-%s: %.3f' % (key, eval_metrics_mean[key])

                print('\n\tIter %d/%d - Eval-Time %.2f sec %s' % (n_iter, n_total_iterations, time.time() - t, message))

    def predict(self, x, sample_functions=True, sample_from_prior=False):
        if x.ndim == 2:
            x, _ = self.process_eval_batch(x, None)

        assert x.shape[0] == self.rp.n_batched_models
        y_pred, pred_dist = self.rp.predict(x, sample_functions, sample_from_prior)

        # unnormalize predictions
        y_pred = tf.multiply(y_pred, self.dataset_sampler.y_std) + self.dataset_sampler.y_mean
        pred_dist = self.affine_transform.apply(pred_dist)

        return y_pred, pred_dist

    def eval(self):
        x, y_true = self.dataset_sampler.val_data
        y_pred, pred_dist = self.predict(x)

        o = self.rp.likelihood.calculate_eval_metrics(y_pred, pred_dist, y_true)
        if self.train_config['plot_functions']:
            import matplotlib.pyplot as plt
            fig, (ax) = plt.subplots(1, 1, figsize=(4, 4))
            # self.dataset.plot_data(train=True, val=False, ax=ax)
            self.dataset_sampler.plot_prediction_functions(model=self, ax=ax,
                                                           plot_pred_std=self.train_config['plot_pred_std'],
                                                           plot_pred_lines=self.train_config['plot_pred_lines'],
                                                           plot_data=True)
            plt.legend()
            plt.show()

        return o, None

    @staticmethod
    def _setup_model_config(model_config, dataset_sampler):
        if 'type' in model_config:
            # determines the NN class and, if necessary, adds image parameters (height, width, depth) to the config
            if model_config['type'] == 'cnn':
                model_config = model_config[model_config['type']]
                model_config['model_class'] = modules.CNN
            elif model_config['fc']:
                model_config = model_config[model_config['type']]
                model_config['model_class'] = modules.FullyConnectedNN
            else:
                raise AssertionError('model type must be either \'cnn\' or \'fc\'')
        else:
            # make sure that the NN class is already specified in the config
            assert 'model_class' in model_config

        return model_config

    def _calib_error(self, pred_dist, y_true):
        # cdf returns a tensor of dimensions (n_batched_models, 1)
        cdf_vals = pred_dist.cdf(y_true)

        if y_true.ndim == 3:
            y_true = y_true[0]

        num_points = y_true.shape[1]
        conf_levels = tf.linspace(0.05, 0.95, 20)
        emp_freq_per_conf_level = tf.reduce_sum(tf.cast(cdf_vals[:, None] <= conf_levels, tf.float32),
                                                axis=0) / num_points

        calib_rmse = tf.sqrt(tf.reduce_mean((emp_freq_per_conf_level - conf_levels) ** 2))
        return calib_rmse

    # Data related methods
    def _prepare_batch(self, task):
        sampler = task.train_batch_sampler
        batch = tf.data.experimental.get_next_as_optional(sampler)
        x, y = batch.get_value()

        x.set_shape(x.shape)
        y.set_shape(y.shape)
        x, y = self._standardize(x, y)

        return x, y, tf.cast(task.n_train_samples, tf.float32), tf.cast(task.batch_size, tf.float32)

    # Randomness
    def _set_seed(self, random_seed):
        if random_seed is not None:
            np.random.seed(random_seed + 1)
            tf.random.set_seed(random_seed + 2)