from abc import ABC, abstractmethod
import numpy as np
import tensorflow as tf
import time

import pacoh.modules as modules

class MetaLearner:
    plot_functions_during_training = False
    meta_train_config = None
    meta_test_config = None

    def __init__(self, meta_train_sampler, meta_test_sampler, model_config, random_seed=None):
        self._set_seed(random_seed)
        self.model_config = self._setup_model_config(model_config, meta_train_sampler)

        self.meta_train_sampler = meta_train_sampler
        self.meta_test_sampler = meta_test_sampler
        self.process_eval_batch = meta_train_sampler.tasks[0].process_eval_batch

    def meta_fit(self):
        n_steps = self.meta_train_config['n_optimization_steps']
        n_iter_per_step = self.meta_train_config['n_iterations_per_step']
        n_total_iterations = self.meta_train_config['n_total_iterations']
        log_period = self.meta_train_config['log_period']

        n_iter = 0

        tf.print("Start meta-training -------------------- ")

        for s in range(n_steps):
            t = time.time()
            for _ in range(n_iter_per_step):
                n_iter += 1
                log_prob = self.step()

                # print training loss
                if n_iter % log_period == 0:
                    avg_log_prob = tf.reduce_mean(log_prob).numpy()
                    message = '\nIter %d/%d - Time %.2f sec' % (n_iter, n_total_iterations, time.time() - t)
                    message += ' - Train-Loss: %.5f' % avg_log_prob
                    print(message)
                    t = time.time()

                    if self.plot_functions_during_training:
                       self.meta_train_sampler.plot_prediction_functions(self, plot_pred_lines=True,
                                                                        sample_functions=True,
                                                                         plot_pred_std=False, sample_from_prior=True,
                                                                         plot_data=True)
            # run validation and print results
            if s + 1 < n_steps:
                # If it is not the last step
                t = time.time()
                message = ""
                eval_metrics_mean, eval_metrics_std = self.meta_eval()
                for key in eval_metrics_mean:
                    message += '- Val-%s: %.3f +- %.3f' % (key, eval_metrics_mean[key], eval_metrics_std[key])

                print('\nStep %d/%d - Time %.2f sec %s' % (s + 1, n_steps, time.time() - t, message))

    def meta_eval(self, print_results=False, max_tasks_parallel=5):
        """
        meta-testing functionality - Runs posterior inference for the tasks of the meta_test_sampler
        and reports evaluation metrics. For the posterior inference the context data is used while the evaluation
        metrics are computed on the left-out test sets

        Args:
            print_results (bool): whether to print the final evaluation results
            max_tasks_parallel (int): maximum number of tasks to evaluate on in parallel. If
                                      max_tasks_parallel < n_test_tasks

        Returns: (eval_metrics_mean, eval_metrics_std) each a dict of evaluation results

        """
        eval_tasks = self.meta_test_sampler.tasks
        eval_tasks_batches = _split_into_batches(eval_tasks, max_batch_size=max_tasks_parallel)

        # Start training the BNN models
        print("\tStart meta-test posterior inference ------------------")
        if len(eval_tasks_batches) > 1:
            print("\tPerforming meta-testing in %i batches"%len(eval_tasks_batches))

        eval_metrics_dict_per_task = []
        for eval_batch_id, task_batch in enumerate(eval_tasks_batches):
            print("\tMeta-Test batch #%i consisting of %i tasks----"%(eval_batch_id+1, len(task_batch)))
            # initialize evaluation models and define their training step
            eval_models, eval_models_step = self._setup_meta_test_models_and_step(meta_test_tasks=task_batch)
            self._meta_test_training_loop(task_batch, eval_models, eval_models_step)

            t = time.time()
            _, _, eval_metrics_grouped = self._meta_test_models_eval(task_batch, eval_models)
            eval_metrics_dict_per_task.append(eval_metrics_grouped)

        eval_metrics_mean, eval_metrics_std = self._aggregate_eval_metrics_across_tasks(eval_metrics_dict_per_task)
        if print_results:
            message = ""
            for key in eval_metrics_mean:
                message += '- Val-%s: %.3f +- %.3f' % (key, eval_metrics_mean[key], eval_metrics_std[key])
            print('\n\tMeta-Test - Time %.2f sec %s' % (time.time() - t, message))

        return eval_metrics_mean, eval_metrics_std

    def _setup_optimizer(self, training_config):
        if training_config['optimizer'] == 'adam':
            self.optimizer = tf.keras.optimizers.Adam(learning_rate=training_config['learning_rate'])
        else:
            raise Exception("Unexpected optimizer choice")

        return self.optimizer

    def _standardize(self, x, y=None):
        x = tf.divide((x - self.meta_train_sampler.x_mean), self.meta_train_sampler.x_std)
        if y is not None:
            y = tf.divide((y - self.meta_train_sampler.y_mean), self.meta_train_sampler.y_std)
            return x, y
        else:
            return x

    def _unstandardize(self, y):
        y = tf.multiply(y, self.meta_train_sampler.y_std) + self.meta_train_sampler.y_mean
        return y

    @abstractmethod
    def step(self):
        raise NotImplementedError

    @abstractmethod
    def predict(self, x, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def _setup_meta_test_models_and_step(self, meta_test_tasks):
        raise NotImplementedError

    def _meta_test_training_loop(self, test_tasks, eval_models, eval_models_step):
        n_steps = self.meta_test_config['n_optimization_steps']
        n_iter_per_step = self.meta_test_config['n_iterations_per_step']
        n_total_iterations = self.meta_test_config['n_total_iterations']
        log_period = self.meta_test_config['log_period']

        n_iter = 0
        for s in range(n_steps):
            t = time.time()
            for _ in range(n_iter_per_step):
                n_iter += 1
                log_prob = eval_models_step()

                # print meta-test training loss
                if n_iter % log_period == 0:
                    avg_log_prob = tf.reduce_mean(log_prob).numpy()
                    message = '\tMeta-Test - Iter %d/%d - Time %.2f sec' % (
                        n_iter, n_total_iterations, time.time() - t)
                    message += ' - Train-Log-Prob: %.3f' % avg_log_prob
                    print(message)
                    t = time.time()

            # meta-test evaluation
            if s + 1 < n_steps:
                t = time.time()
                eval_metrics_mean, eval_metrics_std, _ = self._meta_test_models_eval(test_tasks, eval_models)
                message = ""
                for key in eval_metrics_mean:
                    message += '- Val-%s: %.3f +- %.3f' % (key, eval_metrics_mean[key], eval_metrics_std[key])

                print('\n\tStep %d/%d - Time %.2f sec %s' % (s + 1, n_steps, time.time() - t, message))

    def _meta_test_models_eval(self, test_tasks, eval_models):
        eval_metrics_grouped = {}
        for m in eval_models:
            res, _ = m.eval()
            for key in res:
                l = eval_metrics_grouped.get(key, [])
                l.append(res[key].numpy())
                eval_metrics_grouped[key] = l

        eval_metrics_mean = {}
        eval_metrics_std = {}
        for key in eval_metrics_grouped:
            eval_metrics_mean[key] = np.mean(eval_metrics_grouped[key])
            eval_metrics_std[key] = np.std(eval_metrics_grouped[key])

        return eval_metrics_mean, eval_metrics_std, eval_metrics_grouped

    @staticmethod
    def _setup_model_config(model_config, dataset_sampler):
        # determines the NN class and, if necessary, adds image parameters (height, width, depth) to the config
        if model_config['type'] == 'cnn':
            model_config = model_config[model_config['type']]
            # for key in ['img_height', 'img_width', 'img_depth']:
            #     model_config[key] = getattr(dataset_sampler, key)
            model_config['model_class'] = modules.CNN
        elif model_config['type'] == 'fc':
            model_config = model_config[model_config['type']]
            model_config['model_class'] = modules.FullyConnectedNN
        else:
            raise AssertionError('model type must be either \'cnn\' or \'fc\'')
        return model_config

    @staticmethod
    def _aggregate_eval_metrics_across_tasks(eval_metrics_dict_per_task):
        eval_metrics_mean, eval_metrics_std = {}, {}
        metric_keys = set.union(*[set(metrics_dict.keys()) for metrics_dict in eval_metrics_dict_per_task])
        for key in metric_keys:
            metric_list = []
            for eval_metrics_dict in eval_metrics_dict_per_task:
                if type(eval_metrics_dict[key]) == list:
                    metric_list.extend(eval_metrics_dict[key])
                else:
                    metric_list.append(eval_metrics_dict[key])
            eval_metrics_mean[key] = np.mean(metric_list)
            eval_metrics_std[key] = np.std(metric_list)
        return eval_metrics_mean, eval_metrics_std

    def _set_seed(self, random_seed):
        if random_seed is not None:
            np.random.seed(random_seed + 1)
            tf.random.set_seed(random_seed + 2)

""" --- helper functions --- """



def _split_into_batches(list, max_batch_size):
    import math
    n_elements = len(list)
    if max_batch_size == -1:
        max_batch_size = n_elements
    n_batches = math.ceil(n_elements / float(max_batch_size))
    remainder = n_elements % n_batches
    batches = []
    idx = 0
    for i in range(n_batches):
        if i < remainder:
            batch_size = n_elements // n_batches + 1
        else:
            batch_size = n_elements // n_batches
        batches.append(list[idx:idx + batch_size])
        idx += batch_size

    assert all([len(batch) <= max_batch_size for batch in batches])
    assert np.sum([len(batch) for batch in batches]) == n_elements
    return batches
