import functools
import numpy as np

class MetaBanditTask:
    x_domain = None

    def mean_reward(self, action):
        raise NotImplementedError

    def noisy_reward(self, action):
        raise NotImplementedError

    def _action_to_x(self, action):
        return self.x_domain[action]

    @property
    @functools.lru_cache()
    def n_arms(self):
        return self.x_domain.shape[0]

    @property
    @functools.lru_cache()
    def best_action(self):
        return np.argmax(self.mean_reward(np.arange(self.n_arms).flatten()))

    @property
    @functools.lru_cache()
    def best_reward(self):
        return np.max(self.mean_reward(np.arange(self.n_arms)).flatten())

    def average_regret(self, action_list):
        reward_mean = np.mean(self.mean_reward(action_list))
        return self.best_reward - reward_mean

    def maximum_regret(self, rewards_list, action_list):
        current_best_action = action_list[np.argmax(rewards_list)]
        current_max_reward = np.max(self.mean_reward(current_best_action))
        return self.best_reward - current_max_reward

class MetaCauchyBandit(MetaBanditTask):
    x_domain = np.expand_dims(np.arange(-10, 10, step=0.01), axis=-1)

    def __init__(self, noise_std=0.05, random_seed=534):
        self.random_state = np.random.RandomState(random_seed)
        self.noise_std = noise_std

    def mean_reward(self, action):
        x = self._action_to_x(action)
        return self._mean(x)

    def noisy_reward(self, action):
        r = self.mean_reward(action)
        return r + np.random.normal(0, scale=self.noise_std, size=r.shape)

    def generate_meta_train_data(self, n_tasks=20, n_samples=20):
        meta_train_tuples = []
        for i in range(n_tasks):
            x_idx = self.random_state.choice(self.x_domain.shape[0], n_samples)
            X = self.x_domain[x_idx]
            Y = np.expand_dims(self._mean(X) + self._gp_fun_from_prior(X), axis=-1)
            Y = Y + self.random_state.normal(scale=self.noise_std, size=Y.shape)
            meta_train_tuples.append((X, Y))
        return meta_train_tuples

    def generate_meta_test_data(self, n_tasks=1, n_samples_context=20, n_samples_test=200):
        assert n_tasks==1
        assert n_samples_test > 0
        meta_test_tuples = []
        for i in range(n_tasks):
            x_idx = self.random_state.choice(self.x_domain.shape[0], n_samples_context + n_samples_test)
            X = self.x_domain[x_idx]
            Y = np.expand_dims(self._mean(X), axis=-1)
            Y = Y + self.random_state.normal(scale=self.noise_std, size=Y.shape)
            meta_test_tuples.append(
                (X[:n_samples_context], Y[:n_samples_context], X[n_samples_context:], Y[n_samples_context:]))

        return meta_test_tuples

    def _action_to_x(self, action):
        return self.x_domain[action]

    def _mean(self, x):
        loc1 = -1 * np.ones(x.shape[-1])
        loc2 = 2 * np.ones(x.shape[-1])
        cauchy1 = 1 / (np.pi * (1 + (np.linalg.norm(x - loc1, axis=-1)) ** 2))
        cauchy2 = 1 / (np.pi * (1 + (np.linalg.norm(x - loc2, axis=-1)) ** 2))
        return 4 * cauchy1 + 3 * cauchy2 + 1

    def _gp_fun_from_prior(self, X):
        assert X.ndim == 2

        n = X.shape[0]

        def kernel(a, b, lengthscale):
            sqdist = np.sum(a ** 2, 1).reshape(-1, 1) + np.sum(b ** 2, 1) - 2 * np.dot(a, b.T)
            return np.exp(-.5 * (1 / lengthscale) * sqdist)

        K_ss = kernel(X, X, 0.5)
        L = np.linalg.cholesky(K_ss + 1e-8 * np.eye(n))
        f = np.dot(L, self.random_state.normal(scale=0.2, size=(n, 1))).flatten()
        y = f + self.random_state.normal(scale=self.noise_std, size=f.shape)
        return y

class MetaMHCBandit(MetaBanditTask):

    def __init__(self, cv_split_id=0, allele_id=0):
        from pacoh.datasets.regression_datasets import MHCDataset
        self.mhc_ds = MHCDataset(cv_split_id=cv_split_id)
        self.allele_id = allele_id

        self.x_domain, self.rewards = self.mhc_ds._load_data()[self.mhc_ds.test_task_ids[allele_id]]
        self.rewards = self.rewards.flatten()

    def generate_meta_train_data(self, n_tasks=5, n_samples=-1):
        return self.mhc_ds.generate_meta_train_data(n_tasks, n_samples)

    def generate_meta_test_data(self, n_tasks=1, n_samples_context=5, n_samples_test=-1):
        assert n_tasks == 1
        meta_test_tuples = self.mhc_ds.generate_meta_test_data(n_tasks=2, n_samples_context=n_samples_context,
                                                               n_samples_test=n_samples_test)
        return [meta_test_tuples[self.allele_id]]

    def mean_reward(self, action):
        return self.rewards[action]

    def noisy_reward(self, action):
        return self.rewards[action]

    @property
    @functools.lru_cache()
    def n_arms(self):
        return self.x_domain.shape[0]

    @property
    @functools.lru_cache()
    def best_action(self):
        return np.argmax(self.mean_reward(np.arange(self.n_arms).flatten()))

    @property
    @functools.lru_cache()
    def best_reward(self):
        return np.max(self.mean_reward(np.arange(self.n_arms)).flatten())

    def average_regret(self, action_list):
        reward_mean = np.mean(self.mean_reward(action_list))
        return self.best_reward - reward_mean

    def maximum_regret(self, rewards_list, action_list):
        current_best_action = action_list[np.argmax(rewards_list)]
        current_max_reward = np.max(self.mean_reward(current_best_action))
        return self.best_reward - current_max_reward

if __name__ == '__main__':
    mhc_bandit = MetaMHCBandit()