import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import json
from matplotlib import pyplot as plt
import numpy as np

from pacoh.experiments.utils import NumpyEncoder

from config import DATA_DIR

BANDIT_EXP_DIR = os.path.join(DATA_DIR, 'bandit_exps')


META_SEED = 55


LAYER_SIZE = 64
N_LAYERS = 4
ACTIVATION = 'relu'
OUTPUT_DIM = 1



def get_config(learn_likelihood=True, prior_std=0.5, n_particles=20, n_train_iter=4000, bandwidth=100.,
               prior_weight=1.0):
    config = {
        'likelihood_config': {'class': {'learn_likelihood_variables': False},
                              'reg': {'log_var': 2 * np.log(0.01),
                                      'learn_likelihood_variables': learn_likelihood},
                              'type': 'reg'
                              },

        'model_config': {'fc': {'input_dim': None,
                                'output_dim': OUTPUT_DIM,
                                'layer_units': [LAYER_SIZE for _ in range(N_LAYERS)] + [OUTPUT_DIM],
                                'activations': [ACTIVATION for _ in range(N_LAYERS)] + [None]
                                },
                         'type': 'fc'},

        'prior_config': {'gaussian': {'model': {'mean_mu': 0.0,
                                                'bias_log_var_mu': 2 * np.log(2.0),
                                                'weight_log_var_mu': 2 * np.log(prior_std)  # 2 * np.log(0.5)
                                                },
                                      'likelihood': {'log_var_mu': 2 * np.log(0.01),
                                                     'log_var_log_var': 2 * np.log(2.0)
                                                     },
                                      'init_std': 0.1,
                                      'add_noise': True,
                                      'init_from_models': True,
                                      'n_samples': 5,  # number of samples from prior for estimating the mll
                                      },
                         'type': 'gaussian'
                         },

        'posterior_config': {'n_samples': n_particles,
                             'use_init_std': True,
                             'type': 'particle'
                             },

        'meta_test_config': {'n_total_iterations': n_train_iter,
                             'n_train_loss_logs': 5,
                             'n_eval_runs': 0,
                             'optimizer': 'adam',
                             'learning_rate': 0.002,
                             'plot_functions': False,
                             'plot_pred_lines': False,
                             'plot_pred_std': False,
                             'prior_prob_weight': prior_weight,
                             'bandwidth': bandwidth
                             }

    }

    prior_config = config['prior_config']
    config['prior_config'] = prior_config = prior_config[prior_config['type']]

    return config


def bandit_experiment(bandit_task, bnn_config, n_iter=5, mode='thomson_sampling', plot=False, exp_dir='',
                      bandit_seed=90):
    from pacoh.datasets.bandits import MetaCauchyBandit, MetaMHCBandit
    from pacoh.algorithms.BNN_SVGD_Bandit import setup_meta_learner, bandit_loop

    if len(exp_dir) > 0:
        """ Save config """
        if not os.path.isdir(exp_dir):
            os.mkdir(exp_dir)
        assert os.path.isdir(exp_dir)

        config_path = os.path.join(exp_dir, 'model_config.json')
        with open(config_path, 'w') as f:
            json.dump(bnn_config, f)


    """ Setup Bandit """
    if bandit_task == 'cauchy':
        bandit_task = MetaCauchyBandit()
    elif 'mhc' in bandit_task:
        cv_split_id = int(bandit_task[-3])
        allele_id = int(bandit_task[-1])
        bandit_task = MetaMHCBandit(cv_split_id=cv_split_id, allele_id=allele_id)
    else:
        raise NotImplementedError

    """ Bandit loop """

    action_list, reward_list, max_regret_list, avg_regret_list = bandit_loop(bandit_task, config,
                                                                             meta_reg_model=None,
                                                                             mode=mode,
                                                                             plot_pred_dist=plot,
                                                                             n_iter=n_iter, random_seed=bandit_seed)

    result_dict = {'bandit_seed': bandit_seed,
                   'meta_seed': META_SEED,
                   'actions': action_list,
                   'rewards': reward_list,
                   'max_regret': max_regret_list,
                   'avg_regret': avg_regret_list}


    if plot:
        plt.plot(range(len(max_regret_list)), max_regret_list)
        plt.title('max regret')
        plt.show()

        plt.plot(range(len(avg_regret_list)), avg_regret_list)
        plt.title('avg regret')
        plt.show()

    """ Save result """
    if len(exp_dir) > 0:
        config_path = os.path.join(exp_dir, 'result.json')
        with open(config_path, 'w') as f:
            json.dump(result_dict, f, cls=NumpyEncoder)

    return result_dict


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description='bnn bandit exp')
    parser.add_argument('--mode', type=str, default='thomson_sampling', help='what bandit algo to use')
    parser.add_argument('--bandit_seed', type=int, default=90, help='')
    parser.add_argument('--task', type=str, default='mhc_0_0', help='which bandit task to use')
    parser.add_argument('--n_iter', type=int, default=3, help='how many iterations')
    parser.add_argument('--learn_likelihood', type=bool, default=True, help='')
    parser.add_argument('--prior_std', type=float, default=0.5, help='')
    parser.add_argument('--n_particles', type=int, default=20, help='')
    parser.add_argument('--n_train_iter', type=int, default=5000, help='')
    parser.add_argument('--bandwidth', type=float, default=100., help='')
    parser.add_argument('--prior_weight', type=float, default=1., help='')
    args = parser.parse_args()

    if not os.path.isdir(BANDIT_EXP_DIR):
        os.mkdir(BANDIT_EXP_DIR)

    exp_base_dir = os.path.join(BANDIT_EXP_DIR, 'bnn_bandit_%s_%s_%i_iter'%(args.task, args.mode, args.n_iter))

    if not os.path.isdir(exp_base_dir):
        os.mkdir(exp_base_dir)

    config = get_config(learn_likelihood=args.learn_likelihood,
                        prior_std=args.prior_std,
                        n_particles=args.n_particles,
                        n_train_iter=args.n_train_iter,
                        bandwidth=args.bandwidth,
                        prior_weight=args.prior_weight)

    config_hash = str(hash(str(config)))
    exp_dir = os.path.join(exp_base_dir, config_hash)

    if not os.path.isdir(exp_dir):
        os.mkdir(exp_dir)

    # save config
    config_path = os.path.join(exp_dir, 'config.json')
    with open(config_path, 'w') as f:
        json.dump(vars(args), f)

    from pacoh.experiments.utils import Tee
    log_dir = os.path.join(exp_dir, 'log.out')

    with Tee(log_dir):
        bandit_experiment(args.task, config, n_iter=args.n_iter, plot=False, exp_dir=exp_dir, bandit_seed=args.bandit_seed)