import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import json

from matplotlib import pyplot as plt
import numpy as np

from config import DATA_DIR

from pacoh.experiments.utils import NumpyEncoder

BANDIT_EXP_DIR = os.path.join(DATA_DIR, 'bandit_exps')


META_SEED = 55


""" META LEARNING CONFIG """

LAYER_SIZE = 64
N_LAYERS = 4
ACTIVATION = 'relu'
OUTPUT_DIM = 1


def get_config(learning_mode='both', covar_module='SE', num_iter_fit=2000):
    config = {
        'learning_mode': learning_mode,
        'covar_module': covar_module,
        'mean_module': 'constant',
        'num_iter_fit': num_iter_fit,
        'lr_params': 1e-2,

    }
    return config


def bandit_experiment(bandit_task, meta_learn_config, n_iter=5, meta_learn=True, mode='thomson_sampling',
                      plot=False, exp_dir='', bandit_seed=90):
    from pacoh.datasets.bandits import MetaCauchyBandit, MetaMHCBandit
    from pacoh.algorithms.GP_Bandit import setup_meta_learner, bandit_loop

    if len(exp_dir) > 0:
        """ Save config """
        if not os.path.isdir(exp_dir):
            os.mkdir(exp_dir)
        assert os.path.isdir(exp_dir)

        config_path = os.path.join(exp_dir, 'model_config.json')
        with open(config_path, 'w') as f:
            json.dump(meta_learn_config, f)


    """ Setup Bandit """
    if bandit_task == 'cauchy':
        bandit_task = MetaCauchyBandit()
    elif 'mhc' in bandit_task:
        cv_split_id = int(bandit_task[-3])
        allele_id = int(bandit_task[-1])
        bandit_task = MetaMHCBandit(cv_split_id=cv_split_id, allele_id=allele_id)
    else:
        raise NotImplementedError

    meta_train_data = bandit_task.generate_meta_train_data()
    meta_test_data = bandit_task.generate_meta_test_data()

    """ Meta-Learn prior """
    if meta_learn:
        meta_reg_model = setup_meta_learner(meta_train_data, meta_test_data, meta_learn_config, random_seed=META_SEED)
        meta_reg_model.meta_fit(log_period=100)
    else:
        meta_reg_model = None

    """ Bandit loop """

    action_list, reward_list, max_regret_list, avg_regret_list = bandit_loop(bandit_task, meta_reg_model,
                                                                             plot_pred_dist=plot,
                                                                             mode=mode,
                                                                             n_iter=n_iter, random_seed=bandit_seed)

    result_dict = {'bandit_seed': bandit_seed,
                   'meta_seed': META_SEED,
                   'actions': action_list,
                   'rewards': reward_list,
                   'max_regret': max_regret_list,
                   'avg_regret': avg_regret_list}

    if plot:
        plt.plot(range(len(max_regret_list)), max_regret_list)
        plt.title('max regret')
        plt.show()

        plt.plot(range(len(avg_regret_list)), avg_regret_list)
        plt.title('avg regret')
        plt.show()

    """ Save result """
    if len(exp_dir) > 0:
        config_path = os.path.join(exp_dir, 'result.json')
        with open(config_path, 'w') as f:
            json.dump(result_dict, f, cls=NumpyEncoder)

    return result_dict


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description='gp bandit exp')
    parser.add_argument('--mode', type=str, default='ucb', help='what bandit algo to use')
    parser.add_argument('--task', type=str, default='mhc_0_0', help='which bandit task to use')
    parser.add_argument('--n_iter', type=int, default=50, help='how many iterations')
    parser.add_argument('--bandit_seed', type=int, default=90, help='')
    parser.add_argument('--learning_mode', type=str, default='both', help='')
    parser.add_argument('--covar_module', type=str, default='linear', help='')
    parser.add_argument('--n_train_iter', type=int, default=2000, help='')


    args = parser.parse_args()

    if not os.path.isdir(BANDIT_EXP_DIR):
        os.mkdir(BANDIT_EXP_DIR)

    if args.learning_mode == 'vanilla':
        exp_base_dir = os.path.join(BANDIT_EXP_DIR, 'gp_bandit_%s_%s_%s_%i_iter_' % (args.task, args.mode, args.covar_module, args.n_iter))
    elif args.learning_mode == 'both':
        exp_base_dir = os.path.join(BANDIT_EXP_DIR, 'meta_gp_bandit_%s_%s_%s_%i_iter' % (args.task, args.mode, args.covar_module, args.n_iter))
    else:
        raise NotImplementedError


    if not os.path.isdir(exp_base_dir):
        os.mkdir(exp_base_dir)

    config = get_config(learning_mode=args.learning_mode,
                        covar_module=args.covar_module,
                        num_iter_fit=args.n_train_iter)

    config_hash = str(hash(str(config)))
    exp_dir = os.path.join(exp_base_dir, config_hash)

    if not os.path.isdir(exp_dir):
        os.mkdir(exp_dir)

    # save config
    config_path = os.path.join(exp_dir, 'config.json')
    with open(config_path, 'w') as f:
        json.dump(vars(args), f)

    from pacoh.experiments.utils import Tee
    log_dir = os.path.join(exp_dir, 'log.out')

    with Tee(log_dir):
        bandit_experiment(args.task, config, n_iter=args.n_iter, mode=args.mode, plot=False, exp_dir=exp_dir,
                          bandit_seed=args.bandit_seed)