import os
import json
import glob
from config import DATA_DIR
import pandas as pd

BANDIT_EXP_DIR = os.path.join(DATA_DIR, 'bandit_exps')

def collect_exp_results(exp_name, verbose=True):
    exp_dir = os.path.join(BANDIT_EXP_DIR, exp_name)
    no_results_counter = 0

    exp_dicts = []
    for exp_sub_dir in glob.glob(exp_dir + '/*'):
        config_file = os.path.join(exp_sub_dir, 'config.json')
        results_file = os.path.join(exp_sub_dir, 'result.json')

        if os.path.isfile(config_file) and os.path.isfile(results_file):
            with open(config_file, 'r') as f:
                exp_dict = json.load(f)
            with open(results_file, 'r') as f:
                print(results_file)
                exp_dict.update(json.load(f))
            exp_dicts.append(exp_dict)
        else:
            no_results_counter += 1

    if verbose:
       print('Parsed results %s - found %i folders with results and %i folders without results'
                    %(exp_name, len(exp_dicts), no_results_counter))

    print(exp_dicts)
    return exp_dicts

def load_and_prepare_results(exp_name, top_n=1):
    exp_dicts = collect_exp_results(exp_name)
    df = pd.DataFrame(exp_dicts)
    cols_to_group_by = list(set(df.columns) - {'bandit_seed', 'actions', 'rewards', 'max_regret', 'avg_regret'})

    def series_mean(series):
        return [np.mean(np.stack(series, axis=0), axis=0)]
    def series_std(series):
        return [np.std(np.stack(series, axis=0), axis=0)]
    def series_final(series):
        return np.mean(np.stack(series, axis=0), axis=0)[-1]

    df_aggregated = df.groupby(cols_to_group_by).agg({'avg_regret': [series_mean, series_std, series_final],
                                                      'max_regret': [series_mean, series_std, series_final]})
    # sort by avg regret
    df_aggregated = df_aggregated.sort_values(('avg_regret', 'series_final'))

    best_settings = [dict(df_aggregated.iloc[i]) for i in range(top_n)]
    return best_settings


if __name__ == "__main__":
    from matplotlib import pyplot as plt
    import numpy as np

    plt.rc('text', usetex=True)
    plt.rc('font', family='serif')
    fig_scale = 0.9

    """ 1) plot ucb results """

    method = 'ucb'
    bandit_task = 'mhc_0_0'
    exp_name = 'meta_bnn_bandit_logsumexp_%s_%s_50_iter'%(bandit_task, method)
    best_exp_dicts_meta = load_and_prepare_results(exp_name, top_n=2)

    exp_name = 'bnn_bandit_%s_%s_50_iter'%(bandit_task, method)
    best_exp_dicts = load_and_prepare_results(exp_name, top_n=1)

    kernel = 'linear'
    exp_name = 'meta_gp_bandit_%s_%s_%s_50_iter'%(bandit_task, method, kernel)
    best_exp_dicts_gp = load_and_prepare_results(exp_name, top_n=1)

    for exp_dict_meta, exp_dict, exp_dict_gp in zip(best_exp_dicts_meta, best_exp_dicts, best_exp_dicts_gp):
        fig, axs = plt.subplots(1, 2, figsize=(fig_scale*5.8, fig_scale*2.5))
        for j, metric in enumerate(['avg_regret', 'max_regret']):

            # gp
            avg_regret_mean = exp_dict_gp[(metric, 'series_mean')][0]
            avg_regret_std = exp_dict_gp[(metric, 'series_std')][0]
            axs[j].plot(range(1, len(avg_regret_mean) + 1), avg_regret_mean, "--",  label='GP-UCB')
            axs[j].fill_between(range(1, len(avg_regret_mean) + 1), avg_regret_mean - avg_regret_std,
                                avg_regret_mean + avg_regret_std, alpha=0.2)

            # bnn
            avg_regret_mean = exp_dict[(metric, 'series_mean')][0]
            avg_regret_std = exp_dict[(metric, 'series_std')][0]
            axs[j].plot(range(1, len(avg_regret_mean)+1), avg_regret_mean, ":", linewidth=2.0, label='BNN-UCB')
            axs[j].fill_between(range(1, len(avg_regret_mean)+1), avg_regret_mean-avg_regret_std, avg_regret_mean+avg_regret_std, alpha=0.2)

            # meta_bnn
            avg_regret_mean = exp_dict_meta[(metric, 'series_mean')][0]
            avg_regret_std = exp_dict_meta[(metric, 'series_std')][0]
            axs[j].plot(range(1, len(avg_regret_mean)+1), avg_regret_mean, linewidth=2.0, label='PACOH-UCB')
            axs[j].fill_between(range(1, len(avg_regret_mean)+1), avg_regret_mean-avg_regret_std, avg_regret_mean+avg_regret_std, alpha=0.2)

            axs[j].set_xlabel('iteration')
            axs[j].set_ylabel(metric)

        # lgd = axs[1].legend(axs[1].lines[:3], ["GP", "BNN", "PACOH-DL"])
        # fig.suptitle("UCB")
        # axs[0].set_ylabel('average regret')
        # axs[1].set_ylabel('max regret')
        # fig.tight_layout(rect=(0,0.2,0.5,0.6))
        # fig.show()
        # fig.savefig("plot_%s_%s.pdf"%(bandit_task, method))

        """ 2) plot thomson_sampling results """

        method = 'thomson_sampling'
        bandit_task = 'mhc_0_0'
        exp_name = 'meta_bnn_bandit_logsumexp_%s_%s_50_iter' % (bandit_task, method)
        best_exp_dicts_meta = load_and_prepare_results(exp_name, top_n=2)

        exp_name = 'bnn_bandit_%s_%s_50_iter' % (bandit_task, method)
        best_exp_dicts = load_and_prepare_results(exp_name, top_n=1)

        for exp_dict_meta, exp_dict, exp_dict_gp in zip(best_exp_dicts_meta, best_exp_dicts, best_exp_dicts_gp):
            for j, metric in enumerate(['avg_regret', 'max_regret']):

                # bnn
                avg_regret_mean = exp_dict[(metric, 'series_mean')][0]
                avg_regret_std = exp_dict[(metric, 'series_std')][0]
                axs[j].plot(range(1, len(avg_regret_mean) + 1), avg_regret_mean, ":", linewidth=2.0, label='BNN-TS')
                axs[j].fill_between(range(1, len(avg_regret_mean) + 1), avg_regret_mean - avg_regret_std,
                                    avg_regret_mean + avg_regret_std, alpha=0.2)

                # meta_bnn
                avg_regret_mean = exp_dict_meta[(metric, 'series_mean')][0]
                avg_regret_std = exp_dict_meta[(metric, 'series_std')][0]
                axs[j].plot(range(1, len(avg_regret_mean) + 1), avg_regret_mean, linewidth=2.0, label='PACOH-TS')
                axs[j].fill_between(range(1, len(avg_regret_mean) + 1), avg_regret_mean - avg_regret_std,
                                    avg_regret_mean + avg_regret_std, alpha=0.2)

                axs[j].set_xlabel('iteration')
                axs[j].set_ylabel(metric)

            handles, labels = axs[1].get_legend_handles_labels()
            handles = [handles[0], handles[1], handles[3], handles[2], handles[4]]
            labels = [labels[0], labels[1], labels[3], labels[2], labels[4]]
            lgd = axs[1].legend(handles, labels, handlelength=1.3, handletextpad=0.4, fontsize=10, borderpad=0.3,
                                labelspacing=0.3)
            axs[0].set_ylabel('average regret')
            axs[1].set_ylabel('simple regret')
            axs[0].set_ylim(0.0, 4.7)
            axs[1].set_ylim(0.0, 4.7)
            #fig.tight_layout(rect=(0, 0.2, 0.5, 0.6))
            fig.tight_layout()
            #fig.show()
            fig.savefig("plot_bandits_paper.pdf")
            fig.savefig("/home/jonasrothfuss/Dropbox/Eigene_Dateien/ETH/02_Projects/07_Meta-Learning_NNs/LaTex/figures/plot_bandits_paper.pdf")