# Usage
# python pacoh/experiments/hyperparam_search/display_latex.py --experiment_prefix final --cs_algos pacoh,mlap,maml_reg --cs_datasets sin_20,cauchy,swissfel,physionet_0,physionet_2 --cs_metrics avg_rmse,avg_ll,cal_err
import os
import pandas as pd
import numpy as np
import argparse

from config import DATA_DIR
HPARAM_EXP_DIR = os.path.join(DATA_DIR, 'tune-hparam-pacoh-final')

def get_algo_name(algo_name):
    algo_name = algo_name.lower()
    if 'mlap' in algo_name:
        return 'MLAP'
    elif 'maml' in algo_name:
        return 'BMAML'
    elif 'map' in algo_name:
        return 'MAP'
    elif 'svgd' in algo_name:
        # no map in name only svgd
        return 'SVGD'
    elif 'pacoh' in algo_name:
        return 'PACOH-DL (ours)'
    elif 'np' in algo_name:
        return 'NP'
    elif 'none' in algo_name:
        return 'No meta learning'

    raise Exception("Unexpected algorithm")

def get_caption(metric):
    if metric == "avg_rmse":
        return "Root-Mean-Square Error"
    elif metric == "avg_ll":
        return "Log-Likelihood"
    elif metric == "cal_err":
        return "Calibration Error"
    elif metric == "avg_acc":
        return "Average accuracy"

    raise Exception("Unexpected metric")

def get_algo_subname(csv_file_name):
    csv_file_name = csv_file_name.lower()
    if 'first_order' in csv_file_name:
        return 'First-order'
    elif 'second_order' in csv_file_name:
        return 'Second-order'
    elif 'variational' in csv_file_name:
        return 'Variational'
    elif 'seeger' in csv_file_name:
        return 'Seeger'
    elif 'mcallester' in csv_file_name:
        return 'McAllester'
    elif 'map_svgd' in csv_file_name or 'pacoh_map' in csv_file_name:
        if 'fixed' in csv_file_name:
            return 'MAP-SVGD (FL)'
        return 'MAP-SVGD'
    elif 'svgd_svgd' in csv_file_name or 'pacoh_svgd' in csv_file_name:
        if 'fixed' in csv_file_name:
            return 'SVGD-SVGD (FL)'
        return 'SVGD-SVGD'
    elif 'svgd' in csv_file_name:
        if 'fixed' in csv_file_name:
            return 'SVGD (FL)'
        return 'SVGD'
    elif 'np_reg' in csv_file_name:
        return ''

    raise Exception("Unexpected algorithm:", csv_file_name)

def get_dataset_name(dataset):
    if dataset == 'sin_20' or dataset == 'sine_20':
        return 'Sinusoid-Affine'
    elif dataset == 'old_sine':
        return 'Sinusoid'
    elif dataset == 'swissfel':
        return 'SwissFel'
    elif 'cauchy' in dataset:
        return 'Cauchy'
    elif 'physionet' in dataset:
        variable_list = ['GCS', 'Urine', 'HCT', 'BUN', 'Creatinine', 'DiasABP']

        variable_id = int(dataset.split('_')[-1])
        variable_name = variable_list[variable_id]

        dataset_name = f"Physionet-{variable_name}"
        return dataset_name
    elif 'omniglot_5shot_small' in dataset:
        return "Omniglot-5-shot-S"
    elif 'omniglot_2shot_small' in dataset:
        return "Omniglot-2-shot-S"
    elif 'omniglot_5shot_full' in dataset:
        return "Omniglot-5-shot-F"
    elif 'omniglot_2shot_full' in dataset:
        return "Omniglot-2-shot-F"
    elif 'berkeley' in dataset:
        return "Berkeley-Sensor"


    raise Exception("Unexpected dataset:", dataset)

# Groups different versions of the algorithm under one name
sub_algos_dict = {'maml_reg': ['maml_reg_first_order', 'maml_reg_second_order'],
            'maml_class': ['maml_class_first_order', 'maml_class_second_order'],
             #'maml': ['BMAML'],
             'mlap': ['mlap_mcallester', 'mlap_variational', 'mlap_seeger'],
             'mlap_class': ['mlap_mcallester', 'mlap_variational', 'mlap_seeger'],
             # 'pacoh': ['svgd_svgd', 'MAP_svgd'],
             # 'pacoh': ['pacoh_svgd', 'pacoh_map'],
             #'pacoh': ['pacoh_svgd_learned_ll', 'pacoh_svgd_fixed_ll', 'pacoh_map_learned_ll', 'pacoh_map_fixed_ll'],
             'pacoh': ['pacoh_svgd_learned_ll_logsumexp', 'pacoh_svgd_fixed_ll__logsumexp',
                       'pacoh_map_learned_ll_logsumexp', 'pacoh_map_fixed_ll_logsumexp'],
             'pacoh_class': ['pacoh_svgd_class', 'pacoh_map_class'],
             'nps': ['np_reg'],
             'eMLAP': ['eMLAP_Seeger', 'eMLAP_variational', 'eMLAP_McAllester'],
             'none': ['svgd_fixed_context', 'svgd_learned_context']
             }

def main(args):
    DATASETS = args.cs_datasets.split(',')  # put dataset name here to filer results
    ALGOS = args.cs_algos.split(',')
    METRICS = args.cs_metrics.split(',')
    num_columns = len(DATASETS) + 1

    header = ['\\multicolumn{1}{c|}{%s}' % get_dataset_name(ds) for ds in DATASETS[:-1]]
    header.append('\\multicolumn{1}{c}{%s}' % get_dataset_name(DATASETS[-1]))
    header = ' & '.join(header)
    header = f'& {header} \\\\'

    algo_header = """\\hline
\\multicolumn{%d}{c}{\\vspace{-2mm}}\\\\
\\multicolumn{%d}{c}{\\textbf{%s}}\\\\
\\hline
""" % (num_columns, num_columns, '%s')

    csv_files = [path for path in os.listdir(HPARAM_EXP_DIR) if '.csv' in path and path.startswith(args.experiment_prefix)]
    latex = [header]
    for algo_name in ALGOS:
        latex.append(algo_header % get_algo_name(algo_name))
        csv = ""

        sub_algos = sub_algos_dict[algo_name]
        for sub_algo in sub_algos:
            results = []
            for ds_name in DATASETS:
                found = False
                for csv_name in csv_files:
                    if (sub_algo in csv_name) and (ds_name in csv_name):
                        # Make sure that no more than one file matches
                        if found:
                            print(algo_name, ds_name)
                            print(csv)
                            print(csv_name)
                            raise Exception

                        found = True
                        csv = csv_name
                        CSV_PATH = os.path.join(HPARAM_EXP_DIR, csv_name)

                        # 1) load df
                        df = pd.read_csv(CSV_PATH)

                        # 2) aggregate over seeds
                        hyperparams = list(
                            set(df.columns) - {'Unnamed: 0', 'timesteps_total', 'random_seed', 'randomness_config',
                                                'cal_err', 'avg_ll', 'avg_rmse', 'loss', 'train_obj', 'avg_acc'})

                        if 'maml' not in csv_name:
                            df_aggregated = df.groupby(hyperparams).aggregate(
                                {m: [np.mean, np.std] for m in METRICS})

                            # print(csv_name)
                            # print(df_aggregated.reset_index(drop=True).to_string(), '\n')
                            assert args.criterion in ['avg_ll', 'avg_rmse']

                            if args.criterion == 'avg_ll':
                                opt_idx = df_aggregated[args.criterion]['mean'].idxmax()
                            elif args.criterion == 'avg_rmse':
                                opt_idx = df_aggregated[args.criterion]['mean'].idxmin()
                        else:
                            if METRICS[0] not in ['avg_rmse', 'avg_acc']:
                                found = False
                                continue

                            df_aggregated = df.groupby(hyperparams).aggregate(
                                {m: [np.mean, np.std] for m in METRICS if m == METRICS[0]})

                            if METRICS[0] == 'avg_rmse':
                                opt_idx = df_aggregated[METRICS[0]]['mean'].idxmin()
                            elif METRICS[0] == 'avg_acc':
                                opt_idx = df_aggregated[METRICS[0]]['mean'].idxmax()

                        mean = df_aggregated.loc[opt_idx][METRICS[0]]['mean']
                        std = df_aggregated.loc[opt_idx][METRICS[0]]['std']
                        res_string = f"${mean:.03f} \pm {std:.03f}$"

                # Check if there is at least one match
                if not found:
                    res_string = 'N/A'

                results.append(res_string)

            results_str = " & ".join(results)
            algo_name = get_algo_subname(sub_algo)
            latex.append('%s & %s \\\\' % (algo_name, results_str))

    footer = """\\hline\n"""

    latex.append(footer)
    latex_code = '\n'.join(latex)
    start_tag = f"""\\begin{{table*}}[h]
\\begin{{subtable}} %{{\linewidth}}
\centering\caption{{
{get_caption(METRICS[0])}
}}
    
\\begingroup
\setlength{{\\tabcolsep}}{{4pt}} % Default value: 6pt
\\renewcommand{{\\arraystretch}}{{0.92}}

\\begin{{tabular}}{{c||"""
    for _ in range(len(DATASETS)-1):
        start_tag += 'c|'
    start_tag += 'c}'
    end_tag = """
\end{tabular}
\endgroup
\end{subtable}
\end{table*}"""
    latex_code = '\n'.join([start_tag, latex_code, end_tag])

    with open("latex_table.txt", "w") as f:
        f.write(latex_code)

    print(latex_code)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run meta mll hyper-parameter search.')
    parser.add_argument("--experiment_prefix", type=str, required=True)
    parser.add_argument("--cs_datasets", type=str, required=True)
    parser.add_argument("--cs_algos", type=str, required=True)
    parser.add_argument("--criterion", type=str, default='avg_ll')
    parser.add_argument("--cs_metrics", type=str, required=True)
    args = parser.parse_args()
    cs_metrics = args.cs_metrics.split(',')
    for m in cs_metrics:
        print("\n\n%=====================", m.upper(), "START =====================\n")
        args.cs_metrics = f'{m},{args.criterion}'
        main(args)
        print("\n%=====================", m.upper(), "END =====================\n\n")