import os
import pandas as pd
import numpy as np
import argparse
import collections
from pprint import pprint

import json
from pandas.io.json import json_normalize

from config import DATA_DIR
HPARAM_EXP_DIR = os.path.join(DATA_DIR, 'tune-hparam-pacoh')


expected_algos = {'pacoh_svgd_learned_ll', 'pacoh_svgd_fixed_ll', 'pacoh_map_learned_ll', 'pacoh_map_fixed_ll'}

def flatten(d, parent_key='', sep='/'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)

def process_dict_string(s):
    s = s.replace("'", '\"').replace('None', 'null').replace('True', 'true').replace('False', 'false')
    dict = json.loads(s)
    flat_dict = flatten(dict, sep='/')
    return flat_dict

def aggregate_results(configs):
    results = {}

    if type(configs) == list:
        keys = set().union(*[set(d.keys()) for d in configs])

    for key in keys:
        agg_list = [d[key] for d in configs]

        agg_dict = {}

        if type(agg_list[0]) != list:
            agg_dict['all_values'] = set(agg_list)
        else:
            agg_dict['all_values'] = agg_list

        try:
            min_ = min(agg_list)
            max_ = max(agg_list)
            agg_dict['min'] = min_
            agg_dict['max'] = max_
        except:
            pass

        results[key] = agg_dict

    return results

def main(args):
    DATASETS = args.cs_datasets.split(',')  # put dataset name here to filer results
    ALGOS = args.cs_algos.split(',')

    assert len(set(ALGOS)) == len(ALGOS) and set(ALGOS) <= expected_algos

    range_per_file = {}

    csv_files = [path for path in os.listdir(HPARAM_EXP_DIR) if '.csv' in path and path.startswith(args.prefix)]
    for algo_name in ALGOS:
        csv = ""

        for ds_name in DATASETS:
            found = False
            temp = {}

            for csv_name in csv_files:
                if (algo_name in csv_name) and (ds_name in csv_name):
                    # Make sure that no more than one file matches
                    if found:
                        print(f"Two csv files were matched with algorithm `{algo_name}` and dataset `{ds_name}`")
                        print(f"The two files are: `{csv}` and `{csv_name}`")
                        raise Exception

                    found = True
                    csv = csv_name
                    CSV_PATH = os.path.join(HPARAM_EXP_DIR, csv_name)

                    # 1) load df
                    df = pd.read_csv(CSV_PATH)

                    columns = ['data_config', 'hyper_posterior_config', 'hyper_prior_config', 'inference_training_config', 'likelihood_config',
                               'model_config', 'posterior_config', 'prior_config', 'training_config']

                    for col in columns:
                        flattened_configs = [process_dict_string(run) for run in df[col]]
                        res = aggregate_results(flattened_configs)
                        temp[col] = res

                    range_per_file[csv_name] = temp

                    if args.show_per_dataset_results:
                        pprint.pprint(temp)

            # # Check if there is at least one match
            if not found:
                print(f"No file found for algo `{algo_name}` and dataset `{ds_name}`")

    return range_per_file

def merge_per_file_results(dicts):
    config_names = set().union(*[d.keys() for file_name, d in dicts.items()])

    final_configs = {}

    for cn in config_names:
        config = {}
        configs = [d[cn] for file_name, d in dicts.items()]

        keys = set().union(*[set(d.keys()) for d in configs])

        for key in keys:
            all_data = [d[key] for d in configs]

            agg_dict = {}

            if type(all_data[0]['all_values']) == list:
                agg_dict['all_values'] = [d['all_values'] for d in all_data]
            else:
                agg_dict['all_values'] = set().union(*[d['all_values'] for d in all_data])

            try:
                min_ = min([d['min']for d in all_data])
                max_ = max([d['max'] for d in all_data])
                agg_dict['min'] = min_
                agg_dict['max'] = max_
            except:
                pass

            config[key] = agg_dict

        final_configs[cn] = config

    return final_configs

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run meta mll hyper-parameter search.')
    parser.add_argument("--prefix", type=str, required=True)
    parser.add_argument("--cs_datasets", type=str, required=True)
    parser.add_argument("--cs_algos", type=str, required=True)
    parser.add_argument("--show_per_dataset_results", action="store_true", default=False)

    args = parser.parse_args()

    range_per_file = main(args)
    res = merge_per_file_results(range_per_file)

    print("Overall result")
    pprint(res)
