import ray
import copy
import pandas as pd
import sys
import os
import pacoh.experiments.hyperparam_search.util as util
import pacoh.experiments.hyperparam_search.tuners as tuners

from pacoh.experiments.custom_tune import HyperOptSearch
import pacoh.experiments.custom_tune as custom_tune
from ray.tune import Analysis
from datetime import datetime

import argparse

from config import BASE_DIR, DATA_DIR
HPARAM_EXP_DIR = os.path.join(DATA_DIR, 'tune-hparam-pacoh')

SEED = 28
# N_THREADS_PER_RUN = 1
TEST_SEEDS = [28, 29, 30, 31, 32]


def main(args):
    # get rid of tf debug infos and warnings
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

    # ray.init(num_cpus=args['num_cpus'], memory=500 * 1024 ** 3, object_store_memory=200 * 1024 ** 3)
    ray.init(num_cpus=args['num_workers'], memory=(args['mem_gb']-30) * 1024 ** 3, object_store_memory=20 * 1024 ** 3)

    def train_reg(config, reporter):
        args = config.pop('args')
        print(args)
        sys.path.append(BASE_DIR)

        print('Init config:', config)
        # 1) Update config for current run
        util.process_config(config)
        print('Processed config: ', config)

        # 2) setup model
        import pacoh.experiments.hyperparam_search.tuners as tuners
        modelTuner = tuners.get_tuner_class(args)
        modelTuner.init_run(config)

        # 3) train and evaluate model
        for i in range(modelTuner.n_optimization_steps):
            results = modelTuner.run_iters(modelTuner.n_iterations_per_step, reporter=reporter)
            reporter(**results)

    @ray.remote
    def train_test(config):
        args = config.pop('args')
        args['testing'] = True

        print('Init config:', config)
        # 1) Update config for current run
        util.process_config(config)
        print('Processed config: ', config)

        # 2) setup model
        import pacoh.experiments.hyperparam_search.tuners as tuners
        modelTuner = tuners.get_tuner_class(args)
        modelTuner.init_run(config)

        # 3) train and evaluate model
        results = modelTuner.run_iters(modelTuner.n_total_iterations)
        config.update(results)

        return config

    modelTuner = tuners.get_tuner_class(args)
    if args['load_analysis_from'] != '':
        assert os.path.isdir(args['load_analysis_from']), 'load_analysis_from must be a valid directory'
        print('Loading existing tune analysis results from %s' % args['load_analysis_from'])
        analysis = Analysis(args['load_analysis_from'])
        exp_name = os.path.basename(args['load_analysis_from'])
    else:
        exp_name = '%s_%s_%s_%s' % (EXP_PREFIX, args['dataset'], args['meta_algo_type'], args['inference_algo_type'])

        # determine metric to optimize for
        if len(args['metric']) > 0:
            metric = args['metric']
            mode = "max" if ('ll' in args['metric'] or 'acc' in args['metric']) else "min"
        else:
            metric = modelTuner.metric
            mode = modelTuner.mode

        if args['test_only']:
            print("--------Running only the test step--------")
            path = os.path.join(HPARAM_EXP_DIR, exp_name)
            print('Loading existing tune analysis results from %s' % path)
            analysis = Analysis(path)
        else:
            space = modelTuner.get_search_space()
            base_config = modelTuner.get_base_config()
            base_config['args'] = args


            config = {
                "num_samples": modelTuner.num_samples,
                "config": base_config,
                "stop": {
                    "timesteps_total": 300000
                },
            }
            print(config)

            # Run hyper-parameter search

            algo = HyperOptSearch(
                space,
                max_concurrent=args['num_workers'],
                metric=metric,
                mode=mode)

            analysis = custom_tune.run(train_reg, name=exp_name, search_alg=algo, verbose=1, max_failures=2, raise_on_failed_trial=False,
                                local_dir=HPARAM_EXP_DIR, resume=args['resume'], **config)

    from pacoh.experiments.hyperparam_search.util import select_best_configs
    best_configs = select_best_configs(analysis, metric=metric, mode=mode, N=args['n_test_runs'])

    test_configs = []
    for config in best_configs:
        for seed in TEST_SEEDS:
            test_config = copy.deepcopy(config)
            test_config['random_seed'] = seed
            test_configs.append(test_config)

    result_dicts = ray.get([train_test.remote(config) for config in test_configs])

    result_df = pd.DataFrame(result_dicts)
    print(result_df.to_string())

    csv_file_name = os.path.join(HPARAM_EXP_DIR, '%s_%s.csv' % (exp_name, datetime.now().strftime("%b_%d_%Y_%H:%M:%S")))
    result_df.to_csv(csv_file_name)
    print("\nSaved result csv to %s" % csv_file_name)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run meta mll hyper-parameter search.')
    parser.add_argument('--dataset', type=str, default='sin_20', help='dataset')
    parser.add_argument('--mem_gb', type=int, default=100, help='Number of workers')
    parser.add_argument('--meta_algo_type', type=str, default='np_reg')
    parser.add_argument('--inference_algo_type', type=str, default='svgd')
    parser.add_argument('--n_test_runs', type=int, default=5, help='number of test runs')
    parser.add_argument("--resume", action="store_true", help="Whether to resume the experiments from a checkpoint")
    parser.add_argument("--load_analysis_from", type=str, default='')
    parser.add_argument("--experiment_prefix", type=str, required=True)
    parser.add_argument('--num_cpus', type=int, default=1, help='Number of cpus used')
    parser.add_argument('--num_workers', type=int, default=-1, help='Number of workers')
    parser.add_argument('--metric', type=str, default='', help='What metric to optimize for')
    parser.add_argument("--test_only", action="store_true", help="Whether to perform the test step only", default=False)

    args = parser.parse_args()

    if args.num_workers == -1:
        args.num_workers = args.num_cpus

    EXP_PREFIX = args.experiment_prefix
    args.no_cuda = True

    print('Running', os.path.abspath(__file__), '\n')
    print("--- Experiment Settings ---")
    print("Experiment prefix:", EXP_PREFIX)
    print("Dataset:", args.dataset)
    print("Num CPUS:", args.num_cpus)
    print("Num workers:", args.num_workers)
    print("Available memory (in gigabytes):", args.mem_gb)
    print("Meta algorithm type:", args.meta_algo_type)
    print("Meta-test algorithm type:", args.inference_algo_type)
    print("Num test runs:", args.n_test_runs)
    print("Resume:", args.resume)
    print("Test only:", args.test_only)
    print('\n')

    args_dict = vars(args)
    print(args_dict)
    main(args_dict)
