import os
import sys
import argparse

from datetime import datetime
from config import BASE_DIR
sys.path.append(BASE_DIR)

from pacoh.experiments.utils import generate_launch_commands

import pacoh.experiments.hyperparam_search.hyperparam_sweep as hyperparam_sweep
import math

# Datasets for which the data batching has been implemented
SUPPORTED_DATASETS = {'cauchy_20', 'sin_20', 'physionet_0', 'physionet_1', 'physionet_2', 'swissfel',
                      'omniglot_5shot_full', 'omniglot_2shot_full', 'omniglot_5shot_small',
                      'omniglot_2shot_small', 'omniglot_5shot_full_all_tasks', 'omniglot_2shot_full_all_tasks',
                      'omniglot_5shot_small_all_tasks',
                      'omniglot_2shot_small_all_tasks', 'mhc_0', 'mhc_1',
                      'berkeley'}

# Meta-training (hyper-posterior inference) algorithm IDs
# (or IDs for a particular training setting of an algorithm)
# for which a tunner class has been implemented
SUPPORTED_META_ALGOS = {'pacoh_svgd_learned_ll', 'pacoh_svgd_fixed_ll', 'pacoh_svgd_class_cnn',
                        'pacoh_svgd_class_fc',

                        'pacoh_map_learned_ll', 'pacoh_map_fixed_ll', 'pacoh_map_class_cnn',
                        'pacoh_map_class_fc',

                        'pacoh_svgd_learned_ll_logsumexp', 'pacoh_svgd_fixed_ll_logsumexp',
                        'pacoh_svgd_class_cnn_logsumexp','pacoh_svgd_class_fc_logsumexp',

                        'pacoh_map_learned_ll_logsumexp', 'pacoh_map_fixed_ll_logsumexp',
                        'pacoh_map_class_cnn_logsumexp', 'pacoh_map_class_fc_logsumexp',

                        'mlap_mcallester_mse', 'mlap_seeger_mse',
                        'mlap_variational_mse', 'mlap_mcallester_class_cnn', 'mlap_variational_class_cnn',
                        'mlap_seeger_class_cnn',

                        'maml_reg_first_order', 'maml_reg_second_order',
                        'maml_class_first_order', 'maml_class_second_order',

                        'bmaml_reg_learned_ll', 'bmaml_reg_fixed_ll', 'bmaml_class',
                        'bmaml_reg_learned_ll_chaser', 'bmaml_reg_fixed_ll_chaser', 'bmaml_class_chaser',

                        'np_reg',
                        
                        'none'}

# Supported posterior inference algorithms
SUPPORTED_INFERENCE_ALGOS = {'svgd',
                             'svgd_fixed_context', 'svgd_fixed_full',
                             'svgd_learned_context', 'svgd_learned_full',
                             'svgd_class_fixed_context'}

def main(args):
    ### Process the arguments related to the executing machines and their resources
    available_memory = args.mem_gb
    n_cpus = args.n_cpus
    n_gpus = args.n_gpus
    memory_per_core = math.floor((available_memory - 2) * 1024 / n_cpus)

    dry_run = args.dry_run
    assert not (args.run_locally and args.run_oci)
    run_on_cluster = not args.run_locally

    for attr in ['mem_gb', 'n_cpus', 'n_gpus', 'dry_run', 'run_locally']:
        delattr(args, attr)
    ###

    command_list = []
    booleans = []

    # Process any expected boolean parameters that need to be passed to the hyperparameter_sweep script
    for key in ['resume', 'test_only']:
        if getattr(args, key):
            booleans.append(key)

    # Generate commands
    for meta_algo in args.meta_algos:
        for inference_algo in args.inference_algos:
            exp_config = {
                'dataset': args.datasets,
                'meta_algo_type': [meta_algo],
                'inference_algo_type': [inference_algo],
                'num_cpus': [n_cpus],
                'mem_gb': [available_memory],
                'experiment_prefix': [args.experiment_prefix],
                'num_workers': [args.n_workers],
                'metric': [args.metric]
            }

            if args.load_analysis_from != '':
                exp_config['load_analysis_from'] = [args.load_analysis_from]

            command_list += generate_launch_commands(hyperparam_sweep, exp_config, check_flags=False, booleans=booleans)

    print(command_list)

    # Run commands
    if args.run_oci:
        answer = input("About to submit %i compute jobs to the oci cloud. Proceed? [yes/no]\n" % len(command_list))
        if answer == 'yes':
            import oci_launcher.mode as mode
            import oci_launcher.mount as mount
            from oci_launcher.launch import launch_python
            from pacoh.experiments.utils import AsyncExecutor

            def launch_command_oci(command):
                target_file = command.split(" ")[1]
                cmd_args = " ".join(command.split(" ")[2:])

                REMOTE_PYTHON_INTERPRETER = '/home/ubuntu/miniconda3/envs/meta-nn/bin/python -u'

                mode_oci = mode.OCIMode(oci_shape='BM.Standard.E2.64')

                MODE = mode_oci

                mounts = [
                    mount.MountLocal(local_dir=BASE_DIR + '/', mount_point='/home/ubuntu/meta-learning-NNs',
                                     pythonpath=True),
                    mount.MountDropbox(mount_point='meta-learning-NNs/data', dropbox_path='/meta-learning-NNs_data',
                                       output=True, skip_folders='omniglot_resized,physionet2012,sensor_data,swissfel,2018_10_31')
                ]

                datetime_str = datetime.now().strftime("%d-%m-%y_%H:%M:%S.txt")
                launch_python(target=target_file, mount_points=mounts,
                              target_mount_dir='/home/ubuntu/meta-learning-NNs',
                              mode=MODE, args=cmd_args, verbose=True,
                              python_cmd=REMOTE_PYTHON_INTERPRETER, tmux=True,
                              stdout_file='/home/ubuntu/meta-learning-NNs/data/%s_%s_%s.out'%(args.experiment_prefix, datetime_str, str(command.__hash__())))

            exec = AsyncExecutor(n_jobs=len(command_list))
            exec.run(launch_command_oci, command_list)

    elif run_on_cluster:
        cluster_cmds = []
        for python_cmd in command_list:
            bsub_cmd = f'bsub' \
                       f' -W {args.time}' \
                       f' -R "rusage[mem={memory_per_core}]"' \
                       f' -n {n_cpus}' \
                       f' -R "span[hosts=1]"'

            if n_gpus > 0 :
                bsub_cmd += f' -R "rusage[ngpus_excl_p={n_gpus}]"'

            cluster_cmds.append(bsub_cmd + ' ' + python_cmd)

        answer = input("About to submit %i compute jobs to the cluster. Proceed? [yes/no]\n" % len(cluster_cmds))
        if answer == 'yes':
            for cmd in cluster_cmds:
                if dry_run:
                    print(cmd)
                else:
                    os.system(cmd)

    else:
        answer = input("About to run %i compute jobs in a sequential for loop. Proceed? [yes/no]\n" % len(command_list))
        if answer == 'yes':
            for cmd in command_list:
                if dry_run:
                    print(cmd)
                else:
                    os.system(cmd)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run meta mll hyper-parameter search.')
    parser.add_argument("--resume", action="store_true", help="Whether to resume the experiments from a checkpoint")
    parser.add_argument("--time", type=str, default='23:50')
    parser.add_argument("--load_analysis_from", type=str, default='')
    parser.add_argument("--cs_datasets", type=str, default='sin_20')
    parser.add_argument("--cs_meta_algos", type=str, default='pacoh_svgd_fixed_ll,')
    parser.add_argument("--cs_inference_algos", type=str, default='svgd,')
    parser.add_argument("--mem_gb", type=int, default='256')
    parser.add_argument("--n_cpus", type=int, default='20')
    parser.add_argument("--n_workers", type=int, default='-1')
    parser.add_argument("--n_gpus", type=int, default='1')
    parser.add_argument("--run_locally", action="store_true", help="Whether to run the experiments locally instead of "
                                                                   "submitting the jobs to the cluster", default=False)
    parser.add_argument("--run_oci", action="store_true", help="Whether to run the commands in the oracle cloud", default=False)
    parser.add_argument('--metric', type=str, default='', help='What metric to optimize for')
    parser.add_argument("--dry_run", action="store_true", help="Print commands instead of running them")
    parser.add_argument("--experiment_prefix", type=str, help="A prefix to identify all the runs from this experiment", required=True)
    parser.add_argument("--test_only", action="store_true", help="Whether to perform the test step only", default=False)

    args = parser.parse_args()
    args.datasets = [i.strip() for i in args.cs_datasets.split(',') if i != '']
    del args.cs_datasets
    assert len(set(args.datasets)) == len(args.datasets) and set(args.datasets) <= SUPPORTED_DATASETS

    args.meta_algos = [i.strip() for i in args.cs_meta_algos.split(',') if i != '']
    del args.cs_meta_algos
    assert len(set(args.meta_algos)) == len(args.meta_algos) and set(args.meta_algos) <= SUPPORTED_META_ALGOS

    args.inference_algos = [i.strip() for i in args.cs_inference_algos.split(',') if i != '']
    del args.cs_inference_algos
    assert len(set(args.inference_algos)) == len(args.inference_algos) and set(args.inference_algos) <= SUPPORTED_INFERENCE_ALGOS
    # Run generation that supports multiple inference algorithm is currently implemented only for baselines
    assert len(args.inference_algos) == 1 or (len(args.meta_algos) == 1 and args.meta_algos[0] == 'none')

    main(args)
