from abc import ABC, abstractmethod

class ModelTuner(ABC):
    @abstractmethod
    def __init__(self, args):
        pass

    @abstractmethod
    def init_run(self, config, random_seed=None):
        raise NotImplementedError

    @abstractmethod
    def run_iters(self, n_iter):
        raise NotImplementedError

    @abstractmethod
    def get_search_space(self):
        raise NotImplementedError

    @abstractmethod
    def get_base_config(self):
        raise NotImplementedError


def get_tuner_class(args):
    from .tuner_pacoh_svgd_regression import PACOH_SVGD_Regression_Tuner_FixedLikelihood,PACOH_SVGD_Regression_Tuner_LearnableLikelihood
    from .tuner_pacoh_svgd_classification import PACOH_SVGD_Classification_Tuner
    from .tuner_mlap_regression import MLAP_McAllester_MSE, MLAP_Seeger_MSE, MLAP_Variational_MSE
    from .tuner_mlap_classification import MLAP_Seeger_Classification_CNN, MLAP_Variational_Classification_CNN, \
        MLAP_McAllester_Classification_CNN
    from .tuner_maml_regression import MAML_Regression_Tuner
    from .tuner_maml_classification import MAML_Classification_Tuner
    from .tuner_maml_classification_ntasks import MAML_Classification_Tuner_NTASKS
    from .tuner_np_regression import NP_Regression_Tuner
    from .tuner_bmaml_regression import BMAML_Regression_Tuner
    from .tuner_bmaml_classification import BMAML_Classification_Tuner

    # PACOH SVGD
    if 'pacoh_svgd' in args['meta_algo_type'] and not 'class' in args['meta_algo_type']:
        args['dataset_type'] = 'reg'
        args['prior_type'] = 'gaussian'
        args['posterior_type'] = 'particle'
        args['likelihood_type'] = 'reg'
        args['model_type'] = 'fc'
        args['MAP'] = False

        if 'logsumexp' in args['meta_algo_type']:
            args['logsumexp'] = True
        else:
            args['logsumexp'] = False

        if 'learned_ll' in args['meta_algo_type']:
            args['learn_likelihood_variables'] = True
            return PACOH_SVGD_Regression_Tuner_LearnableLikelihood(args)
        elif 'fixed_ll' in args['meta_algo_type']:
            args['learn_likelihood_variables'] = False
            return PACOH_SVGD_Regression_Tuner_FixedLikelihood(args)
        else:
            raise NotImplementedError

    elif 'pacoh_svgd_class' in args['meta_algo_type']:
        args['MAP'] = False
        if 'fc' in args['meta_algo_type']:
            args['model_type'] = 'fc'
        elif 'cnn' in args['meta_algo_type']:
            args['model_type'] = 'cnn'
        else:
            raise NotImplementedError

        if 'logsumexp' in args['meta_algo_type']:
            args['logsumexp'] = True
        else:
            args['logsumexp'] = False

        return PACOH_SVGD_Classification_Tuner(args)

    # PACOH MAP
    if 'pacoh_map' in args['meta_algo_type'] and not 'class' in args['meta_algo_type']:
        args['dataset_type'] = 'reg'
        args['prior_type'] = 'gaussian'
        args['posterior_type'] = 'particle'
        args['likelihood_type'] = 'reg'
        args['model_type'] = 'fc'
        args['MAP'] = True

        if 'logsumexp' in args['meta_algo_type']:
            args['logsumexp'] = True
        else:
            args['logsumexp'] = False

        if 'learned_ll' in args['meta_algo_type']:
            args['learn_likelihood_variables'] = True
            return PACOH_SVGD_Regression_Tuner_LearnableLikelihood(args)
        elif 'fixed_ll' in args['meta_algo_type']:
            args['learn_likelihood_variables'] = False
            return PACOH_SVGD_Regression_Tuner_FixedLikelihood(args)
        else:
            raise NotImplementedError

    elif 'pacoh_map_class' in args['meta_algo_type']:
        args['MAP'] = True
        if 'fc' in args['meta_algo_type']:
            args['model_type'] = 'fc'
        elif 'cnn' in args['meta_algo_type']:
            args['model_type'] = 'cnn'
        else:
            raise NotImplementedError

        if 'logsumexp' in args['meta_algo_type']:
            args['logsumexp'] = True
        else:
            args['logsumexp'] = False

        return PACOH_SVGD_Classification_Tuner(args)

    # MLAP

    elif args['meta_algo_type'] == 'mlap_mcallester_mse':
        args['dataset_type'] = 'reg'
        args['likelihood_type'] = 'reg'
        args['model_type'] = 'fc'
        args['learn_likelihood_variables'] = False
        args['use_mse_as_loss'] = True
        args['bound_type'] = 'mcallester'

        return MLAP_McAllester_MSE(args)

    elif args['meta_algo_type'] == 'mlap_seeger_mse':
        args['dataset_type'] = 'reg'
        args['likelihood_type'] = 'reg'
        args['model_type'] = 'fc'
        args['learn_likelihood_variables'] = False
        args['use_mse_as_loss'] = True
        args['bound_type'] = 'seeger'
        return MLAP_Seeger_MSE(args)

    elif args['meta_algo_type'] == 'mlap_variational_mse':
        args['dataset_type'] = 'reg'
        args['likelihood_type'] = 'reg'
        args['model_type'] = 'fc'
        args['learn_likelihood_variables'] = False
        args['use_mse_as_loss'] = True
        args['bound_type'] = 'variational'
        return MLAP_Variational_MSE(args)

    elif args['meta_algo_type'] == 'mlap_seeger_class_cnn':
        args['model_type'] = 'cnn'
        args['bound_type'] = 'seeger'
        return MLAP_Seeger_Classification_CNN(args)

    elif args['meta_algo_type'] == 'mlap_variational_class_cnn':
        args['model_type'] = 'cnn'
        args['bound_type'] = 'variational'
        return MLAP_Variational_Classification_CNN(args)

    elif args['meta_algo_type'] == 'mlap_mcallester_class_cnn':
        args['model_type'] = 'cnn'
        args['bound_type'] = 'mcallester'
        return MLAP_McAllester_Classification_CNN(args)

    # MAML

    elif args['meta_algo_type'] == 'maml_reg_first_order':
        args['first_order'] = True
        args['dataset_type'] = 'reg'
        return MAML_Regression_Tuner(args)

    elif args['meta_algo_type'] == 'maml_reg_second_order':
        args['first_order'] = False
        args['dataset_type'] = 'reg'
        return MAML_Regression_Tuner(args)

    elif args['meta_algo_type'] == 'maml_class_first_order':
        args['first_order'] = True
        args['dataset_type'] = 'class'
        return MAML_Classification_Tuner(args)

    elif args['meta_algo_type'] == 'maml_class_second_order':
        args['first_order'] = False
        args['dataset_type'] = 'class'
        return MAML_Classification_Tuner(args)

    # MAML n_tasks
    elif args['meta_algo_type'] == 'maml_class_ntasks':
        args['first_order'] = False
        args['dataset_type'] = 'class'
        return MAML_Classification_Tuner_NTASKS(args)

    elif args['meta_algo_type'] == 'np_reg':
        return NP_Regression_Tuner(args)

    # BMAML
    elif 'bmaml_reg' in args['meta_algo_type']:

        if 'learned_ll' in args['meta_algo_type']:
            args['learn_likelihood_variables'] = True
        elif 'fixed_ll' in args['meta_algo_type']:
            args['learn_likelihood_variables'] = False
        else:
            raise NotImplementedError

        if 'chaser' in args['meta_algo_type']:
            args['chaser'] = True
        else:
            args['chaser'] = False

        args['dataset_type'] = 'reg'
        return BMAML_Regression_Tuner(args)

    elif args['meta_algo_type'] == 'bmaml_class':
        if 'chaser' in args['meta_algo_type']:
            args['chaser'] = True
        else:
            args['chaser'] = False

        args['dataset_type'] = 'class'
        return BMAML_Classification_Tuner(args)

    # BNN Baseline

    elif args['meta_algo_type'] == 'none' or args['meta_algo_type'] is None:
        return get_bnn_baseline_tunner_class(args)
    else:
        raise NotImplementedError

def get_bnn_baseline_tunner_class(args):
    from .tuner_BNN_regression import BNN_Regression_Tuner
    from .tuner_BNN_classification import BNN_Classification_Tuner

    if args['inference_algo_type'].endswith('full'):
        args['add_meta_data_to_context'] = True
    elif args['inference_algo_type'].endswith('context'):
        args['add_meta_data_to_context'] = False
    else:
        raise NotImplementedError

    if 'learned' in args['inference_algo_type']:
        args['learn_likelihood_variables'] = True
    elif 'fixed' in args['inference_algo_type']:
        args['learn_likelihood_variables'] = False
    else:
        raise NotImplementedError

    if 'svgd' in args['inference_algo_type']:
        if 'class' in args['inference_algo_type']:
            return BNN_Classification_Tuner(args)
        else:
            return BNN_Regression_Tuner(args)
    else:
        raise NotImplementedError
