import copy
import math
import time as time

import numpy as np
import tensorflow as tf
from hyperopt import hp

from pacoh import datasets as datasets, algorithms as algorithms
from .tuner import ModelTuner
import pacoh.experiments.hyperparam_search.util as util

SEED = 28

class BNN_Classification_Tuner(ModelTuner):
    metric = 'avg_ll'
    mode = 'max'
    num_samples = 100  # number of runs for tune
    max_tasks_parallel = 20

    def __init__(self, args):
        self.args = args

    def init_run(self, config):
        config = copy.deepcopy(config)
        data_config = config.pop('data_config')

        prior_config = config['prior_config']
        config['prior_config'] = prior_config = prior_config[prior_config['type']]

        posterior_config = config['posterior_config']
        n_batched_models = posterior_config['n_samples']

        # Setup data
        if data_config['type'] == 'reg':
            _, val_ds = datasets.get_regression_datasets(data_config,
                                                         n_batched_models_train=n_batched_models,
                                                         n_batched_models_test=n_batched_models,
                                                         add_meta_data_to_context=self.args['add_meta_data_to_context'])
        elif data_config['type'] == 'class':
            if self.args['add_meta_data_to_context'] == True:
                raise NotImplementedError
            _, val_ds = datasets.get_classification_datasets(data_config,
                                                             n_batched_models_train=n_batched_models,
                                                             n_batched_models_test=n_batched_models,
                                                             )
        else:
            raise Exception("Unexpected dataset type")

        # Use the first n validation tasks to estimate the loss during hyper_param_tuning
        self.tasks = val_ds.meta_train_tasks

        self.config = config

        if self.args.get('testing', False):
            model = algorithms.BNN_SVGD.BayesianNeuralNetworkSVGD(self.tasks[0], **config)
            self.run_iters = self.run_iters_test
        else:
            print(f"Using the first `{self.max_tasks_parallel}` validation tasks for hyper-parameter tuning.")
            tasks_tuning = self.tasks[:self.max_tasks_parallel]
            self.models, self.step = self._setup_models_and_step(tasks=tasks_tuning)
            model = self.models[0]

        self.n_optimization_steps = model.train_config['n_optimization_steps']
        self.n_iterations_per_step = model.train_config['n_iterations_per_step']
        self.n_total_iterations = model.train_config['n_total_iterations']

        self.log_period = model.train_config['log_period']
        self.n_iter = 0

    def _setup_models_and_step(self, tasks):
        # initialize BNN models with the meta-learned prior for each meta-test task
        self.n_iter = 0

        models = []

        for task in tasks:
            # each meta_test_task corresponds to a dataset sampler
            model = algorithms.BNN_SVGD.BayesianNeuralNetworkSVGD(task, **self.config)
            models.append(model)

        # define parallel training step on all the models
        @tf.function
        def _step_tf(batch_per_model):
            log_likelihood_per_eval_model = tf.stack([tf.reduce_mean(model.step(batch[0], batch[1]))
                                                      for model, batch in zip(models, batch_per_model)])

            return log_likelihood_per_eval_model

        def step():
            batch_per_model = [model.dataset_sampler.get_batch() for model in models]
            log_prob_per_model = _step_tf(batch_per_model)

            return log_prob_per_model

        assert len(models) == len(tasks)
        return models, step

    def eval(self, models, print_results):
        eval_metrics_grouped = {}

        for m in models:
            res, _ = m.eval()
            for key in res:
                l = eval_metrics_grouped.get(key, [])
                l.append(res[key].numpy())
                eval_metrics_grouped[key] = l

        eval_metrics_mean = {}
        eval_metrics_std = {}
        for key in eval_metrics_grouped:
            eval_metrics_mean[key] = np.mean(eval_metrics_grouped[key])
            eval_metrics_std[key] = np.std(eval_metrics_grouped[key])

        if print_results:
            message = ""
            for key in eval_metrics_mean:
                message += '- Val-%s: %.3f +- %.3f' % (key, eval_metrics_mean[key], eval_metrics_std[key])
            print('\n\t%s' % message)

        return eval_metrics_mean, eval_metrics_std, eval_metrics_grouped

    def run_iters(self, n_iter_per_step, reporter=None):
        t = time.time()
        n_iter = self.n_iter
        for _ in range(n_iter_per_step):
            n_iter += 1
            log_prob = self.step()
            avg_log_prob = tf.reduce_mean(log_prob).numpy()
            if n_iter % self.log_period == 0:
                message = '\nIter %d/%d - Time %.2f sec' % (n_iter, self.n_total_iterations, time.time() - t)
                message += ' - Train-Avg-Log-Prob: %.3f' % avg_log_prob
                print(message)
                t = time.time()
                if reporter is not None:
                    reporter(timesteps_total=n_iter, train_obj=avg_log_prob, avg_ll=math.nan, avg_acc=math.nan,
                             cal_err=math.nan)

        self.n_iter = n_iter

        eval_metrics_mean, eval_metrics_std, _ = self.eval(self.models, print_results=True)

        results = {'timesteps_total': self.n_iter, 'train_obj': avg_log_prob}
        results.update(eval_metrics_mean)
        return results

    def _meta_test_training_loop(self, step, n_total_iterations):
        n_iter = 0
        t = time.time()
        for _ in range(n_total_iterations):
            n_iter += 1
            log_prob = step()

            # print meta-test training loss
            if n_iter % self.log_period == 0:
                avg_log_prob = tf.reduce_mean(log_prob).numpy()
                message = '\tMeta-Test - Iter %d/%d - Time %.2f sec' % (
                    n_iter, n_total_iterations, time.time() - t)
                message += ' - Train-Log-Prob: %.3f' % avg_log_prob
                print(message)
                t = time.time()

    def run_iters_test(self, n_total_iterations):
        tasks = self.tasks
        task_batches = util._split_into_batches(tasks, max_batch_size=self.max_tasks_parallel)

        # Start training the BNN models
        print("\tStart meta-test posterior inference ------------------")
        if len(task_batches) > 1:
            print("\tPerforming meta-testing in %i batches" % len(task_batches))

        eval_metrics_dict_per_task = []
        for eval_batch_id, task_batch in enumerate(task_batches):
            print("\tMeta-Test batch #%i consisting of %i tasks----"%(eval_batch_id+1, len(task_batch)))
            # initialize evaluation models and define their training step
            models, step = self._setup_models_and_step(tasks=task_batch)
            self._meta_test_training_loop(step, n_total_iterations)

            t = time.time()
            _, _, eval_metrics_grouped = self.eval(models, print_results=False)
            eval_metrics_dict_per_task.append(eval_metrics_grouped)

        eval_metrics_mean, eval_metrics_std = util._aggregate_eval_metrics_across_tasks(eval_metrics_dict_per_task)

        message = ""
        for key in eval_metrics_mean:
            message += '- Val-%s: %.3f +- %.3f' % (key, eval_metrics_mean[key], eval_metrics_std[key])
        print('\n\tMeta-Test - Time %.2f sec %s' % (time.time() - t, message))

        results = {'timesteps_total': self.n_iter, 'train_obj': math.nan}
        results.update(eval_metrics_mean)
        return results

    def get_base_config(self):
        likelihood_config = {'class': {'learn_likelihood_variables': False},
                             'reg': {'log_var': 2 * np.log(0.1),
                                     'learn_likelihood_variables': self.args['learn_likelihood_variables']},
                             'type': 'class'
                             }

        if 'omniglot' in self.args['dataset']:
            n_classes = 5
        else:
            raise NotImplementedError("Unknown omniglot dataset")

        n_total_iterations = 10000

        model_config = {'cnn': {'activations': ['relu', 'softmax'],
                                'output_dim': n_classes},
                        'type': 'cnn'}

        data_config = {'dataset': self.args['dataset'],
                       'type': 'class',
                       'num_train_tasks': None,
                       'num_test_valid_tasks': None,
                       'num_train_samples': None,
                       'num_context_samples': None,
                       'num_test_valid_samples': None,
                       'standardization': False,
                       'tiled': True,
                       'batch_size': 4,
                       'meta_batch_size': -1,
                       'posterior_inference_batch_size': 4,
                       }

        model_gaussian_prior = {'mean_mu': 0.0,
                                'bias_log_var_mu': 2 * np.log(1.0),
                                'weight_log_var_mu': 2 * np.log(0.5)
                                }

        likelihood_gaussian_prior = {'log_var_mu': 2 * np.log(0.1),
                                     'log_var_log_var': 2 * np.log(2.0)
                                     }

        prior_config = {'gaussian': {'model': model_gaussian_prior,
                                     'likelihood': likelihood_gaussian_prior,
                                     'init_std': 0.1,
                                     'n_batched_priors': 1,
                                     'add_noise': False
                                     },
                        'type': 'gaussian'
                        }

        posterior_config = {'n_samples': 5,
                            'use_init_std': False,
                            'type': 'particle'
                            }

        train_config = {'n_total_iterations': n_total_iterations,
                        'n_train_loss_logs': 5,
                        'n_eval_runs': 0,
                        'optimizer': 'adam',
                        'learning_rate': 0.002,
                        'plot_functions': False,
                        'plot_pred_lines': False,
                        'plot_pred_std': False,
                        'prior_prob_weight': 1.0,
                        'bandwidth': 10.0
                        }

        config = {'model_config': model_config,
                  'likelihood_config': likelihood_config,
                  'data_config': data_config,
                  'prior_config': prior_config,
                  'posterior_config': posterior_config,
                  'train_config': train_config,
                  'random_seed': None
                  }

        return config

    def get_search_space(self):
        # Slash to go a level deeper in the hierarchy
        # Colon to set multiple parameters to the same value

        if 'svgd' in self.args['inference_algo_type']:
            model_log_var_mu = 'prior_config/gaussian/model/bias_log_var_mu;' \
                               'prior_config/gaussian/model/weight_log_var_mu'

            space = {
                # prior log_var for the bias and kernel parameters
                model_log_var_mu: hp.uniform(model_log_var_mu, math.log(1e-1), math.log(1e1)),


                # "train_config/prior_prob_weight": hp.loguniform("train_config/prior_prob_weight",
                #                                                 math.log(1e-6),
                #                                                 math.log(1e0)),

                "train_config/learning_rate": hp.loguniform("train_config/learning_rate",
                                                                math.log(5e-4),
                                                                math.log(5e-3)),

                # bandwidths
                "train_config/bandwidth":
                    hp.loguniform("train_config/bandwidth", math.log(1e0),
                                  math.log(1e3)),

                "data_config/posterior_inference_batch_size":  hp.choice("data_config/posterior_inference_batch_size", [4, 10]),
            }

            return space
        else:
            raise NotImplementedError