import copy
import math
import time as time

import numpy as np
import tensorflow as tf
from hyperopt import hp

from pacoh import datasets as datasets, algorithms as algorithms
from .tuner import ModelTuner

SEED = 28

SAVE_FIGS = False

# FC NN CONFIG
LAYER_SIZE = 64
N_LAYERS = 4
ACTIVATION = 'relu'


class BMAML_Regression_Tuner(ModelTuner):
    metric = 'avg_ll'
    mode = 'max'
    num_samples = 200 # number of runs for tune

    def __init__(self, args):
        self.args = args

    def init_run(self, config):

        """ --- pre-process config ---- """
        config = copy.deepcopy(config)
        data_config = config['data_config']

        # set activations
        """ --- setup datasets and samplers ---- """

        n_particles = config['meta_train_config']['n_particles']

        if data_config['type'] == 'reg':
            train_ds, val_ds = datasets.get_regression_datasets(data_config,
                                                                n_batched_models_train=n_particles,
                                                                n_batched_models_test=n_particles)
        elif data_config['type'] == 'class':
            train_ds, val_ds = datasets.get_classification_datasets(data_config,
                                                                    n_batched_models_train=n_particles,
                                                                    n_batched_models_test=n_particles)
        else:
            raise Exception("Unexpected dataset type")

        del config['data_config']

        self.meta_model = algorithms.BMAML(meta_train_sampler=train_ds,
                                           meta_test_sampler=val_ds, **config)

        self.n_optimization_steps = self.meta_model.meta_train_config['n_optimization_steps']
        self.n_iterations_per_step = self.meta_model.meta_train_config['n_iterations_per_step']
        self.n_total_iterations = self.meta_model.meta_train_config['n_total_iterations']

        self.log_period = self.meta_model.meta_train_config['log_period']
        self.n_iter = 0

    def run_iters(self, n_iter_per_step, reporter=None):
        t = time.time()
        n_iter = self.n_iter
        for _ in range(n_iter_per_step):
            n_iter += 1
            loss = self.meta_model.step()
            loss = tf.reduce_mean(loss).numpy()
            if n_iter % self.log_period == 0:
                message = '\nIter %d/%d - Time %.2f sec' % (n_iter, self.n_total_iterations, time.time() - t)
                message += ' - Train-Loss: %.3f' % loss
                print(message)
                t = time.time()
                if reporter is not None:
                    reporter(timesteps_total=n_iter, loss=loss, cal_err=math.nan, avg_ll=math.nan, avg_rmse=math.nan)

        self.n_iter = n_iter

        eval_metrics_mean, eval_metrics_std = self.meta_model.meta_eval(print_results=True)

        results = {'timesteps_total': self.n_iter, 'loss': loss}
        results.update(eval_metrics_mean)
        return results

    def get_search_space(self):
        # Slash to go a level deeper in the hierarchy
        # Colon to set multiple parameters to the same value
        space = {
            # learning rates for meta-inference and inference
            "meta_train_config/n_total_iterations": hp.choice("meta_train_config/n_total_iterations", [10000, 30000]),
            "meta_train_config/inner_learning_rate": hp.loguniform("meta_train_config/inner_learning_rate", math.log(5e-3),
                                                           math.log(5e-2)),
            "meta_train_config/bandwidth": hp.loguniform("meta_train_config/bandwidth", math.log(1e-3),
                                                                   math.log(1.0)),
            "meta_train_config/prior_factor": hp.loguniform("meta_train_config/prior_factor", math.log(1e-3),
                                                         math.log(1.0)),
            "meta_train_config/n_inner_steps": hp.choice("meta_train_config/n_inner_steps", [1,2,3]),
            "meta_test_config/n_steps_extra": hp.choice("meta_test_config/n_steps_extra", [0,1,2])
        }
        if self.args['chaser']:
            space["meta_train_config/leader_lr_prefactor"] = hp.loguniform("meta_train_config/leader_lr_prefactor", math.log(0.1), math.log(1.0))
            space["meta_train_config/n_leader_steps_extra"] = hp.choice("meta_train_config/n_leader_steps_extra", [1,2])

        if not self.args['learn_likelihood_variables']:
            space["likelihood_config/reg/log_var"] = hp.loguniform("likelihood_config/reg/log_var", np.log(0.01),
                                                                   np.log(0.1)),
        return space

    def get_base_config(self):

        if 'physionet' in self.args['dataset']:
            batch_size = posterior_inference_batch_size = 6
        elif 'sin' in self.args['dataset'] or 'cauchy' in self.args['dataset'] or 'swissfel' in self.args['dataset']:
            batch_size = posterior_inference_batch_size = -1
        elif 'mhc' in self.args['dataset']:
            batch_size = posterior_inference_batch_size = -1
        elif 'berkeley' in self.args['dataset']:
            batch_size = posterior_inference_batch_size = -1
        else:
            raise NotImplementedError("Unknown regression dataset %s"%self.args['dataset'])

        likelihood_config = {'class': {'learn_likelihood_variables': False},
                             'reg': {'log_var': 2 * np.log(0.1),
                                     'learn_likelihood_variables': self.args['learn_likelihood_variables']},
                             'type': 'reg'
                             }


        model_config = {'fc': {'input_dim': None,
                               'output_dim': 1,
                               'layer_units': [LAYER_SIZE for _ in range(N_LAYERS)] + [1],
                               'activations': [ACTIVATION for _ in range(N_LAYERS)] + [None]
                               },
                        'cnn': {},
                        'type': 'fc'}

        assert(len(model_config['fc']['layer_units']) == len(model_config['fc']['activations']))

        model_gaussian_prior = {'mean_mu': 0.0,
                                'bias_log_var_mu': 2 * np.log(2.0),
                                'weight_log_var_mu': 2 * np.log(1.0)
                                }

        likelihood_gaussian_prior = {'log_var_mu': 2 * np.log(0.1),
                                     'log_var_log_var': 2 * np.log(0.5)
                                     }

        prior_config = {'model': model_gaussian_prior,
                        'likelihood': likelihood_gaussian_prior,
                        'init_std': 0.1,
                        'n_batched_priors': 1,
                        'add_noise': False,
                        }


        data_config = {'dataset': self.args['dataset'],
                       'type': self.args['dataset_type'],
                       'num_train_tasks': None,
                       'num_test_valid_tasks': None,
                       'num_train_samples': None,
                       'num_context_samples': None,
                       'num_test_valid_samples': None,
                       'standardization': True,
                       'tiled': True,
                       'batch_size': batch_size,
                       'posterior_inference_batch_size': posterior_inference_batch_size,
                       'meta_batch_size': 5,
                       }

        meta_train_config = {'n_total_iterations': 30000,
                             'n_train_loss_logs': 50,
                             'n_eval_runs': 0,
                             'optimizer': 'adam',
                             'learning_rate': 0.001,
                             'inner_learning_rate': 0.01,
                             'n_inner_steps': 1,
                             'n_leader_steps_extra': 1,
                             'n_particles': 5,
                             'bandwidth': 0.01,
                             'chaser_loss': self.args['chaser'],
                             'leader_lr_prefactor': 1.0,
                             'split_batch': True,
                             'prior_factor': 0.001,
                             }

        meta_test_config = {'n_steps_extra': 0}

        config = {'model_config': model_config,
                  'likelihood_config': likelihood_config,
                  'data_config': data_config,
                  'prior_config': prior_config,
                  'meta_train_config': meta_train_config,
                  'meta_test_config': meta_test_config,
                  'random_seed': None
                 }

        return config