import copy
import math
import time as time

import numpy as np
import tensorflow as tf
from hyperopt import hp

from pacoh import datasets as datasets, algorithms as algorithms
from .tuner import ModelTuner

SEED = 28

SAVE_FIGS = False

# FC NN CONFIG
LAYER_SIZE = 64
N_LAYERS = 4
ACTIVATION = 'relu'


class MAML_Classification_Tuner_NTASKS(ModelTuner):
    metric = 'avg_acc'
    mode = 'max'
    num_samples = 500 # number of runs for tune

    def __init__(self, args):
        self.args = args

    def init_run(self, config):

        """ --- pre-process config ---- """

        config = copy.deepcopy(config)
        data_config = config['data_config']

        # set activations
        """ --- setup datasets and samplers ---- """
        assert 'num_train_tasks' in config['data_config']
        data_config = config['data_config']

        from pacoh.datasets.classification_datasets import OmniglotMetaDataset
        from pacoh.datasets.data_sampler import MetaDatasetSampler
        dataset = OmniglotMetaDataset(n_classes=5, ignore_alphabet_boundaries=True,
                                      random_state=np.random.RandomState(28))

        data_train = dataset.generate_meta_train_data(
            n_tasks=data_config['num_train_tasks'],
            n_samples=-1)

        data_valid = dataset.generate_meta_test_data(
            n_tasks=40,
            n_samples_context=25,
            n_samples_test=-1
        )

        meta_train_sampler = MetaDatasetSampler(data_train, per_task_batch_size=data_config['batch_size'],
                                                meta_batch_size=data_config['meta_batch_size'],
                                                n_batched_models=None, tiled=False,
                                                flatten_data=False, standardization=False)

        meta_test_sampler = MetaDatasetSampler(data_valid, per_task_batch_size=25,
                                               meta_batch_size=-1, n_batched_models=None,
                                               tiled=False, flatten_data=False, standardization=False)

        meta_train_sampler.set_original_shape(dataset.original_shape)
        meta_test_sampler.set_original_shape(dataset.original_shape)

        del config['data_config']

        self.meta_model = algorithms.MAML(meta_train_sampler=meta_train_sampler,
                                          meta_test_sampler=meta_test_sampler, **config)

        self.n_optimization_steps = self.meta_model.meta_train_config['n_optimization_steps']
        self.n_iterations_per_step = self.meta_model.meta_train_config['n_iterations_per_step']
        self.n_total_iterations = self.meta_model.meta_train_config['n_total_iterations']

        self.log_period = self.meta_model.meta_train_config['log_period']
        self.n_iter = 0

    def run_iters(self, n_iter_per_step, reporter=None):
        t = time.time()
        n_iter = self.n_iter
        for _ in range(n_iter_per_step):
            n_iter += 1
            loss = self.meta_model.step()
            loss = tf.reduce_mean(loss).numpy()
            if n_iter % self.log_period == 0:
                message = '\nIter %d/%d - Time %.2f sec' % (n_iter, self.n_total_iterations, time.time() - t)
                message += ' - Train-Loss: %.3f' % loss
                print(message)
                t = time.time()
                if reporter is not None:
                    reporter(timesteps_total=n_iter, loss=loss, avg_acc=math.nan)

        self.n_iter = n_iter

        eval_metrics_mean, eval_metrics_std = self.meta_model.meta_eval(print_results=True)

        results = {'timesteps_total': self.n_iter, 'loss': loss}
        results.update(eval_metrics_mean)
        return results

    def get_search_space(self):
        # Slash to go a level deeper in the hierarchy
        # Colon to set multiple parameters to the same value
        if '2shot' in self.args['dataset']:
            max_batch_size = 2 * 5
        elif '5shot' in self.args['dataset']:
            max_batch_size = 2 * 5 * 5

        space = {

            # learning rates for meta-inference and inference
            "meta_train_config/learning_rate": hp.loguniform("meta_train_config/learning_rate", math.log(1e-4),
                                                           math.log(1e-3)),

            "meta_train_config/inner_learning_rate": hp.loguniform("meta_test_config/learning_rate", math.log(5e-4),
                                                           math.log(5e-1)),

            # batch sizes
            # "data_config/meta_batch_size": hp.choice("data_config/meta_batch_size", [2, 5, 10]),
            #
            # "data_config/batch_size": hp.choice("data_config/batch_size", [5, 10, max_batch_size]),
            "data_config/batch_size": hp.choice("data_config/batch_size", [25, max_batch_size]),

            "meta_train_config/first_order": hp.choice("meta_train_config/first_order", [True, False]),

            "meta_train_config/split_batch": hp.choice("meta_train_config/split_batch", [True, False]),

            # num inner steps
            "meta_train_config/n_inner_steps;meta_test_config/n_optimization_steps":
                hp.choice("meta_train_config/n_inner_steps;meta_test_config/n_optimization_steps", [1, 2, 3, 4, 5]),
        }

        return space


    def get_base_config(self):
        if '80' in self.args['dataset']:
            n_tasks = 80
        elif '160' in self.args['dataset']:
            n_tasks = 160
        elif '320' in self.args['dataset']:
            n_tasks = 320
        elif '640' in self.args['dataset']:
            n_tasks = 640
        elif '2000' in self.args['dataset']:
            n_tasks = 2000
        else:
            raise NotImplementedError

        batch_size = 25
        if 'omniglot' in self.args['dataset']:
            meta_batch_size = 5
            n_classes = 5
        else:
            raise NotImplementedError("Unknown classification dataset %s"%self.args['dataset'])

        likelihood_config = {'class': {'learn_likelihood_variables': False},
                             'type': 'class'
                             }


        model_config = {'fc': {'input_dim': None,
                               'output_dim': n_classes,
                               'layer_units': [LAYER_SIZE for _ in range(N_LAYERS)] + [n_classes],
                               'activations': [ACTIVATION for _ in range(N_LAYERS)] + ['sigmoid']
                               },
                        'cnn': {'activations': [ACTIVATION, 'softmax'],
                                'output_dim': 1 if n_classes == 2 else n_classes},
                        'type': 'cnn'}

        assert(len(model_config['fc']['layer_units']) == len(model_config['fc']['activations']))


        data_config = {'dataset': self.args['dataset'],
                       'num_train_tasks': n_tasks,
                       'tiled': False,
                       'batch_size': batch_size,
                       'posterior_inference_batch_size': -1,
                       'meta_batch_size': meta_batch_size,
                       }

        meta_train_config = {'n_total_iterations': 100000,
                             'n_train_loss_logs': 20,
                             'n_eval_runs': 0,
                             'optimizer': 'adam',
                             'learning_rate': 0.001,
                             'inner_learning_rate': 0.01,
                             'n_inner_steps': 1,
                             'split_batch': True,
                             'first_order': False
                             }

        meta_test_config = {'n_optimization_steps': 1}


        config = {'model_config': model_config,
                  'likelihood_config': likelihood_config,
                  'data_config': data_config,
                  'meta_train_config': meta_train_config,
                  'meta_test_config': meta_test_config,
                  'random_seed': None
                 }

        return config