import copy
import math
import time as time

import numpy as np
import tensorflow as tf
from hyperopt import hp

from pacoh.algorithms.Meta_BNN_MLAP import MetaBayesianNeuralNetwork_MLAP
from pacoh import datasets as datasets
from .tuner import ModelTuner

SEED = 28

SAVE_FIGS = False

# FC NN CONFIG
LAYER_SIZE = 32
N_LAYERS = 4
ACTIVATION = 'relu'



class MLAP_Regression_Tuner(ModelTuner):
    metric = 'avg_ll'
    mode = 'max'
    num_samples = 100 # number of runs for tune

    def __init__(self, args):
        self.args = args

    def init_run(self, config):

        """ --- pre-process config ---- """

        config = copy.deepcopy(config)
        data_config = config['data_config']

        posterior_config = config['posterior_config']
        config['posterior_config'] = posterior_config = posterior_config[posterior_config['type']]

        prior_config = copy.deepcopy(posterior_config)
        prior_config['n_batched_priors'] = 1
        config['prior_config'] = prior_config

        # set activations
        if 'activation_fn' in config['model_config']['fc']:
            activation_fn = config['model_config']['fc']['activation_fn']
            del config['model_config']['fc']['activation_fn']
            config['model_config']['fc']['activations'] = [activation_fn for _ in range(N_LAYERS)] + [None]

        """ --- setup datasets and samplers ---- """

        n_batched_models_train = posterior_config['n_samples']
        n_batched_models_test = posterior_config['n_samples']

        if data_config['type'] == 'reg':
            train_ds, val_ds = datasets.get_regression_datasets(data_config,
                                                                n_batched_models_train=n_batched_models_train,
                                                                n_batched_models_test=n_batched_models_test)
        elif data_config['type'] == 'class':
            train_ds, val_ds = datasets.get_classification_datasets(data_config,
                                                                    n_batched_models_train=n_batched_models_train,
                                                                    n_batched_models_test=n_batched_models_test)
        else:
            raise Exception("Unexpected dataset type")

        del config['data_config']

        n_batched_models_train *= train_ds.n_tasks
        config['posterior_config']['n_batched_priors'] = train_ds.n_tasks
        config['n_batched_models'] = n_batched_models_train

        self.meta_model = MetaBayesianNeuralNetwork_MLAP(meta_train_sampler=train_ds,
                                                         meta_test_sampler=val_ds, **config)

        self.n_optimization_steps = self.meta_model.meta_train_config['n_optimization_steps']
        self.n_iterations_per_step = self.meta_model.meta_train_config['n_iterations_per_step']
        self.n_total_iterations = self.meta_model.meta_train_config['n_total_iterations']

        self.log_period = self.meta_model.meta_train_config['log_period']
        self.n_iter = 0

    def run_iters(self, n_iter_per_step, reporter=None):
        t = time.time()
        n_iter = self.n_iter
        for _ in range(n_iter_per_step):
            n_iter += 1
            log_prob = self.meta_model.step() # step returns the per task log_likelihood or-mean squared error
            avg_log_prob = tf.reduce_mean(log_prob).numpy()
            if n_iter % self.log_period == 0:
                message = '\nIter %d/%d - Time %.2f sec' % (n_iter, self.n_total_iterations, time.time() - t)
                message += ' - Train-Log-Prob: %.3f' % avg_log_prob
                print(message)
                t = time.time()
                if reporter is not None:
                    reporter(timesteps_total=n_iter, train_obj=avg_log_prob, avg_ll=math.nan, avg_rmse=math.nan,
                             cal_err=math.nan)

        self.n_iter = n_iter
        # TODO: Kill if avg_log_prob is NaN
        eval_metrics_mean, eval_metrics_std = self.meta_model.meta_eval(print_results=True)

        results = {'timesteps_total': self.n_iter, 'train_obj': avg_log_prob}
        results.update(eval_metrics_mean)
        return results

    def get_search_space(self):
        # Slash to go a level deeper in the hierarchy
        # Colon to set multiple parameters to the same value
        batch_size_range = [-1, 5, 10]
        posterior_inference_batch_size_range = [-1, 5, 10]
        meta_batch_size_range = [5, 10]

        if self.args['dataset'] in ['physionet_0', 'physionet_2']:
            batch_size_range = batch_size_range[1:]
            posterior_inference_batch_size_range = posterior_inference_batch_size_range[1:]

        space = {
            # Prior and posterior variance
            # kappaPrior
            "hyper_prior_config/std": hp.loguniform("hyper_prior_config/std", math.log(1e2), math.log(1e4)),

            # kappaPosterior
            "hyper_posterior_config/std": hp.loguniform("hyper_posterior_config/std", math.log(1e-6),
                                                        math.log(1e-2)),

            # Task and meta complexity
            "meta_train_config/task_complexity_weight": hp.loguniform(
                    "meta_train_config/task_complexity_weight",
                    math.log(1e-3), math.log(1e1)),

            "meta_train_config/meta_complexity_weight": hp.loguniform(
                "meta_train_config/meta_complexity_weight",
                math.log(1e-3), math.log(1e0)),

            "meta_test_config/task_complexity_weight": hp.loguniform(
                "meta_test_config/task_complexity_weight",
                math.log(1e-3), math.log(1e1)),

            # Learning rates
            # base learning rates
            "meta_train_config/learning_rate": hp.loguniform("meta_train_config/learning_rate", math.log(5e-4),
                                                             math.log(1e-2)),

            "meta_test_config/learning_rate": hp.loguniform("meta_test_config/learning_rate", math.log(5e-4),
                                                            math.log(1e-2)),

            # prior parameters' learning rate scaling
            "meta_train_config/prior_lr_scale": hp.loguniform("meta_train_config/prior_lr_scale",
                                                            math.log(1e-1), math.log(1e0)),

            # variance parameters' learning rate scalling
            "meta_train_config/variance_lr_scale": hp.loguniform("meta_train_config/variance_lr_scale",
                                                            math.log(1e0), math.log(1e1))

            # STUFF THAT MAY BE INCLUDED IN THE HYPERPARAM SEARCH
            # activation function
            # "model_config/fc/activation_fn": hp.choice("model_config/fc/activation_fn", ['relu', 'elu', 'tanh']),
            # posterior_config
            # "posterior_config/n_samples": hp.choice("posterior_config/n_samples", [5, 10, 20]),
            # "data_config/meta_batch_size": hp.choice("data_config/meta_batch_size", meta_batch_size_range),
            # "data_config/posterior_inference_batch_size_range":
            #     hp.choice("data_config/posterior_inference_batch_size_range", posterior_inference_batch_size_range)
        }

        return space

    def get_base_config(self):
        likelihood_config = {'class': {'learn_likelihood_variables': True},
                             'reg': {'log_var': 2 * np.log(0.01),
                                     'learn_likelihood_variables': self.args['learn_likelihood_variables']
                                     },
                             'type': self.args['likelihood_type'],
                             'use_mse_as_loss': self.args['use_mse_as_loss']
                             }

        model_config = {'fc': {'input_dim': None,
                               'output_dim': 1,
                               'layer_units': [LAYER_SIZE for _ in range(N_LAYERS)] + [1],
                               'activations': [ACTIVATION for _ in range(N_LAYERS)] + [None],
                               },
                        'cnn': {},
                        'type': self.args['model_type']}

        assert(len(model_config['fc']['layer_units']) == len(model_config['fc']['activations']))

        n_total_iterations = 10000
        if 'physionet' in self.args['dataset']:
            batch_size = 6
        elif 'sin' in self.args['dataset']:
            batch_size = 5
        elif 'cauchy' in self.args['dataset']:
            batch_size = 5
        elif 'swissfel' in self.args['dataset']:
            batch_size = 20
        elif 'mhc' in self.args['dataset']:
            batch_size = 10
        elif 'berkeley' in self.args['dataset']:
            batch_size = 20
            n_total_iterations = 20000
        else:
            raise AssertionError("Unknown dataset")


        data_config = {'dataset': self.args['dataset'],
                       'type': self.args['dataset_type'],
                       'num_train_tasks': None,
                       'num_test_valid_tasks': None,
                       'num_train_samples': None,
                       'num_context_samples': None,
                       'num_test_valid_samples': None,
                       'standardization': True,
                       'tiled': True,
                       'batch_size': batch_size,
                       'meta_batch_size': -1,
                       'posterior_inference_batch_size': batch_size,
                       }

        # hyper-prior
        # std <=> kappa_Pscript
        hyper_prior_config = {'mean': 0.0,
                              'std': 4064.6619982881634}

        # std <=> kappa_Qscript
        hyper_posterior_config = {'mean_mu': 0.0,
                                  'log_var_mu': -6.0,
                                  'std': 0.0001444503653271869,
                                  'init_std': 0.01,
                                  }

        # prior
        model_gaussian_prior = {'mean_mu': 0.0,
                                'bias_log_var_mu': -5,
                                'weight_log_var_mu': -5
                                }

        likelihood_gaussian_prior = {'log_var_mu': 2 * np.log(0.01),
                                     'log_var_log_var': 2 * np.log(1.0)
                                     }

        posterior_config = {'gaussian': {'model': model_gaussian_prior,
                                     'likelihood': likelihood_gaussian_prior,
                                     'init_std': 0.1,
                                     'add_noise': True,
                                     'init_from_models': True,
                                     'n_samples': 5,  # number of samples from prior for estimating the mll
                                     },
                        'type': 'gaussian'
                        }


        meta_train_config = {'n_total_iterations': n_total_iterations,
                             'n_train_loss_logs': 20,
                             'n_eval_runs': 0,
                             'optimizer': 'adam',
                             'learning_rate': 1e-2,
                             'plot_functions': False,
                             'plot_pred_lines': False,
                             'plot_pred_std': False,

                             'variance_lr_scale': 5.0,
                             'prior_lr_scale': 1.0,
                             'task_complexity_weight': 0.26870006891471826,
                             'meta_complexity_weight': 0.08033980400235861,
                             'bound_type': self.args['bound_type'],
                             'delta': 0.005564546625796155,
                             }

        meta_test_config = {'n_total_iterations': 4000,
                            'n_train_loss_logs': 5,
                            'n_eval_runs': 0,
                            'optimizer': 'adam',
                            'learning_rate': 1e-3,
                            'plot_functions': False,
                            'plot_pred_lines': True,
                            'plot_pred_std': True,

                            'variance_lr_scale': 2.0,
                            'task_complexity_weight': 0.26870006891471826,
                            'bound_type': self.args['bound_type'],
                            'delta': 0.005564546625796155,
                            }

        config = {'model_config': model_config,
                  'likelihood_config': likelihood_config,
                  'data_config': data_config,
                  'hyper_prior_config': hyper_prior_config,
                  'hyper_posterior_config': hyper_posterior_config,
                  'posterior_config': posterior_config,
                  'meta_train_config': meta_train_config,
                  'meta_test_config': meta_test_config,
                  'random_seed': None
                  }

        return config


class MLAP_Seeger_MSE(MLAP_Regression_Tuner):
    def __init__(self, args):
        super().__init__(args)

    def get_search_space(self):
        space = super().get_search_space()
        space.update({'meta_train_config/delta;meta_test_config/delta':
                          hp.loguniform('meta_train_config/delta;meta_test_config/delta', math.log(1e-4), math.log(1e-1))})
        return space


class MLAP_McAllester_MSE(MLAP_Regression_Tuner):
    def __init__(self, args):
        super().__init__(args)

    def get_search_space(self):
        space = super().get_search_space()
        space.update({'meta_train_config/delta;meta_test_config/delta':
                      hp.loguniform('meta_train_config/delta;meta_test_config/delta', math.log(1e-4), math.log(1e-1))})
        return space


class MLAP_Variational_MSE(MLAP_Regression_Tuner):

    def __init__(self, args):
        super().__init__(args)

    def get_search_space(self):
        space = super().get_search_space()
        return space