import copy
import math
import time as time

import numpy as np
import tensorflow as tf
from hyperopt import hp

from pacoh.algorithms.Meta_BNN_MLAP import MetaBayesianNeuralNetwork_MLAP
from pacoh import datasets as datasets
from .tuner import ModelTuner

SEED = 28

SAVE_FIGS = False

# FC NN CONFIG
LAYER_SIZE = 32
N_LAYERS = 4
ACTIVATION = 'relu'

N_THREADS = 2

class NP_Regression_Tuner(ModelTuner):
    metric = 'avg_ll'
    mode = 'max'
    num_samples = 100 # number of runs for tune

    def __init__(self, args):
        self.args = args

    def init_run(self, config):
        import torch
        torch.set_num_threads(N_THREADS)

        """ --- pre-process config ---- """
        config = copy.deepcopy(config)
        data_config = config['data_config']
        model_config = config['model_config']
        meta_train_config = config['meta_train_config']

        """ --- setup datasets and samplers ---- """

        from pacoh.datasets.regression_datasets import provide_data
        meta_train_data, meta_test_data, _ = provide_data(dataset=data_config['dataset'])
        self.meta_test_data = meta_test_data

        from pacoh_gp.NPR_meta import NPRegressionMetaLearned
        self.meta_model = NPRegressionMetaLearned(meta_train_data,
                                                  context_split_ratio=meta_train_config['context_split_ratio'],
                                                  lr_params=meta_train_config['lr'],
                                                  num_iter_fit=meta_train_config['n_total_iterations'],
                                                  weight_decay=meta_train_config['weight_decay'],
                                                  r_dim=model_config['r_dim'],
                                                  z_dim=model_config['z_dim'],
                                                  h_dim=model_config['h_dim'],
                                                  task_batch_size=data_config['meta_batch_size'],
                                                  random_seed=config['random_seed'])

        self.n_total_iterations = meta_train_config['n_total_iterations']
        self.n_iterations_per_step = meta_train_config['n_total_iterations']
        self.n_optimization_steps = 1
        self.log_period = meta_train_config['log_period']

        self.n_iter = 0

    def run_iters(self, n_iter_per_step, reporter=None):
        t = time.time()
        n_iter = self.n_iter
        for _ in range(n_iter_per_step):
            n_iter += 1
            avg_loss = self.meta_model.meta_fit(log_period=10000, n_iter=1, verbose=False)
            if n_iter % self.log_period == 0:
                message = '\nIter %d/%d - Time %.2f sec' % (n_iter, self.n_total_iterations, time.time() - t)
                message += ' - Train-Loss: %.3f' % avg_loss
                print(message)
                t = time.time()
                if reporter is not None:
                    reporter(timesteps_total=n_iter, train_obj=avg_loss, avg_ll=math.nan, avg_rmse=math.nan,
                             cal_err=math.nan)

        self.n_iter = n_iter
        avg_ll, avg_rmse, cal_err = self.meta_model.eval_datasets(self.meta_test_data, flatten_y=False)

        results = {'timesteps_total': self.n_iter, 'train_obj': avg_loss}
        results.update({'avg_ll': avg_ll, 'avg_rmse': avg_rmse, 'cal_err': cal_err})
        return results

    def get_search_space(self):
        # Slash to go a level deeper in the hierarchy
        # Colon to set multiple parameters to the same value
        space = {
            # Prior and posterior variance
            # kappaPrior
            "model_config/r_dim;model_config/z_dim;model_config/h_dim":
                hp.choice("model_config/r_dim;model_config/z_dim;model_config/h_dim", [50, 100, 200]),

            "meta_train_config/weight_decay": hp.loguniform("meta_train_config/weight_decay", math.log(1e-4),
                                                             math.log(5e-1)),

            "meta_train_config/lr": hp.loguniform("meta_train_config/lr", math.log(5e-4),
                                                             math.log(1e-2)),
            }

        return space

    def get_base_config(self):

        data_config = {'dataset': self.args['dataset'],
                       'meta_batch_size': 5,
                       }

        model_config = {'r_dim': 50,
                        'z_dim': 50,
                        'h_dim': 50,}

        meta_train_config = {'n_total_iterations': 30000,
                             'log_period': 3000,
                             'context_split_ratio': 0.5,
                             'n_train_loss_logs': 20,
                             'weight_decay': 1e-2,
                             'lr': 1e-3,
                             }


        config = {'model_config': model_config,
                  'data_config': data_config,
                  'meta_train_config': meta_train_config,
                  'random_seed': None
                  }

        return config
