import copy
import math
import time as time

import numpy as np
import tensorflow as tf
from hyperopt import hp

from pacoh import datasets as datasets, algorithms as algorithms
from .tuner import ModelTuner

SEED = 28

SAVE_FIGS = False

# FC NN CONFIG
LAYER_SIZE = 32
N_LAYERS = 4
ACTIVATION = 'relu'



class PACOH_SVGD_Regression_Tuner(ModelTuner):
    metric = 'avg_ll'
    mode = 'max'
    num_samples = 200 # number of runs for tune

    def __init__(self, args):
        self.args = args

    def init_run(self, config):

        """ --- pre-process config ---- """

        config = copy.deepcopy(config)
        data_config = config['data_config']

        prior_config = config['prior_config']
        config['prior_config'] = prior_config = prior_config[prior_config['type']]

        config['posterior_config'].update(config['posterior_config'].pop(config['posterior_config']['type']))
        posterior_config = config['posterior_config']

        hyper_posterior_config = config['hyper_posterior_config']
        config['hyper_posterior_config'] = hyper_posterior_config = hyper_posterior_config[hyper_posterior_config['type']]

        # set activations
        if 'activation_fn' in config['model_config']['fc']:
            activation_fn = config['model_config']['fc']['activation_fn']
            del config['model_config']['fc']['activation_fn']
            config['model_config']['fc']['activations'] = [activation_fn for _ in range(N_LAYERS)] + [None]

        """ --- setup datasets and samplers ---- """

        n_batched_models_train = hyper_posterior_config['n_samples'] * prior_config['n_samples']
        n_batched_models_test = hyper_posterior_config['n_samples'] * posterior_config['n_samples']

        if data_config['type'] == 'reg':
            train_ds, val_ds = datasets.get_regression_datasets(data_config,
                                                                n_batched_models_train=n_batched_models_train,
                                                                n_batched_models_test=n_batched_models_test)
        elif data_config['type'] == 'class':
            train_ds, val_ds = datasets.get_classification_datasets(data_config,
                                                                    n_batched_models_train=n_batched_models_train,
                                                                    n_batched_models_test=n_batched_models_test)
        else:
            raise Exception("Unexpected dataset type")

        del config['data_config']

        self.meta_model = algorithms.MetaBayesianNeuralNetwork_SVGD(meta_train_sampler=train_ds,
                                                                    meta_test_sampler=val_ds, **config)

        self.n_optimization_steps = self.meta_model.meta_train_config['n_optimization_steps']
        self.n_iterations_per_step = self.meta_model.meta_train_config['n_iterations_per_step']
        self.n_total_iterations = self.meta_model.meta_train_config['n_total_iterations']

        self.log_period = self.meta_model.meta_train_config['log_period']
        self.n_iter = 0

    def run_iters(self, n_iter_per_step, reporter=None):
        t = time.time()
        n_iter = self.n_iter
        for _ in range(n_iter_per_step):
            n_iter += 1
            log_prob = self.meta_model.step()
            avg_log_prob = tf.reduce_mean(log_prob).numpy()
            if n_iter % self.log_period == 0:
                message = '\nIter %d/%d - Time %.2f sec' % (n_iter, self.n_total_iterations, time.time() - t)
                message += ' - Train-Avg-LL: %.3f' % avg_log_prob
                print(message)
                t = time.time()
                if reporter is not None:
                    reporter(timesteps_total=n_iter, train_obj=avg_log_prob, avg_ll=math.nan, avg_rmse=math.nan,
                             cal_err=math.nan)

        self.n_iter = n_iter

        eval_metrics_mean, eval_metrics_std = self.meta_model.meta_eval(print_results=True)

        results = {'timesteps_total': self.n_iter, 'train_obj': avg_log_prob}
        results.update(eval_metrics_mean)
        return results

    def get_search_space(self):
        # Slash to go a level deeper in the hierarchy
        # Colon to set multiple parameters to the same value
        batch_size_range = [-1, 5, 10]
        posterior_inference_batch_size_range = [-1, 5, 10]
        meta_batch_size_range = [5, 10]

        if self.args['dataset'] in ['physionet_0', 'physionet_2']:
            batch_size_range = batch_size_range[1:]
            posterior_inference_batch_size_range = posterior_inference_batch_size_range[1:]

        if self.args['inference_algo_type'] == 'svgd':
            hyper_prior_all_stds = 'hyper_prior_config/bias_mean_std;' \
                                   'hyper_prior_config/kernel_mean_std;' \
                                   'hyper_prior_config/bias_log_var_std;' \
                                   'hyper_prior_config/kernel_log_var_std'
            space = {
                # hyper-prior settings
                hyper_prior_all_stds: hp.loguniform(hyper_prior_all_stds, math.log(1e-2), math.log(1e2)), # TODO: math.log(1e-2), math.log(1e2)

                "hyper_prior_config/log_var_mean": hp.uniform("hyper_prior_config/log_var_mean", -5, 0),

                # prior prob weights for meta-inference and inference
                "meta_train_config/prior_prob_weight": hp.loguniform("meta_train_config/prior_prob_weight",
                                                                     math.log(1e-6),
                                                                     math.log(1e0)),

                "meta_test_config/prior_prob_weight": hp.loguniform("meta_test_config/prior_prob_weight",
                                                                    math.log(1e-6),
                                                                    math.log(1e0)),

                # learning rates for meta-inference and inference
                "meta_train_config/learning_rate": hp.loguniform("meta_train_config/learning_rate", math.log(5e-4),
                                                                 math.log(5e-3)),

                "meta_test_config/learning_rate": hp.loguniform("meta_test_config/learning_rate", math.log(5e-4),
                                                                math.log(5e-3)),

                # bandwidths
                "meta_train_config/bandwidth;meta_test_config/bandwidth":
                    hp.loguniform("meta_train_config/bandwidth;meta_test_config/bandwidth", math.log(1e-1), math.log(1e3)),

                # activation function
                "model_config/fc/activation_fn": hp.choice("model_config/fc/activation_fn", ['relu', 'elu', 'tanh']),

                # STUFF THAT MAY BE INCLUDED IN THE HYPERPARAM SEARCH
                # posterior_config
                # "posterior_config/n_samples": hp.choice("posterior_config/n_samples", [5, 10, 20]),
                # "data_config/meta_batch_size": hp.choice("data_config/meta_batch_size", meta_batch_size_range),
                # "data_config/posterior_inference_batch_size_range":
                #     hp.choice("data_config/posterior_inference_batch_size_range", posterior_inference_batch_size_range)
            }

            if not self.args['learn_likelihood_variables']:
                space["likelihood_config/reg/log_var"] = hp.loguniform("likelihood_config/reg/log_var", np.log(0.01), np.log(0.1)),

            return space
        else:
            raise NotImplementedError

    def get_base_config(self):
        likelihood_config = {'class': {'learn_likelihood_variables': True},
                             'reg': {'log_var': 2 * np.log(0.01),
                                     'learn_likelihood_variables': self.args['learn_likelihood_variables']},
                             'type': self.args['likelihood_type']
                             }

        model_config = {'fc': {'input_dim': None,
                               'output_dim': 1,
                               'layer_units': [LAYER_SIZE for _ in range(N_LAYERS)] + [1],
                               'activations': [ACTIVATION for _ in range(N_LAYERS)] + [None],
                               },
                        'cnn': {},
                        'type': self.args['model_type']}

        assert(len(model_config['fc']['layer_units']) == len(model_config['fc']['activations']))

        n_total_iterations = 40000

        if 'physionet' in self.args['dataset']:
            batch_size = 6
        elif 'sin' in self.args['dataset']:
            batch_size = 5
        elif 'cauchy' in self.args['dataset']:
            batch_size = 5
        elif 'swissfel' in self.args['dataset']:
            batch_size = 20
        elif 'mhc' in self.args['dataset']:
            batch_size = 10
            n_total_iterations = 80000
        elif 'berkeley' in self.args['dataset']:
            n_total_iterations = 60000
            batch_size = 20
        else:
            raise AssertionError("Unknown dataset")


        data_config = {'dataset': self.args['dataset'],
                       'type': self.args['dataset_type'],
                       'num_train_tasks': None,
                       'num_test_valid_tasks': None,
                       'num_train_samples': None,
                       'num_context_samples': None,
                       'num_test_valid_samples': None,
                       'standardization': True,
                       'tiled': True,
                       'batch_size': batch_size,
                       'meta_batch_size': 5,
                       'posterior_inference_batch_size': batch_size,
                       }

        # hyper-prior

        hyper_prior_config = {'mean_mean': 0.0,  # mean of the hyper-prior for the means
                              'bias_mean_std': 3,  # std of the hyper-prior for the mean of the biases # 1.3692749870080771
                              'kernel_mean_std': 1,  # std of the hyper-prior for the mean of the kernels # 1.3692749870080771
                              'log_var_mean': -1,  # mean of the hyper-prior for the log-variances of the prior # -2.5474841072521235
                              'bias_log_var_std': 1,  # std of the hyper-prior for the log-variances of the prior # 1.3692749870080771
                              'kernel_log_var_std': 1,  # std of the hyper-prior for the log-variances of the prior # 1.3692749870080771
                              'likelihood_log_var_mean_mean': -4,
                              'likelihood_log_var_mean_std': 1,
                              'likelihood_log_var_log_var_mean': -2,
                              'likelihood_log_var_log_var_std': 0.2
                              }

        # hyper-posterior
        if self.args.get('MAP', False):
            n_particles_hyper_posterior = 1
        else:
            n_particles_hyper_posterior = 5

        # hyper-posterior
        hyper_posterior_config = {'particle': {'n_samples': n_particles_hyper_posterior,
                                               },
                                  'type': 'particle'
                                  }
        # prior
        model_gaussian_prior = {'mean_mu': 0.0,
                                'bias_log_var_mu': -5,  # init mean for the log_var of the bias
                                'weight_log_var_mu': -5
                                }

        likelihood_gaussian_prior = {'log_var_mu': 2 * np.log(0.01),
                                     'log_var_log_var': 2 * np.log(1.0)
                                     }

        prior_config = {'gaussian': {'model': model_gaussian_prior,
                                     'likelihood': likelihood_gaussian_prior,
                                     'init_std': 0.1,
                                     'add_noise': True, # Whether to add noise to the trainable variables of the prior module
                                     'init_from_models': True, # Whether to use a sample of models to initialize prior variables (not implemented for likelihood)
                                     'n_samples': 10,  # number of samples from prior for estimating the mll
                                     },
                        'type': 'gaussian'
                        }

        # posterior
        posterior_config = {'particle': {'n_samples': 5,
                                         'use_init_std': False
                                         },
                            'type': 'particle'
                            }

        meta_train_config = {'n_total_iterations': n_total_iterations,
                             'n_train_loss_logs': 20,
                             'n_eval_runs': 0,
                             'optimizer': 'adam',
                             'learning_rate': 1e-3,
                             'plot_functions': False,
                             'plot_pred_lines': False,
                             'logsumexp_estimate': self.args['logsumexp'],
                             'plot_pred_std': False,
                             'prior_prob_weight': 0.1,
                             'bandwidth': 10
                             }

        meta_test_config = {'n_total_iterations': 3000,
                            'n_train_loss_logs': 5,
                            'n_eval_runs': 0,
                            'optimizer': 'adam',
                            'learning_rate': 1e-3,
                            'plot_functions': False,
                            'plot_pred_lines': False,
                            'plot_pred_std': False,
                            'prior_prob_weight': 0.1,
                            'bandwidth': 10  # if this goes to zero repulsive term goes to zero
                            }

        config = {'model_config': model_config,
                  'likelihood_config': likelihood_config,
                  'data_config': data_config,
                  'hyper_prior_config': hyper_prior_config,
                  'hyper_posterior_config': hyper_posterior_config,
                  'prior_config': prior_config,
                  'posterior_config': posterior_config,
                  'meta_train_config': meta_train_config,
                  'meta_test_config': meta_test_config,
                  'random_seed': None
                  }

        return config


class PACOH_SVGD_Regression_Tuner_FixedLikelihood(PACOH_SVGD_Regression_Tuner):

    def __init__(self, args):
        super().__init__(args)

    def get_search_space(self):
        space = super().get_search_space()
        space.update({'likelihood_config/reg/log_var':
                          hp.uniform('likelihood_config/reg/log_var', 2 * np.log(0.005), 2 * np.log(0.1))})
        return space


class PACOH_SVGD_Regression_Tuner_LearnableLikelihood(PACOH_SVGD_Regression_Tuner):

    def __init__(self, args):
        super().__init__(args)

    def get_search_space(self):
        space = super().get_search_space()
        space.update({'hyper_prior_config/likelihood_log_var_mean_mean':
                          hp.uniform('hyper_prior_config/likelihood_log_var_mean_mean', 2 * np.log(0.05),
                                     2 * np.log(0.1))})
        return space
