import pandas as pd
import numpy as np
import os

def select_best_configs(analysis, metric, mode='max', N=5):
    try:
        rows = analysis._retrieve_rows(metric=metric, mode=mode)
    except:
        # Get header
        missing_header = []
        for path, df in analysis.trial_dataframes.items():
            try:
                df[metric]
                header = df.columns
            except:
                missing_header.append(path)

        # Read problematic results again
        for path in missing_header:
            df = pd.read_csv(os.path.join(path, 'progress.csv'), header=None)
            df.columns = header
            analysis.trial_dataframes[path] = df

        rows = {}
        for path, df in analysis.trial_dataframes.items():
            if mode == "max":
                idx = df[metric].idxmax()
            elif mode == "min":
                idx = df[metric].idxmin()
            else:
                idx = -1
            if np.isnan(idx):
                print('NaN value in experiment: ', path)
                continue
            rows[path] = df.iloc[idx].to_dict()

    all_configs = analysis.get_all_configs()
    reverse = mode == 'max'
    best_paths = sorted(rows, key=lambda k: rows[k][metric], reverse=reverse)[:N]
    best_configs = [all_configs[path] for path in best_paths]
    return best_configs

def process_parameter(config, pn, value):
    parts = pn.split('/')
    loc = parts[:-1]
    key = parts[-1]

    d = config
    for i in loc:
        if i == '':
            continue

        d = d[i]

    d[key] = value


def process_config(config):
    d = config.copy()

    for pn in d:
        # Check if this is a parameter to set
        if '/' in pn:
            parts = pn.split(";")
            value = config[pn]

            if len(parts) > 1:
                # more than one parameter needs to be initialized with this value
                for part in parts:
                    process_parameter(config, part, value)
            else:
                process_parameter(config, pn, value)

            del config[pn]

    return config

""" --- helper functions --- """

def _aggregate_eval_metrics_across_tasks(eval_metrics_dict_per_task):
    eval_metrics_mean, eval_metrics_std = {}, {}
    metric_keys = set.union(*[set(metrics_dict.keys()) for metrics_dict in eval_metrics_dict_per_task])
    for key in metric_keys:
        metric_list = []
        for eval_metrics_dict in eval_metrics_dict_per_task:
            metric_list.extend(eval_metrics_dict[key])
        eval_metrics_mean[key] = np.mean(metric_list)
        eval_metrics_std[key] = np.std(metric_list)
    return eval_metrics_mean, eval_metrics_std

def _split_into_batches(list, max_batch_size):
    import math
    n_elements = len(list)
    if max_batch_size == -1:
        max_batch_size = n_elements
    n_batches = math.ceil(n_elements / float(max_batch_size))
    remainder = n_elements % n_batches
    batches = []
    idx = 0
    for i in range(n_batches):
        if i < remainder:
            batch_size = n_elements // n_batches + 1
        else:
            batch_size = n_elements // n_batches
        batches.append(list[idx:idx + batch_size])
        idx += batch_size

    assert all([len(batch) <= max_batch_size for batch in batches])
    assert np.sum([len(batch) for batch in batches]) == n_elements
    return batches


