import tensorflow as tf
import numpy as np
import copy

import pacoh.datasets as datasets
import pacoh.algorithms as algorithms


N_CLASSES = 5  # K-way
DATASET = 'omniglot_5shot_small'

likelihood_config = {'class': {'learn_likelihood_variables': False},
                     'type': 'class'
                     }

LAYER_SIZE = 120
N_LAYERS = 3
ACTIVATION = 'relu'

model_config = {'fc': {'input_dim': None,
                       'output_dim': 1 if N_CLASSES == 2 else N_CLASSES,
                       'layer_units': [LAYER_SIZE for _ in range(N_LAYERS)] + [1],
                       'activations': [ACTIVATION for _ in range(N_LAYERS)] + ['softmax']
                       },
                'cnn': {'activations': ['elu', 'sigmoid' if N_CLASSES == 2 else 'softmax'],
                        'output_dim': 1 if N_CLASSES == 2 else N_CLASSES},
                'type': 'cnn'}

assert (len(model_config['fc']['layer_units']) == len(model_config['fc']['activations']))

data_config = {'dataset': DATASET,
               'n_classes': N_CLASSES,
               'num_train_tasks': None,
               'num_test_valid_tasks': None,
               'num_train_samples': None,
               'num_context_samples': None,
               'num_test_valid_samples': None,
               'standardization': False,
               'tiled': True,
               'batch_size': 10,
               'posterior_inference_batch_size': 10,
               'meta_batch_size': -1
               }

model_gaussian_prior = {'mean_mu': 0.0,
                        'bias_log_var_mu': -5, #2 * np.log(1.0), # mean for the log_var of the bias
                        'weight_log_var_mu': -5 #2 * np.log(0.5)
                        }
# hyper-prior
# std <=> kappa_Pscript
hyper_prior_config = {'mean': 0.0,
                      'std': 4064.6619982881634}

# std <=> kappa_Qscript
hyper_posterior_config = {'mean_mu': 0.0,
                          'log_var_mu': -6.0,
                          'std': 0.0001444503653271869,
                          'init_std': 0.01,
                          }

# prior
model_gaussian_prior = {'mean_mu': 0.0,
                        'bias_log_var_mu': -5,
                        'weight_log_var_mu': -5
                        }

likelihood_gaussian_prior = {'log_var_mu': 2 * np.log(0.01),
                             'log_var_log_var': 2 * np.log(1.0)
                             }

posterior_config = {'gaussian': {'model': model_gaussian_prior,
                                 'likelihood': likelihood_gaussian_prior,
                                 'init_std': 0.1,
                                 'add_noise': True,
                                 'init_from_models': True,
                                 'n_samples': 3,  # number of samples from prior for estimating the mll
                                 },
                    'type': 'gaussian'
                    }

meta_train_config = {'n_total_iterations': 20,
                     'n_train_loss_logs': 5,
                     'n_eval_runs': 0,
                     'optimizer': 'adam',
                     'learning_rate': 1e-2,
                     'plot_functions': False,
                     'plot_pred_lines': False,
                     'plot_pred_std': False,

                     'variance_lr_scale': 5.0,
                     'prior_lr_scale': 1.0,
                     'task_complexity_weight': 0.26870006891471826,
                     'meta_complexity_weight': 0.08033980400235861,
                     'bound_type': 'variational',
                     'delta': 0.005564546625796155,
                     }

meta_test_config = {'n_total_iterations': 20,
                    'n_train_loss_logs': 5,
                    'n_eval_runs': 0,
                    'optimizer': 'adam',
                    'learning_rate': 1e-3,
                    'plot_functions': False,
                    'plot_pred_lines': True,
                    'plot_pred_std': True,

                    'variance_lr_scale': 2.0,
                    'task_complexity_weight': 0.26870006891471826,
                    'bound_type': 'variational',
                    'delta': 0.005564546625796155,
                    }

config = {'model_config': model_config,
          'likelihood_config': likelihood_config,
          'data_config': data_config,
          'hyper_prior_config': hyper_prior_config,
          'hyper_posterior_config': hyper_posterior_config,
          'posterior_config': posterior_config,
          'meta_train_config': meta_train_config,
          'meta_test_config': meta_test_config,
          'random_seed': None
          }

# hyper_posterior_config = hyper_posterior_config[hyper_posterior_config['type']]

def main():
    # tf.config.experimental_run_functions_eagerly(True)

    # setup data set and corresponding samplers

    data_config = config['data_config']

    posterior_config = config['posterior_config']
    config['posterior_config'] = posterior_config = posterior_config[posterior_config['type']]

    prior_config = copy.deepcopy(posterior_config)
    prior_config['n_batched_priors'] = 1
    config['prior_config'] = prior_config

    n_batched_models_train = posterior_config['n_samples']
    n_batched_models_test = posterior_config['n_samples']

    train_ds, val_ds = datasets.get_classification_datasets(data_config, n_batched_models_train=n_batched_models_train,
                                                            n_batched_models_test=n_batched_models_test)

    tf.print("data samplers ready")
    del config['data_config']
    n_batched_models_train *= train_ds.n_tasks
    config['posterior_config']['n_batched_priors'] = train_ds.n_tasks
    config['n_batched_models'] = n_batched_models_train

    meta_model = algorithms.MetaBayesianNeuralNetwork_MLAP(meta_train_sampler=train_ds,
                                                     meta_test_sampler=val_ds, **config)

    tf.print('model created')

    meta_model.meta_fit()

    meta_model.meta_eval(print_results=True)

if __name__ == '__main__':
    main()