import tensorflow as tf
import numpy as np

import pacoh.datasets as datasets
import pacoh.algorithms as algorithms

N_CLASSES = 10  # K-way
DATASET = 'fmnist'

likelihood_config = {'class': {'learn_likelihood_variables': False},
                     'type': 'class'
                     }

model_config = {'fc': {'input_dim': None,
                       'output_dim': 1 if N_CLASSES == 2 else N_CLASSES,
                       'layer_units': [120, 120, 120, 1 if N_CLASSES == 2 else N_CLASSES],
                        'activations': ['relu', 'relu', 'relu', 'softmax']
                       },
                'cnn': {'activations': ['relu', 'sigmoid' if N_CLASSES == 2 else 'softmax'],
                        'output_dim': 1 if N_CLASSES == 2 else N_CLASSES},
                'type': 'cnn'}

data_config = {'dataset': DATASET,
                'n_classes': N_CLASSES,
                'num_train_tasks': 2,
                'num_test_valid_tasks': 2,
                'num_train_samples': 2000,
                'num_context_samples': 100,
                'num_test_valid_samples': 1000,
                'tiled': True,
                'batch_size': 20,
                'posterior_inference_batch_size': 20,
                'meta_batch_size': 5
               }

model_gaussian_prior = {'mean_mu': 0.0,
                        'bias_log_var_mu': -5, #2 * np.log(1.0), # mean for the log_var of the bias
                        'weight_log_var_mu': -5 #2 * np.log(0.5)
                        }

hyper_prior_config = {'mean_mean': 0.0,
                     'bias_mean_std': 3,
                     'kernel_mean_std': 1,
                     'log_var_mean': -1,
                     'bias_log_var_std': 1,
                     'kernel_log_var_std': 1,
                     'likelihood_log_var_mean_mean': -1,
                     'likelihood_log_var_mean_std': 1,
                     'likelihood_log_var_log_var_mean': -2,
                     'likelihood_log_var_log_var_std': 0.1
                     }



prior_config = {'gaussian': {'model': model_gaussian_prior,
                             'likelihood': {},
                             'init_std': 0.02,
                             'add_noise': True,
                             'n_samples': 5   # number of samples from prior to estimate the mll
                             },
                'type': 'gaussian'
                }

hyper_posterior_config = {'particle': {'n_samples': 3,
                                 'use_init_std': True
                                 },
                    'type': 'particle'
                    }

posterior_config = {'particle': {'n_samples': 3,
                                 'use_init_std': True
                                 },
                    'type': 'particle'
                    }

meta_train_config = {'epochs': 100,
                   'n_train_loss_logs': 50,
                   'n_eval_runs': 5,
                   'optimizer': 'adam',
                   'learning_rate': 0.002,
                    'logsumexp_estimate': True,
                   'plot_functions': False,
                   'plot_pred_lines': True,
                   'plot_pred_std': True,
                   'weight_decay': 0.0,
                   'prior_prob_weight': 0.001,
                   'bandwidth': 10
                     }

meta_test_config = {'epochs': 100,
                             'n_train_loss_logs': 10,
                             'n_eval_runs': 100,
                             'optimizer': 'adam',
                             'learning_rate': 0.01,
                             'plot_functions': False,
                             'plot_pred_lines': True,
                             'plot_pred_std': True,
                             'weight_decay': 1e-5,
                             'prior_prob_weight': 1 / 10000,
                    'bandwidth': 10
                    }

prior_config = prior_config[prior_config['type']]
posterior_config = posterior_config[posterior_config['type']]
hyper_posterior_config = hyper_posterior_config[hyper_posterior_config['type']]

def main():
    #tf.config.experimental_run_functions_eagerly(True)

    # setup data set and corresponding samplers
    n_batched_models_train = hyper_posterior_config['n_samples'] * prior_config['n_samples']
    n_batched_models_test = hyper_posterior_config['n_samples'] * posterior_config['n_samples']
    train_ds, val_ds = datasets.get_classification_datasets(data_config, n_batched_models_train=n_batched_models_train,
                                                            n_batched_models_test=n_batched_models_test)

    tf.print("data samplers ready")
    meta_reg_model = algorithms.MetaBayesianNeuralNetwork_SVGD(train_ds, val_ds, model_config, likelihood_config,
                                                               prior_config, posterior_config,
                                                               hyper_prior_config, hyper_posterior_config,
                                                               meta_train_config, meta_test_config)

    tf.print('model created')

    meta_reg_model.meta_fit()

    meta_reg_model.meta_eval(print_results=True)

if __name__ == '__main__':
    main()