import tensorflow as tf
import numpy as np

import pacoh.datasets as datasets
import pacoh.algorithms as algorithms

N_CLASSES = 5  # K-way
DATASET = 'omniglot_5shot_full'

likelihood_config = {'class': {'learn_likelihood_variables': False},
                     'type': 'class'
                     }

LAYER_SIZE = 120
N_LAYERS = 3
ACTIVATION = 'relu'

model_config = {'fc': {'input_dim': None,
                       'output_dim': 1 if N_CLASSES == 2 else N_CLASSES,
                       'layer_units': [LAYER_SIZE for _ in range(N_LAYERS)] + [1],
                       'activations': [ACTIVATION for _ in range(N_LAYERS)] + ['softmax']
                       },
                'cnn': {'activations': ['elu', 'sigmoid' if N_CLASSES == 2 else 'softmax'],
                        'output_dim': 1 if N_CLASSES == 2 else N_CLASSES},
                'type': 'cnn'}

data_config = {'dataset': DATASET,
               'n_classes': N_CLASSES,
               'num_train_tasks': None,
               'num_test_valid_tasks': None,
               'num_train_samples': None,
               'num_context_samples': None,
               'num_test_valid_samples': None,
               'standardization': False,
               'tiled': True,
               'batch_size': 10,
               'posterior_inference_batch_size': 10,
               'meta_batch_size': 5
               }

model_gaussian_prior = {'mean_mu': 0.0,
                        'bias_log_var_mu': -5, #2 * np.log(1.0), # mean for the log_var of the bias
                        'weight_log_var_mu': -5 #2 * np.log(0.5)
                        }

hyper_prior_config = {'mean_mean': 0.0,
                     'bias_mean_std': 3,
                     'kernel_mean_std': 1,
                     'log_var_mean': -1,
                     'bias_log_var_std': 1,
                     'kernel_log_var_std': 1,
                     }



prior_config = {'gaussian': {'model': model_gaussian_prior,
                             'likelihood': {},
                             'init_std': 0.05,
                             'add_noise': True,
                             'init_from_models': True,
                             'n_samples': 5   # number of samples from prior to estimate the mll
                             },
                'type': 'gaussian'
                }

hyper_posterior_config = {'particle': {'n_samples': 3,
                                 'use_init_std': True
                                 },
                    'type': 'particle'
                    }

posterior_config = {'particle': {'n_samples': 3,
                                 'use_init_std': True
                                 },
                    'type': 'particle'
                    }

training_config = {'epochs': 2000,
                   'n_train_loss_logs': 100,
                   'n_eval_runs': 10,
                   'optimizer': 'adam',
                   'learning_rate': 0.001,
                   'plot_functions': False,
                   'plot_pred_lines': True,
                   'plot_pred_std': True,
                   'weight_decay': 0.0,
                   'prior_prob_weight': 0.01,
                   'bandwidth': 10
                   }

inference_training_config = {'epochs': 1000,
                             'n_train_loss_logs': 10,
                             'n_eval_runs': 0,
                             'optimizer': 'adam',
                             'learning_rate': 0.001,
                             'plot_functions': False,
                             'plot_pred_lines': False,
                             'plot_pred_std': False,
                             'prior_prob_weight': 1 / 1000,
                             'bandwidth': 10,
                             }

prior_config = prior_config[prior_config['type']]
posterior_config = posterior_config[posterior_config['type']]
hyper_posterior_config = hyper_posterior_config[hyper_posterior_config['type']]

def main():
    #tf.config.experimental_run_functions_eagerly(True)

    # setup data set and corresponding samplers
    n_batched_models_train = hyper_posterior_config['n_samples'] * prior_config['n_samples']
    n_batched_models_test = hyper_posterior_config['n_samples'] * posterior_config['n_samples']
    train_ds, val_ds = datasets.get_classification_datasets(data_config, n_batched_models_train=n_batched_models_train,
                                                            n_batched_models_test=n_batched_models_test)

    tf.print("data samplers ready")
    meta_reg_model = algorithms.MetaBayesianNeuralNetwork_SVGD(train_ds, val_ds, model_config, likelihood_config,
                                                               prior_config, posterior_config,
                                                               hyper_prior_config, hyper_posterior_config,
                                                               training_config, inference_training_config)

    tf.print('model created')

    meta_reg_model.meta_fit()

    meta_reg_model.meta_eval(print_results=True)

if __name__ == '__main__':
    main()