import tensorflow as tf
import tensorflow_probability as tfp

from .batched_module import TFModuleBatched


class GaussianHyperPrior(tf.Module):
    def __init__(self, batched_prior_module, config, name):
        super().__init__(name=name)
        self.config = config
        self.prior_module_names = [prior.name for prior in batched_prior_module.priors]
        self.n_batched_priors = len(self.prior_module_names)

        prior_module = batched_prior_module.priors[0]
        variables = prior_module.variables

        self.batched_variable_sizes = batched_prior_module.variable_sizes()
        self.batched_variable_names = [v.name for v in batched_prior_module.variables]

        self.base_variable_sizes = prior_module.base_variable_sizes
        self.base_variable_names = prior_module.base_variable_names

        for var in variables:
            self.process_variable(var)

    def process_variable(self, var):
        if 'model_parameters' in var.name:
            sizes = self.base_variable_sizes[0]
            names = self.base_variable_names[0]

            if 'mean' in var.name:
                # prior for model_parameters_mean
                means = []
                stds = []

                for v_size, v_name in zip(sizes, names):
                    mean = tf.ones((1, v_size), tf.float32) * self.config['mean_mean']

                    if 'bias' in v_name:
                        std = tf.ones((1, v_size), tf.float32) * self.config['bias_mean_std']
                    elif 'kernel' in v_name:
                        std = tf.ones((1, v_size), tf.float32) * self.config['kernel_mean_std']
                    else:
                        raise Exception("Unexpected parameter")

                    means.append(mean)
                    stds.append(std)

                means = tf.concat(means, axis=1)
                means = tf.cast(tf.squeeze(means), tf.float32)
                stds = tf.concat(stds, axis=1)
                stds = tf.cast(tf.squeeze(stds), tf.float32)

                if tf.rank(means) == 0:
                    means = tf.expand_dims(means, 0)

                dist = tfp.distributions.Independent(
                    tfp.distributions.Normal(means, stds),
                    reinterpreted_batch_ndims=1)

                @tf.function
                def log_prob(parameters):
                    return dist.log_prob(parameters)

            elif 'log_var' in var.name:
                # prior for model_parameters_mean
                means = []
                stds = []

                for v_size, v_name in zip(sizes, names):
                    mean = tf.ones((1, v_size), tf.float32) * self.config['log_var_mean']

                    if 'bias' in v_name:
                        std = tf.ones((1, v_size), tf.float32) * self.config['bias_log_var_std']
                    elif 'kernel' in v_name:
                        std = tf.ones((1, v_size), tf.float32) * self.config['kernel_log_var_std']
                    else:
                        raise Exception("Unexpected parameter")

                    means.append(mean)
                    stds.append(std)

                means = tf.concat(means, axis=1)
                means = tf.cast(tf.squeeze(means), tf.float32)
                stds = tf.concat(stds, axis=1)
                stds = tf.cast(tf.squeeze(stds), tf.float32)

                if tf.rank(means) == 0:
                    means = tf.expand_dims(means, 0)

                dist = tfp.distributions.Independent(
                    tfp.distributions.Normal(means, stds),
                    reinterpreted_batch_ndims=1)

                @tf.function
                def log_prob(parameters):
                    return dist.log_prob(parameters)
            else:
                raise Exception("Unexpected variable name")

        elif 'likelihood_parameters' in var.name:
            sizes = self.base_variable_sizes[1]

            if 'mean' in var.name:
                # prior for likelihood_parameters_mean
                means = tf.ones(sizes[0], tf.float32) * self.config['likelihood_log_var_mean_mean']
                stds = tf.ones(sizes[0], tf.float32) * self.config['likelihood_log_var_mean_std']

                means = tf.concat(means, axis=1)
                means = tf.cast(tf.squeeze(means), tf.float32)
                stds = tf.concat(stds, axis=1)
                stds = tf.cast(tf.squeeze(stds), tf.float32)

                if tf.rank(means) == 0:
                    means = tf.expand_dims(means, 0)

                dist = tfp.distributions.Independent(
                    tfp.distributions.Normal(means, stds),
                    reinterpreted_batch_ndims=1)

                @tf.function
                def log_prob(parameters):
                    return dist.log_prob(parameters)

            elif 'log_var' in var.name:
                # prior for likelihood_parameters_log_var
                means = tf.ones(sizes[0], tf.float32) * self.config['likelihood_log_var_log_var_mean']
                stds = tf.ones(sizes[0], tf.float32) * self.config['likelihood_log_var_log_var_std']
                means = tf.concat(means, axis=1)
                means = tf.cast(tf.squeeze(means), tf.float32)
                stds = tf.concat(stds, axis=1)
                stds = tf.cast(tf.squeeze(stds), tf.float32)

                if tf.rank(means) == 0:
                    means = tf.expand_dims(means, 0)

                dist = tfp.distributions.Independent(
                    tfp.distributions.Normal(means, stds),
                    reinterpreted_batch_ndims=1)

                @tf.function
                def log_prob(parameters):
                    return dist.log_prob(parameters)

            else:
                raise Exception("Unexpected variable name")
        else:
            raise Exception("Unexpeted variable name")

        suffix = var.name.split('/')[1]
        for prior_module_name in self.prior_module_names:
            name = f'{prior_module_name}/{suffix}'
            setattr(self, name, log_prob)

    @tf.function
    def log_prob_vectorized(self, params_vectorized):
        tf.assert_equal(params_vectorized.shape[1], self.n_batched_priors)
        parameters = tf.reshape(params_vectorized, (-1,))
        param_split = tf.split(parameters, self.batched_variable_sizes)
        log_probs = tf.reshape([getattr(self, v_name)(v) for v, v_name in  zip(param_split, self.batched_variable_names)],
                               (self.n_batched_priors, -1))
        log_probs = tf.reduce_sum(log_probs, axis=-1)
        return log_probs

    @tf.function
    def log_prob(self, variables):
        log_probs = tf.reshape([getattr(self, v.name)(v) for v in variables], (self.n_batched_priors, -1))
        log_probs = tf.reduce_sum(log_probs, axis=-1, keepdims=True)
        return log_probs


class ParticleHyperPosterior(TFModuleBatched):
    def __init__(self, config, name, prior=None, init_particles=None):
        super().__init__(name=name)
        assert not (prior is not None and init_particles is not None)

        self.config = config
        if prior is not None:
            sample = prior.sample(config['n_samples'], config['use_init_std'])
            self.particles = tf.Variable(sample, dtype=tf.float32)  # [#priors, #samples_per_prior, #params]
        elif init_particles is not None:
            self.particles = tf.Variable(init_particles, dtype=tf.float32)
        else:
            raise AssertionError("Either prior or init_particles must be specified")

    def sample(self, **kwargs):
        return self.particles

    def log_prob(**kwargs):
        raise Exception("Not implemented for particle dist")
