import config
import tensorflow as tf
import copy

import pacoh.modules as modules

from pacoh import util as util
from .random_predictor import _check_model_likelihood_compatibility

class BaseRandomisedPredictor:
    def __init__(self, model_config, likelihood_config, n_batched_models, batch_sample):
        # Set variables
        self.model_config = model_config
        self.likelihood_config = likelihood_config
        self.n_batched_models = n_batched_models

        # make sure that the likelihood is consistent with the NN setup
        _check_model_likelihood_compatibility(model_config, likelihood_config)

        # setup the NN model and likelihood
        self.model = modules.BatchedNNsModel(model_config=self.model_config,
                                             n_batched_models=self.n_batched_models)
        self.model.initialize_variables(batch_sample[0], batch_sample[1])

        self.likelihood = util.get_likelihood(config=likelihood_config,
                                              output_dim=self.model_config['output_dim'],
                                              n_batched_models=self.n_batched_models)

        # check whether likelihood has trainable variables
        self.model_variables_only = not len(self.likelihood.trainable_variables) > 0
        self.split_sizes = None

    def predict(self, x, sample_functions=True, sample_from_prior=False):
        if sample_functions:
            self.sample_and_set_functions(sample_from_prior)

        y_hat = self.model(x)
        pred_dist = self.likelihood.get_pred_dist(y_hat)
        return y_hat, pred_dist

    @tf.function
    def sample_functions(self, sample_from_prior):
        """

        Args:
            sample_from_prior (): boolean

        Returns:
            A matrix with dimensions [n_batched_models, n_parameters_model (+ n_parameters_likelihood)].
        """

        if sample_from_prior:
            sample = self.prior_module.sample(self.n_samples_per_prior)
        else:
            sample = self.posterior_module.sample(self.n_samples_per_posterior)

        sample = tf.reshape(sample, (self.n_batched_models, sample.shape[-1]))
        return sample

    @tf.function
    def sample_and_set_functions(self, sample_from_prior):
        sample = self.sample_functions(sample_from_prior)

        if self.model_variables_only:
            mp = tf.split(sample, self.split_sizes, axis=-1)[0]
            self.model.set_variables_vectorized(mp)
        else:
            mp, lp = tf.split(sample, self.split_sizes, axis=-1)
            self.model.set_variables_vectorized(mp)
            self.likelihood.set_variables_vectorized(lp)

class MLAP_RandomisedPredictor(BaseRandomisedPredictor):
    def __init__(self, model_config, likelihood_config, prior_config, posterior_config, n_batched_models, hyper_prior_config=None,
                 hyper_posterior_config=None, batch_sample=None, meta_learned_prior_module=None):
        super().__init__(model_config, likelihood_config, n_batched_models, batch_sample)
        # Assert that you are either in Meta-Learning setting (you have passed both a hyper-prior and a hyper-posterior config)
        # assert (hyper_prior_config is None and hyper_prior_config is None) or (hyper_prior_config is not None and hyper_prior_config is not None)
        # or you are have passed a learned prior module. But not both.
        # assert (hyper_prior_config or meta_learned_prior_module) and (not hyper_prior_config or not meta_learned_prior_module)

        self.hyper_prior_config = hyper_prior_config
        self.hyper_posterior_config = hyper_posterior_config

        # Setup prior
        self.prior_module = modules.BatchedGaussianPrior(batched_model=self.model, likelihood=self.likelihood,
                                                         config=prior_config, name='gaussian_prior')
        self.split_sizes = self.prior_module.split_sizes
        assert self.n_batched_models % self.prior_module._n_batched_models == 0
        self.n_samples_per_prior = self.n_batched_models // self.prior_module._n_batched_models


        # Copy prior variables and instantiate model from learned prior, if trained version is provided
        if meta_learned_prior_module is not None:
            print("NOTE: Meta-learned prior module has been provided. Thus this distribution would be used to initialize "
                  "the posterior variables.")
            self.prior_module.copy_variables(meta_learned_prior_module)
            self.sample_and_set_functions(sample_from_prior=True)

        # Setup posterior
        self.posterior_module = modules.BatchedGaussianPrior(batched_model=self.model, likelihood=self.likelihood,
                                                             config=posterior_config, name='gaussian_posterior')
        assert self.n_batched_models % self.posterior_module._n_batched_models == 0
        self.n_samples_per_posterior = self.n_batched_models // self.posterior_module._n_batched_models

        # Setup helper variables used in the loss computation
        self._set_up_helper_variables()

        # Setup custom fwd and bwd pass
        self._construct_custom_log_likelihood_fwd_bwd_pass()


    def _set_up_helper_variables(self):
        prior_means = []
        prior_log_vars = []

        for variable in self.prior_module.variables:
            if 'mean' in variable.name:
                prior_means.append(variable)
            else:
                prior_log_vars.append(variable)

        posterior_means = []
        posterior_log_vars = []

        for posterior in self.posterior_module.priors:
            means = []
            log_vars = []
            for variable in posterior.variables:
                if 'mean' in variable.name:
                    means.append(variable)
                else:
                    log_vars.append(variable)

            posterior_means.extend(means)
            posterior_log_vars.extend(log_vars)

            # Assert correct the number and order of the variables matches
            assert len(means) == len(prior_means)

            for prior_variable, posterior_variable in zip(prior_means, means):
                prior_variable_name = prior_variable.name.split('/')[1]
                posterior_variable_name = posterior_variable.name.split('/')[1]

                assert prior_variable_name == posterior_variable_name

        self.prior_means = prior_means
        self.prior_log_vars = prior_log_vars

        self.posterior_means = posterior_means
        self.posterior_log_vars = posterior_log_vars

    # Uses the helper variables and hyper_posterior_config (the fixed std)
    @tf.function
    def expected_KLs(self):
        # Expected KL between the hyperposteriors and a prior sampled from the hyper-posterior
        # Eq. 21 in Ruffels's write-up

        hyper_posterior_std = self.hyper_posterior_config['std']
        kappa_q_2 = hyper_posterior_std * hyper_posterior_std

        prior_means = self.prior_means
        prior_means = tf.reshape(tf.concat(prior_means, axis=0), (1, -1))

        posterior_means = self.posterior_means
        posterior_means = tf.reshape(tf.concat(posterior_means, axis=0), (self.posterior_module._n_batched_models, -1))

        prior_log_vars = self.prior_log_vars
        prior_log_vars = tf.reshape(tf.concat(prior_log_vars, axis=0), (1, -1))
        posterior_log_vars = self.posterior_log_vars
        posterior_log_vars = tf.reshape(tf.concat(posterior_log_vars, axis=0), (self.posterior_module._n_batched_models, -1))

        first_term = prior_log_vars - posterior_log_vars

        posterior_var = tf.exp(posterior_log_vars)
        means_squared = tf.square((posterior_means - prior_means))
        numerator = posterior_var + means_squared + kappa_q_2
        multiplier = tf.exp(-prior_log_vars + 0.5 * kappa_q_2)
        second_term = tf.multiply(numerator, multiplier)

        total = 0.5 * tf.reduce_sum(first_term + second_term - 1, axis=1)

        return total

    # Requires hyper_prior_config and hyper_posterior_config (the fixed stds)
    @tf.function
    def hyper_kl(self):
        # NB: This is a wrong formula (up to a constant multiplier) in both Ruffel's and MLAP's implementation
        kappa_posterior = self.hyper_posterior_config['std']
        kappa_prior = self.hyper_prior_config['std']

        ratio = kappa_posterior / kappa_prior
        ratio = ratio * ratio

        kappa_2 = kappa_prior * kappa_prior
        posterior_norm = tf.reduce_sum([tf.square(v) for v in self.prior_module.variables])

        res = 0.5 * (-tf.math.log(ratio) + ratio + posterior_norm / kappa_2 - 1)

        return res

    def _construct_custom_log_likelihood_fwd_bwd_pass(self):
        """
        Constructs custom forward and backward pass for batched likelihoods
        This is required since otherwise gradients are cut-off when the model's variables are assigned
        """

        @tf.custom_gradient
        def custom_mse_model(model_params, x, y):
            model = self.model

            model.set_variables_vectorized(model_params)

            tape = tf.GradientTape()
            with tape:
                tape.watch(model.variables)
                y_hat = model(x)

                per_model_mse = tf.reduce_mean(tf.reduce_sum(tf.square(y_hat - y), axis=-1), -1)
                per_model_mse = tf.reshape(per_model_mse, (-1, self.n_samples_per_posterior, 1))
                per_posterior_mse = tf.reduce_mean(per_model_mse, -2)

            def grad_fn(dy, variables):
                with tape:
                    tampered_mse = per_posterior_mse * dy

                g_mp = model.concat_and_vectorize_grads(tape.gradient(tampered_mse, model.variables))

                return [g_mp, tf.zeros_like(x), tf.zeros_like(y)], [None] * len(variables)

            return per_posterior_mse, grad_fn

        self.custom_mse_model = custom_mse_model

        # custom forward and backward pass for batched likelihood when only the NN model has learnable variables
        @tf.custom_gradient
        def custom_log_likelihood_model(model_params, x, y):
            model = self.model
            likelihood = self.likelihood

            model.set_variables_vectorized(model_params)

            tape = tf.GradientTape()
            with tape:
                tape.watch(model.variables)
                y_hat = model(x)

                per_model_ll = likelihood.log_prob(y_hat, y)
                per_model_ll = tf.reshape(per_model_ll, (-1, self.n_samples_per_posterior, 1))
                per_posterior_ll = tf.reduce_mean(per_model_ll, -2)

            def grad_fn(dy, variables):
                with tape:
                    tampered_ll = per_posterior_ll * dy

                g_mp = model.concat_and_vectorize_grads(tape.gradient(tampered_ll, model.variables))

                return [g_mp, tf.zeros_like(x), tf.zeros_like(y)], [None] * len(variables)

            return per_posterior_ll, grad_fn

        self.custom_log_likelihood_model = custom_log_likelihood_model