import tensorflow as tf

from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Conv2D, Flatten, Reshape, MaxPooling2D

from .batched_module import TFModuleBatched


class CNN(TFModuleBatched):
    requires_input_reshape = True

    def __init__(self, config, name):
        super().__init__(name=name)
        self.config = config

        # img_height = config['img_height']
        # img_width = config['img_width']
        # img_depth = config['img_depth']
        # self.img_shape = (img_height, img_width, img_depth)

        output_dim = config['output_dim']

        assert len(config['activations']) == 2
        kernel_activation = config['activations'][0]
        output_activation = config['activations'][-1]

        if output_dim == 1:
            assert output_activation == 'sigmoid'
        else:
            assert output_activation == 'softmax'

        self.model = Sequential([
            Conv2D(16, 5, padding='valid', activation=kernel_activation),
            MaxPooling2D(),
            Conv2D(32, 3, padding='valid', activation=kernel_activation),
            MaxPooling2D(),
            Conv2D(64, 3, padding='valid', activation=kernel_activation),
            MaxPooling2D(),
            Flatten(),
            Dense(256, activation=kernel_activation),
            Dense(output_dim, activation=output_activation)
        ])

    @tf.function
    def __call__(self, inputs):
        return self.model(inputs)


class FullyConnectedNN(TFModuleBatched):
    requires_input_reshape = False

    def __init__(self, config, name):
        super().__init__(name=name)
        self.layers = []
        self.config = config

        for n_units, activation in zip(config['layer_units'], config['activations']):
            self.layers.append(tf.keras.layers.Dense(units=n_units, activation=activation))

    @tf.function
    def __call__(self, inputs):
        x = inputs
        for layer in self.layers:
            x = layer(x)
        return x


class BatchedNNsModel(TFModuleBatched):
    def __init__(self, model_config, n_batched_models, name='batched_NN'):
        super().__init__(name=name, n_batched_models=n_batched_models)
        self.n_batched_models = n_batched_models
        self.output_dim = model_config['output_dim']

        # For CNNs, get image shape to reshape the input back into a the original image format
        # self.reshape_input = model_config["model_class"].requires_input_reshape
        # if self.reshape_input:
        #     self.img_shape = (model_config['img_height'], model_config['img_width'], model_config['img_depth'])

        self.models = []
        for i in range(n_batched_models):
            self.models.append(model_config['model_class'](model_config, name=f'{self.name}_{i}'))

    def initialize_variables(self, x, y):
        self.__call__(x)

    @tf.function
    def __call__(self, inputs):
        # inputs is a tensor with dimensions (n_batched_models x batch_size x input_shape)
        # outputs is a tensor with dimensions (n_batched_models x batch_size x output_shape)
        # if self.reshape_input:
        #     inputs = tf.reshape(inputs, (inputs.shape[0], inputs.shape[1]) + self.img_shape)
        outputs = tf.stack([self.models[i](inputs[i]) for i in range(self.n_batched_models)])
        return outputs