import tensorflow as tf
import tensorflow_probability as tfp
import math

from .batched_module import TFModuleBatched
from .neural_network import CNN, FullyConnectedNN
from .likelihood import ClassificationLikelihood, GaussianLikelihood

class GaussianPrior(tf.Module):
    # [var for var in gp.variables if 'prior' in var.name]
    def __init__(self, model, likelihood, config, name):
        """
        Gaussian prior for a NN model
        Args:
            model: NN model or list of NN models
            likelihood: likelihood object
            config (dict): configuration dict
            name (str): module name
        """
        super().__init__(name=name)

        self.config = config

        self.base_variable_sizes = []
        self.base_variable_names = []

        # initialize prior associated with the model parameters
        if config.get('init_from_models', False):
            assert type(model) == list
            self._init_model_prior_from_models(models=model, config=config)
        else:
            if type(model) == list:
                model = model[0]
            self._init_model_prior_from_config(model, config)

        # initialize prior associated with the likelihood parameters
        self._init_likelihood_prior(likelihood, config)

        # size of the model vs. likelihood parameters so that they can be separated if necessary
        self.split_sizes = [tf.reduce_sum(v_s).numpy() for v_s in self.base_variable_sizes]

        # add noise to initialization
        if config['add_noise']:
            self._add_noise_to_variables(config)

    def _sample_param(self, param_name, n_samples, use_init_std):
        param_name = f'{self.name}/{param_name}'

        name = f'{param_name}_mean'
        mean = getattr(self, name)

        name = f'{param_name}_log_var'
        log_var = getattr(self, name)

        if use_init_std:
            std = tf.ones_like(log_var) * self.config['init_std']
        else:
            std = tf.math.exp(0.5 * log_var)

        dist = tfp.distributions.Independent(
            tfp.distributions.Normal(mean, std),
            reinterpreted_batch_ndims=1)

        return dist.sample(n_samples)

    def sample(self, n_samples, use_init_std=False):
        """
        Generates n_samples samples from the prior distribution
        Args:
            n_samples (int): number of samples to draw
            use_init_std (bool): whether to use the init std

        Returns: Tensor of shape (n_samples, parameter_size)

        """
        mp = self._sample_param('model_parameters', n_samples, use_init_std)

        if self.learn_likelihood_variables:
            lp = self._sample_param('likelihood_parameters', n_samples, use_init_std)
            return tf.concat([mp, lp], axis=1)
        else:
            return mp

    @tf.function
    def _log_prob(self, param_name, param):
        param_name = f'{self.name}/{param_name}'

        name = f'{param_name}_mean'
        mean = getattr(self, name)

        name = f'{param_name}_log_var'
        log_var = getattr(self, name)
        std = tf.math.exp(0.5 * log_var)

        dist = tfp.distributions.Independent(
            tfp.distributions.Normal(mean, std),
            reinterpreted_batch_ndims=1)

        return dist.log_prob(param)

    @tf.function
    def log_prob(self, parameters):
        split = tf.split(parameters, self.split_sizes, axis=-1)

        log_prob = self._log_prob('model_parameters', split[0])

        if self.learn_likelihood_variables:
            log_prob += self._log_prob('likelihood_parameters', split[1])

        return log_prob

    def _init_model_prior_from_config(self, model, config):
        if isinstance(model, FullyConnectedNN) or isinstance(model, CNN):
            variables = model.variables
            variable_sizes = model.variable_sizes()

            self.base_variable_names.append([v.name for v in variables])
            self.base_variable_sizes.append(variable_sizes)

            means = []
            log_vars = []

            for v, v_size in zip(variables, variable_sizes):
                v_name = v.name

                mean = tf.ones((1, v_size), tf.float32) * config['model']['mean_mu']

                if 'bias' in v_name:
                    log_var = tf.ones((1, v_size), tf.float32) * config['model']['bias_log_var_mu']

                elif 'kernel' in v_name:
                    log_var = tf.ones((1, v_size), tf.float32) * config['model']['weight_log_var_mu']
                else:
                    raise Exception("Unexpected parameter")

                means.append(mean)
                log_vars.append(log_var)

            means = tf.concat(means, axis=1)
            log_vars = tf.concat(log_vars, axis=1)

            name = f'{self.name}/model_parameters_mean'
            setattr(self, name, tf.Variable(tf.squeeze(means), dtype=tf.float32, name=name, trainable=True))

            name = f'{self.name}/model_parameters_log_var'
            setattr(self, name, tf.Variable(tf.squeeze(log_vars), dtype=tf.float32, name=name, trainable=True))
        else:
            raise Exception("Unexpected model class")

    def _init_model_prior_from_models(self, models, config):
        # check that model variables have the same size
        assert len(set([tuple(model.get_variables_vectorized().shape) for model in models])) == 1

        # save variable names and sizes
        self.base_variable_names.append([v.name for v in models[0].variables])
        self.base_variable_sizes.append(models[0].variable_sizes())

        means = []
        log_vars = []

        for variable_tuple in zip(*[model.variables for model in models]):
            # assert that the variables have the same shape and name
            assert len(set([tuple(var.shape) for var in variable_tuple])) == 1
            assert len(set([var.name.split('/')[-1] for var in variable_tuple]))

            var_name = variable_tuple[0].name
            var_shape = variable_tuple[0].shape

            var_stack = tf.stack(variable_tuple, axis=0)
            init_std = tf.math.reduce_std(var_stack, axis=0)  # std of model initializations

            if 'kernel' in var_name:
                # take the initialization of the first model as mean of the prior
                init_mean = var_stack[0]
                # take half of the std across model initializations as std of the prior
                init_log_var = 2 * tf.math.log(0.5 * (init_std + 1e-8))
            elif 'bias' in var_name:
                # sample prior mean of bias
                init_mean = tf.random.normal(var_shape, mean=0.0, stddev=config['init_std'])
                # use init_std for the log_var
                init_log_var = tf.ones(var_shape) * 2 * tf.math.log(0.5 * (config['init_std'] + 1e-8))
            else:
                raise Exception("Unknown variable type")


            means.append(tf.reshape(init_mean, (-1,)))
            log_vars.append(tf.reshape(init_log_var, (-1,)))

        means = tf.concat(means, axis=0)
        log_vars = tf.concat(log_vars, axis=0)

        name = f'{self.name}/model_parameters_mean'
        setattr(self, name, tf.Variable(tf.squeeze(means), dtype=tf.float32, name=name, trainable=True))

        name = f'{self.name}/model_parameters_log_var'
        setattr(self, name, tf.Variable(tf.squeeze(log_vars), dtype=tf.float32, name=name, trainable=True))

    def _init_likelihood_prior(self, likelihood, config):
        self.learn_likelihood_variables = False
        if isinstance(likelihood, GaussianLikelihood):
            if len(likelihood.trainable_variables) > 0:
                assert (likelihood.config['learn_likelihood_variables'])
                self.learn_likelihood_variables = True
                self.base_variable_names.append([likelihood.variables[0].name])
                self.base_variable_sizes.append([likelihood.output_dim])

                mean = tf.ones(likelihood.output_dim, tf.float32) * config['likelihood']['log_var_mu']
                name = f'{self.name}/likelihood_parameters_mean'
                setattr(self, name, tf.Variable(mean, dtype=tf.float32, name=name, trainable=True))

                log_var = tf.ones(likelihood.output_dim, tf.float32) * config['likelihood']['log_var_log_var']
                name = f'{self.name}/likelihood_parameters_log_var'
                setattr(self, name, tf.Variable(log_var, dtype=tf.float32, name=name, trainable=True))
        elif isinstance(likelihood, ClassificationLikelihood):
            pass
        else:
            raise Exception("Unexpected likelihood class")

    def _add_noise_to_variables(self, config):
        if config.get('init_from_models', False):
            # exclude mean and log_var for model parameters, since they have been initialized from the models
            # --> they are already noisy
            vars_to_add_noise = [var for var in self.variables if not 'model_parameters' in var.name]
        else:
            vars_to_add_noise = self.variables
        for v in vars_to_add_noise:
            noise = tf.random.normal(v.shape, mean=0.0, stddev=config['init_std'])
            v.assign_add(noise)

class BatchedGaussianPrior(TFModuleBatched):

    def __init__(self, batched_model, likelihood, config, name='batched_Gaussian_prior'):
        """
        Batched Gaussian priors for the model
        Args:
            batched_model: a batched NN model for which to instantiate a prior
            likelihood: likelihood object
            config (dict): configuration for the prior
            name (str): name of the module
        """
        super().__init__(name=name, n_batched_models=config['n_batched_priors'])
        self.n_batched_priors = config['n_batched_priors']
        self.priors = []

        n_models = len(batched_model.models)
        n_models_per_prior = n_models // self.n_batched_priors
        assert n_models_per_prior > 0

        for i in range(self.n_batched_priors):
            models_for_init = batched_model.models[i*n_models_per_prior:(i+1)*n_models_per_prior]
            self.priors.append(GaussianPrior(models_for_init, likelihood, config, name=f'{self.name}_{i}'))

        self.split_sizes = self.priors[0].split_sizes
        self._variable_sizes = None

        self._construct_vectorized_sampler()

    @tf.function
    def log_prob(self, parameters):
        # parameters should have dimensions [#priors, #samples_per_prior, #params]
        log_prob = tf.stack([self.priors[i].log_prob(parameters[i]) for i in range(self.n_batched_priors)])

        return log_prob


    def sample(self, n_samples, use_init_std=False):
        sample = tf.stack([self.priors[i].sample(n_samples, use_init_std)
                           for i in range(self.n_batched_priors)])
        return sample

    def _construct_vectorized_sampler(self):

        @tf.custom_gradient
        def sample_parametrized(n_samples, variables_vectorized):
            self.set_variables_vectorized(variables_vectorized)

            tape = tf.GradientTape(persistent=True)
            with tape:
                tape.watch(list(self.trainable_variables))
                samples = self.sample(n_samples)

            def grad_fn(dy, variables):
                with tape:
                    tampered_samples = samples * dy
                grads_n_w = tape.gradient(tampered_samples, list(self.trainable_variables))
                grads_to_input = [None, self.concat_and_vectorize_grads(grads_n_w)]
                return grads_to_input, [None] * len(variables)

            return samples, grad_fn

        self.sample_parametrized = sample_parametrized

class ParticleDist(tf.Module):
    def __init__(self, prior, config, name='particle_dist'):
        super().__init__(name=name)

        self.split_sizes = prior.split_sizes
        self.config = config
        sample = prior.sample(config['n_samples'], config['use_init_std'])
        self.particles = tf.Variable(sample, dtype=tf.float32,
                                     trainable=False)  # [#priors, #samples_per_prior, #params]

    def sample(self, n_samples=None):
        assert n_samples is None or n_samples == self.particles.shape[1]
        return self.particles

    def log_prob(**kwargs):
        raise Exception("Not implemented for particle dist")

class GaussianPosteriorDist(tf.Module):
    def __init__(self, prior,  init_from_prior=False, log_var_init=-1, n_prior_samples_init=5,
                 likelihood_log_std_init=-3, name='gaussian_posterior'):
        super().__init__(name=name)

        self.split_sizes = prior.split_sizes

        if init_from_prior:
            self._initalize_from_prior(prior, n_prior_samples_init)
        else:
           self._initialize_manually(prior, log_var_init, likelihood_log_std_init)

    @property
    def stddev(self):
        return tf.math.exp(self.log_stddev)

    @property
    def dist(self):
        return tfp.distributions.Independent(tfp.distributions.Normal(self.mu, self.stddev),
                                             reinterpreted_batch_ndims=1)

    def sample(self, n_samples):
        return tf.transpose(self.dist.sample(n_samples), [1, 0, 2])

    def log_prob(self, param_values):
        return self.dist.log_prob(param_values)

    def parameters(self):
        return [self.mu, self.log_stddev]

    def parameters_model(self):
        model_params_size = self.split_sizes[0]
        return [self.mu[:, :model_params_size], self.log_stddev[:, :model_params_size]]

    def parameters_likelihood(self):
        if len(self.split_sizes) == 2:
            model_params_size = self.split_sizes[0]
            return [self.mu[:, model_params_size:], self.log_stddev[:, model_params_size:]]
        else:
            return None

    def _initalize_from_prior(self, prior, n_prior_samples_init):
        # initialize mu and log_std by sampling from prior and calculating the empirical mean & std
        assert n_prior_samples_init > 1
        sample = prior.sample(n_prior_samples_init)
        self.mu = tf.Variable(tf.reduce_mean(sample, axis=1), name='mu')
        self.log_stddev = tf.Variable(tf.math.log(tf.math.reduce_std(sample, axis=1)), name='log_std')

    def _initialize_manually(self, prior, log_var_init=-1, likelihood_log_std_init=-3):
        param_sample_shape = prior.sample(1).shape
        assert sum(self.split_sizes) == param_sample_shape[-1]

        # init for model parameters
        mp_shape = (param_sample_shape[0], self.split_sizes[0])
        mp_mu_init = tf.random.normal(mp_shape, stddev=0.5 * math.exp(log_var_init))
        mp_log_std_init = log_var_init * tf.ones(mp_shape)

        if len(self.split_sizes) == 1:
            mu_init = mp_mu_init
            log_std_init = mp_log_std_init
        elif len(self.split_sizes) == 2:
            # initialize VI dist for params for likelihood differently
            likelihood_param_size = self.split_sizes[1]
            lp_mu_init = likelihood_log_std_init * tf.ones((1, likelihood_param_size))
            lp_log_std_init = tf.zeros((1, likelihood_param_size))

            # concatenate model and likelihood init parameters
            mu_init = tf.concat([mp_mu_init, lp_mu_init], axis=-1)
            log_std_init = tf.concat([mp_log_std_init, lp_log_std_init], axis=-1)
        else:
            raise AssertionError

        self.mu = tf.Variable(mu_init, name='mu')
        self.log_stddev = tf.Variable(log_std_init, name='log_std')