import tensorflow as tf
import math

import pacoh.modules as modules
from pacoh import util as util


class MetaRandomisedPredictor(tf.Module):
    def __init__(self, model_config, likelihood_config, prior_config, hyper_prior_config,
                 hyper_posterior_config, meta_train_sampler, batch=None, logsumexp_estimate=False, catoni_constant = 10):

        self.model_config = model_config
        self.likelihood_config = likelihood_config
        self.logsumexp_estimate = logsumexp_estimate
        self.catoni_constant = catoni_constant

        # batching settings
        prior_config['n_batched_priors'] = hyper_posterior_config['n_samples']
        self.n_batched_models = prior_config['n_batched_priors'] * prior_config['n_samples']
        self.n_batched_priors = prior_config['n_batched_priors']
        self.n_samples_per_prior = prior_config['n_samples']

        # make sure that the likelihood is consistent with the NN setup
        _check_model_likelihood_compatibility(model_config, likelihood_config)

        # setup the NN model and likelihood
        self.model = modules.BatchedNNsModel(model_config=self.model_config, n_batched_models=self.n_batched_models)
        self._initialize_model_variables(meta_train_sampler, batch)

        self.likelihood = util.get_likelihood(config=likelihood_config, output_dim=self.model_config['output_dim'],
                                              n_batched_models=self.n_batched_models)

        # setup prior, posterior & hyper-prior

        self.prior_module = modules.BatchedGaussianPrior(batched_model=self.model, likelihood=self.likelihood,
                                                         config=prior_config, name='gaussian_prior')
        self.hyper_prior_module = modules.GaussianHyperPrior(self.prior_module, hyper_prior_config,
                                                             name='gaussian_hyper_prior')
        init_particles_hyper = tf.expand_dims(self.prior_module.get_variables_stacked_per_model(), axis=0)
        self.hyper_posterior_module = modules.ParticleHyperPosterior(config=hyper_posterior_config,
                                                                    name='particle_hyper_posterior',
                                                                    init_particles=init_particles_hyper)

        # setup PACOH prefactor (m \ (m+n))
        dataset_sizes = tf.convert_to_tensor(meta_train_sampler.n_train_samples, dtype=tf.float32)
        harmonic_mean_dataset_size = 1. / (tf.reduce_mean(1. / dataset_sizes))
        # if logsumexp_estimate:
        #     self.pre_factor = 1 / ((tf.math.sqrt(harmonic_mean_dataset_size*meta_train_sampler.n_tasks)+1))
        # else:
        #     self.pre_factor = harmonic_mean_dataset_size / (harmonic_mean_dataset_size + meta_train_sampler.n_tasks)

        # return 1 / ((tf.math.sqrt(harmonic_mean_dataset_size * self.meta_train_sampler.n_tasks) + 1)) # original code
        abel_numerator = math.e * self.catoni_constant * math.log(self.catoni_constant) 
        abel_constant = abel_numerator / (abel_numerator + self.catoni_constant * harmonic_mean_dataset_size * (math.e -1))

        self.pre_factor = abel_constant

        self.meta_train_sampler = meta_train_sampler

        # check whether likelihood has trainable variables
        self.model_variables_only = not len(self.likelihood.variables) > 0

    @tf.function
    def sample_functions_from_prior(self):
        sample = self.prior_module.sample(self.n_samples_per_prior)
        if self.model_variables_only:
            mp = tf.split(sample, self.prior_module.split_sizes, axis=-1)[0]
            self.model.set_variables_vectorized(mp)
        else:
            mp, lp = tf.split(sample, self.prior_module.split_sizes, axis=-1)
            self.model.set_variables_vectorized(mp)
            self.likelihood.set_variables_vectorized(lp)

    def predict(self, x, sample_functions=False, sample_from_prior=False):
        if sample_functions:
            assert sample_from_prior
            self.sample_functions_from_prior()

        y_hat = self.model(x)
        pred_dist = self.likelihood.get_pred_dist(y_hat)
        return y_hat, pred_dist

    @tf.function
    def compute_likelihood_across_tasks(self, params, meta_batch_x, meta_batch_y):
        """ compute the average likelihood, i.e. the mean of the likelihood for the points in the batch (x,y)
            If you want an unbiased estimator of the dataset likelihood, set the prefactor to the number of points

            Returns: log_likelihood_across_tasks with shape (meta_batch_size, n_hyper_posterior_samples, n_prior_samples)
                """
        if self.model_variables_only:
            model_params = tf.split(params, self.prior_module.split_sizes, axis=-1)[0]
        else:
            model_params, likelihood_params = tf.split(params, self.prior_module.split_sizes, axis=-1)

        # iterate over tasks
        log_likelihood_list = []
        for i in range(self.meta_train_sampler.meta_batch_size):
            x = meta_batch_x[i]
            y = meta_batch_y[i]

            # NN forward pass
            y_hat = self.model.forward_parametrized(x, model_params)

            # compute likelihood
            if self.model_variables_only:
                log_likelihood = self.likelihood.log_prob(y_hat, y)
            else:
                log_likelihood = self.likelihood.log_prob(y_hat, y, likelihood_params)
            log_likelihood = tf.reshape(log_likelihood, (self.n_batched_priors, self.n_samples_per_prior))
            log_likelihood_list.append(log_likelihood)

        log_likelihood_across_tasks = tf.stack(log_likelihood_list)
        # tf.assert_equal(params.shape[:-1], log_likelihood_across_tasks.shape[1:])
        return log_likelihood_across_tasks

    @tf.function
    def compute_mll_estimate(self, prior_samples, meta_batch_x, meta_batch_y, n_train_samples):
        param_sample = self.prior_module.sample_parametrized(self.n_samples_per_prior, prior_samples)
        log_likelihood = self.compute_likelihood_across_tasks(param_sample, meta_batch_x, meta_batch_y)

        if self.logsumexp_estimate:
            # multiply by sqrt of m and apply logsumexp
            neg_log_likelihood = log_likelihood * tf.math.sqrt(n_train_samples[:, None, None])
            mll = tf.math.reduce_logsumexp(neg_log_likelihood, axis=-1) - tf.math.log(
                tf.cast(self.n_samples_per_prior, dtype=tf.float32))

            # sum over meta-batch size and adjust for number of tasks
            mll_sum = tf.reduce_sum(mll, axis=0) * \
                      (self.meta_train_sampler.n_tasks / self.meta_train_sampler.meta_batch_size)
        else:
            # average over prior samples
            mll = tf.reduce_mean(log_likelihood, axis=-1)
            # sum over meta-batch size and adjust for number of tasks
            mll_sum = tf.reduce_sum(mll, axis=0) * (
                    self.meta_train_sampler.n_tasks / self.meta_train_sampler.meta_batch_size)

        # multiply with pre_factor (m / (m+n)) or 1 / (sqrt(nm) + 1) in case of the logsumexp estimate
        mll_sum *= self.pre_factor
        return mll_sum

    def log_prob_hyper_prior(self, prior_samples, prior_weight):
        return prior_weight * self.hyper_prior_module.log_prob_vectorized(prior_samples)

    def _initialize_model_variables(self, meta_train_sampler, batch=None):
        if batch is None:
            batch_sample = tf.data.experimental.get_next_as_optional(meta_train_sampler.task_sampler)
            x, y, _ = batch_sample.get_value()
            x.set_shape(meta_train_sampler.batch_shapes[0])
            y.set_shape(meta_train_sampler.batch_shapes[1])
        else:
            x, y = batch

        self.model.initialize_variables(x, y)

class RandomisedPredictor(tf.Module):
    """
    Randomized neural network. In particular, the class encapsulates
        - a batched NN model
        - a prior distribution
        - a likelihood function
        - a variational posterior

    """

    def __init__(self, model_config, likelihood_config, prior_config, posterior_config, batch_sample,
                 meta_learned_prior_module=None):
        """
        Args:
            model_config (dict): configuration of the neural network model
            likelihood_config (dict): configuration of the likelihood function (e.g. Gaussian likelihood for regression,
                                      Softmax likelihood for classification)
            prior_config (dict): configuration of the prior distribution
            posterior_config (dict): configuration of the variational posterior
            batch_sample (tuple): (x,y) sample from the DatasetSampler - tuple of tensors
            meta_learned_prior_module: prior module with the same specifications as in prior_config
        """
        self.model_config = model_config
        self.likelihood_config = likelihood_config

        # batching settings
        self.n_batched_models = prior_config['n_batched_priors'] * posterior_config['n_samples']
        self.n_batched_priors = prior_config['n_batched_priors']
        self.n_samples_per_prior = posterior_config['n_samples']

        # make sure that the likelihood is consistent with the NN setup
        _check_model_likelihood_compatibility(model_config, likelihood_config)

        # setup the NN model and likelihood
        self.model = modules.BatchedNNsModel(model_config=self.model_config, n_batched_models=self.n_batched_models)
        self.model.initialize_variables(batch_sample[0], batch_sample[1])
        self.likelihood = util.get_likelihood(config=likelihood_config, output_dim=self.model_config['output_dim'],
                                              n_batched_models=self.n_batched_models)

        # setup the prior
        self.prior_module = modules.BatchedGaussianPrior(batched_model=self.model, likelihood=self.likelihood,
                                                         config=prior_config, name='gaussian_prior')

        # set variables of the prior to the meta-learner prior if provided
        if meta_learned_prior_module is not None:
            self.prior_module.copy_variables(meta_learned_prior_module)
            if posterior_config.get('use_init_std', False):
                posterior_config['use_init_std'] = False

        # setup posterior
        self.n_posterior_samples = posterior_config['n_samples']
        assert posterior_config['type'] in ['particle', 'gaussian']
        if posterior_config['type'] == 'gaussian':
            del posterior_config['type']
            del posterior_config['n_samples']
            self.posterior_module = modules.GaussianPosteriorDist(prior=self.prior_module, **posterior_config)
        elif posterior_config['type'] == 'particle':
            self.posterior_module = modules.ParticleDist(prior=self.prior_module, config=posterior_config,
                                                         name='particle_posterior')
        else:
            raise NotImplementedError

        # check whether likelihood has trainable variables (e.g. the noise variance)
        if len(self.likelihood.trainable_variables) > 0:
            self.model_variables_only = False
        else:
            self.model_variables_only = True

        self.posterior_config = posterior_config
        self.prior_config = prior_config

    @tf.function
    def sample_params(self, n_samples=None, sample_from_prior=False):
        if n_samples is None:
            n_samples = self.n_samples_per_prior
        if sample_from_prior:
            param_sample = self.prior_module.sample(n_samples)
        else:
            param_sample = self.posterior_module.sample(n_samples)
        return param_sample

    @tf.function
    def sample_functions(self, n_samples=None, sample_from_prior=False):
        assert n_samples == self.n_posterior_samples
        param_sample = self.sample_params(n_samples, sample_from_prior)

        if self.model_variables_only:
            mp = tf.split(param_sample, self.posterior_module.split_sizes, axis=-1)[0]
            self.model.set_variables_vectorized(mp)
        else:
            mp, lp = tf.split(param_sample, self.posterior_module.split_sizes, axis=-1)
            self.model.set_variables_vectorized(mp)
            self.likelihood.set_variables_vectorized(lp)

    def prior_log_prob(self, params):
        return self.prior_module.log_prob(params)

    def posterior_log_prob(self, params):
        if hasattr(self.posterior_module, 'log_prob'):
            return self.posterior_module.log_prob(params)
        else:
            raise NotImplementedError('Posterior has np log_prob implementation')

    @tf.function
    def compute_likelihood(self, x, y, params, prefactor=1.0):
        """ compute the average likelihood, i.e. the mean of the likelihood for the points in the batch (x,y)
            If you want an unbiased estimator of the dataset likelihood, set the prefactor to the number of points"""
        if self.model_variables_only:
            model_params = tf.split(params, self.posterior_module.split_sizes, axis=-1)[0]
        else:
            model_params, likelihood_params = tf.split(params, self.posterior_module.split_sizes, axis=-1)

        # NN forward pass
        y_hat = self.model.forward_parametrized(x, model_params)

        # compute likelihood
        if self.model_variables_only:
            log_likelihood = self.likelihood.log_prob(y_hat, y)
        else:
            log_likelihood = self.likelihood.log_prob(y_hat, y, likelihood_params)
        log_likelihood *= prefactor
        log_likelihood = tf.reshape(log_likelihood, params.shape[:-1])

        tf.assert_equal(params.shape[:-1], log_likelihood.shape)
        return log_likelihood

    def predict(self, x, n_samples=None, sample_functions=True, sample_from_prior=False):
        if n_samples is None:
            n_samples = self.n_posterior_samples
        assert n_samples % self.n_posterior_samples == 0, "n_samples must be multiple of " \
                                                       "n_batched_models=%i"%self.n_batched_models

        # run multiple batches of random forward passes
        n_forward_batches = n_samples // self.n_posterior_samples
        y_hat_batches = []
        for _ in range(n_forward_batches):
            if sample_functions:
                self.sample_functions(n_samples=self.n_posterior_samples, sample_from_prior=sample_from_prior)
            y_hat_batches.append(self.model(x))

        # concatenate the predictions
        y_hat = tf.concat(y_hat_batches, axis=0)
        pred_dist = self.likelihood.get_pred_dist(y_hat)
        return y_hat, pred_dist

    def trainable_variables(self):
        return self.posterior_module.variables

def _check_model_likelihood_compatibility(model_config, likelihood_config):
    if 'class' in likelihood_config['type']:
        assert ((model_config['activations'][-1] == 'softmax' and model_config['output_dim'] > 1) or
                (model_config['activations'][-1] == 'sigmoid' and model_config['output_dim'] == 1))
    elif 'reg' in likelihood_config['type']:
        assert (model_config['activations'][-1] in ['linear', None]), 'Last activation must be None or `linear`. ' \
                                                                      'Provided activations" %s'%str(model_config['activations'])
    else:
        raise Exception('Unexpected likelihood')