import pacoh.experiments.hyperparam_search.tuners as tuners
import pacoh.datasets as datasets
import pacoh.algorithms as algorithms
import matplotlib.pyplot as plt
import tensorflow as tf
import os
import pandas as pd
import numpy as np
import json
import copy

from config import DATA_DIR

HPARAM_EXP_DIR = os.path.join(DATA_DIR, 'tune-hparam-pacoh')
DIR = os.path.dirname(os.path.abspath(__file__))
N_LAYERS = 4

def update_config(config, max_n_samples, plot, n_samples, prior_prob_weight=None):
    config = copy.deepcopy(config)

    config['data_config']['num_context_samples'] = config['data_config']['num_train_samples'] = max_n_samples
    config['data_config']['num_test_valid_tasks'] = 7
    config['data_config']['posterior_inference_batch_size'] = min(config['data_config']['posterior_inference_batch_size'], n_samples)

    prior_config = config['prior_config']
    config['prior_config'] = prior_config = prior_config[prior_config['type']]

    config['posterior_config'].update(config['posterior_config'].pop(config['posterior_config']['type']))
    posterior_config = config['posterior_config']

    hyper_posterior_config = config['hyper_posterior_config']
    config['hyper_posterior_config'] = hyper_posterior_config = hyper_posterior_config[
        hyper_posterior_config['type']]

    if 'activation_fn' in config['model_config']['fc']:
        activation_fn = config['model_config']['fc']['activation_fn']
        del config['model_config']['fc']['activation_fn']
        config['model_config']['fc']['activations'] = [activation_fn for _ in range(N_LAYERS)] + [None]

    config['meta_train_config']['plot_functions'] = plot
    config['meta_train_config']['plot_pred_lines'] = plot
    config['meta_train_config']['plot_pred_std'] = plot
    config['meta_train_config']['n_eval_runs'] = 0

    if prior_prob_weight is not None:
        config['meta_test_config']["prior_prob_weight"] = prior_prob_weight

    # config['meta_train_config']['n_total_iterations'] = 3
    # config['meta_test_config']['n_total_iterations'] = 3
    # config['meta_test_config']['bandwidth'] = 0.1

    return config


def main(args, base_config):

    from pacoh.experiments.utils import AsyncExecutor
    import itertools
    exec = AsyncExecutor(n_jobs=40)

    prior_prob_weights = [None, 0.1, 0.5, 1, 2, 3, 4, 5]
    seeds = [22, 23, 24, 25, 26]

    args = list(zip(*list(itertools.product([base_config], prior_prob_weights, seeds))))
    exec.run(generate_plots, *args)


def generate_plots(base_config, prior_prob_weight, seed):
    n_samples = [int(i) for i in args['n_samples'].split(',')]
    max_n_samples = max(n_samples)

    fig, axs = plt.subplots(1, len(n_samples), sharey=True, sharex=True, figsize=(8.5, 2.8))
    axs = axs.flatten()

    for ax, n_s in zip(axs, n_samples):
    # if n_s != 5:
    #     continue
        config = update_config(base_config, max_n_samples, False, n_s, prior_prob_weight)

        json_formatted_str = json.dumps(config, indent=2)
        print(json_formatted_str)

        data_config = config.pop('data_config')

        n_batched_models_train = config['hyper_posterior_config']['n_samples'] * config['prior_config']['n_samples']
        n_batched_models_test = config['hyper_posterior_config']['n_samples'] * config['posterior_config']['n_samples']

        train_ds, val_ds = datasets.get_regression_datasets(data_config,
                                                     n_batched_models_train=n_batched_models_train,
                                                     n_batched_models_test=n_batched_models_test,
                                                     add_meta_data_to_context=args.get('add_meta_data_to_context', False),
                                                     n_samples_to_keep_train=5,
                                                     n_samples_to_keep_val=n_s, seed=123)

        meta_model = algorithms.MetaBayesianNeuralNetwork_SVGD(meta_train_sampler=train_ds,
                                                               meta_test_sampler=val_ds, **config, random_seed=seed)

        meta_model.meta_fit()
        meta_learned_prior = meta_model.rp.prior_module

        task = val_ds.meta_train_tasks[-1]

        reg_model = meta_model.meta_test_inference_class(task, meta_model.model_config, meta_model.likelihood_config,
                                                    meta_model.prior_config, meta_model.posterior_config,
                                                    meta_model.meta_test_config,
                                                    meta_learned_prior_module=meta_learned_prior)
        reg_model.fit()

        task.plot_prediction_functions(model=reg_model, ax=ax,
                                       plot_pred_std=args['plot_pred_std'],
                                       plot_pred_lines=args['plot_pred_lines'],
                                       plot_data=True)
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_title(f'Training with {n_s} samples')

    ####
    axs[0].set_ylim((1.5, 9))

    handles, labels = axs[-1].get_legend_handles_labels()
    axs[-1].legend(handles, ['Test', 'Train'])

    #
    plt.tight_layout(rect=(0, 0, 1, 0.98))
    plt.show()

    #fig.savefig(os.path.join(DIR, f'a_svgd_sine_{prior_prob_weight}_final_2.png'))
    fig.savefig(os.path.join(DIR, f'b_svgd_sine_{prior_prob_weight}_{seed}_final_2.pdf'))

def get_dict_from_string(s):
    s = s.replace("'", '\"').replace('None', 'null').replace('True', 'true').replace('False', 'false')
    d = json.loads(s)
    return d

def get_best_config_from_csv(path, criterion):
    df = pd.read_csv(path)

    hyperparams = list(
        set(df.columns) - {'Unnamed: 0', 'timesteps_total', 'random_seed', 'randomness_config',
                           'cal_err', 'avg_ll', 'avg_rmse', 'loss', 'train_obj', 'avg_acc'})

    df_aggregated = df.groupby(hyperparams, as_index=False).aggregate(
        {m: [np.mean, np.std] for m in ['avg_rmse', 'avg_ll']})

    # print(df.groupby(hyperparams, as_index=False).aggregate(
    #     {m: [np.mean, np.std] for m in ['avg_rmse', 'avg_ll']}).reset_index(drop=True))

    if criterion == 'avg_ll':
        opt_idx = df_aggregated[criterion]['mean'].idxmax()
    elif criterion == 'avg_rmse':
        opt_idx = df_aggregated[criterion]['mean'].idxmin()

    columns = ['data_config', 'hyper_posterior_config', 'hyper_prior_config', 'inference_training_config',
               'likelihood_config', 'model_config', 'posterior_config',
               'prior_config', 'training_config', 'meta_test_config', 'meta_train_config']

    config = {}

    for col in columns:
        try:
            d = get_dict_from_string(df_aggregated.loc[opt_idx][col][0])
            config[col] = d
        except:
            print(f"Fetching parameters for column `{col}` failed.")

    if 'training_config' in config:
        meta_train_config = config.pop('training_config')
        config['meta_train_config'] = meta_train_config

    if 'inference_training_config' in config:
        meta_test_config = config.pop('inference_training_config')
        config['meta_test_config'] = meta_test_config

    print(df_aggregated.loc[opt_idx]['avg_rmse']['mean'])
    print(df_aggregated.loc[opt_idx]['avg_ll']['mean'])

    return config


if __name__ == '__main__':
    # path to the csv output file that the tuner outputs
    _csv_name = 'better.csv' # working.csv _csv_name = "april02_sin_20_pacoh_svgd_fixed_ll_svgd_Apr_03_2020_19:39:41.csv"
    _criterion = 'avg_ll'
    csv_path = os.path.join(HPARAM_EXP_DIR, _csv_name)

    config = get_best_config_from_csv(csv_path, _criterion)

    json_formatted_str = json.dumps(config, indent=2)
    print(json_formatted_str)

    args = {}
    args['dataset'] = 'sin_20'
    args['n_samples'] = '1, 5, 10'
    args['plot_pred_std'] = True
    args['plot_pred_lines'] = True

    ###
    debug = False

    config['meta_train_config']['logsumexp_estimate'] = True
    config['model_config']['fc']['use_bias'] = True
    config['meta_test_config']['n_total_iterations'] = 6000
    config['meta_train_config']['n_total_iterations'] = 6000
    # config['hyper_posterior_config']['particle']['n_samsples'] = 1
    config['meta_test_config']['bandwidth'] = 100

    if debug:
        config['posterior_config']['particle']['n_samples'] = 2
        config['hyper_posterior_config']['particle']['n_samples'] = 1
        config['meta_train_config']['n_total_iterations'] = 10
        config['meta_test_config']['n_total_iterations'] = 10
    ###

    # tf.config.experimental_run_functions_eagerly(True)
    # for b in [500, 1, 1000]:
    #     config['meta_test_config']['bandwidth'] = b
    #     temp = copy.deepcopy(config)
    #     main(args, temp)
    main(args, config)
