import math
import copy
import numpy as np
import tensorflow as tf

import pacoh.modules as modules

def get_likelihood(config, output_dim, n_batched_models):
    if config['type'] == 'class':
        return modules.ClassificationLikelihood(output_dim=output_dim)
    elif config['type'] == 'reg':
        return modules.GaussianLikelihood(config=config['reg'], output_dim=output_dim, n_batched_models=n_batched_models)
    else:
        raise Exception("Unexpected task likelihood requested")


def process_model_config(model_config, dataset):
    if model_config['type'] == 'cnn':
        model_config = model_config[model_config['type']]
        for key in ['img_height', 'img_width', 'img_depth']:
            model_config[key] = getattr(dataset, key)
        model_config['model_class'] = modules.CNN
    else:
        model_config = model_config[model_config['type']]
        model_config['model_class'] = modules.FullyConnectedNN

    return model_config


def process_training_config(training_config, dataset):
    # Dataset only needs to provide the number of steps_per_epoch
    training_config = copy.deepcopy(training_config)

    if 'n_total_iterations' not in training_config:
        epochs = training_config['epochs']
        steps_per_epoch = dataset.steps_per_epoch
        print(f'Training for `{epochs}` epochs, `{steps_per_epoch}` gradient steps each.')

        n_total_iterations = epochs * steps_per_epoch
        training_config['n_total_iterations'] = n_total_iterations

    # a step is comprises the gradient descent iterations between evaluations
    if 'n_optimization_steps' not in training_config:
        if training_config['n_eval_runs'] == 0:
            n_optimization_steps = 1
            n_iterations_per_step = training_config['n_total_iterations']
        else:
            n_optimization_steps = training_config['n_eval_runs'] + 1
            n_iterations_per_step = math.floor(training_config['n_total_iterations'] / n_optimization_steps)

        training_config['n_optimization_steps'] = n_optimization_steps
        training_config['n_iterations_per_step'] = n_iterations_per_step

        print(f'Training for {training_config["n_total_iterations"]} iterations, '
              f'structured in `{training_config["n_optimization_steps"]}` optimization steps, '
              f'where each step consists of `{training_config["n_iterations_per_step"]}` gradient descent iterations.')

    if 'log_period' not in training_config:
        if training_config['n_train_loss_logs'] != 0:
            training_config['log_period'] = max(1, math.floor(training_config['n_total_iterations'] / training_config['n_train_loss_logs']))
        else:
            training_config['log_period'] = training_config['n_total_iterations'] + 1

        print(f'The training objective will be reported every `{ training_config["log_period"]}` gradient iterations')

    return training_config
