import tensorflow as tf
import tensorflow.keras as keras
from tensorflow.keras import layers
import tensorflow.keras.backend as keras_backend
tf.keras.backend.set_floatx('float32')
import tensorflow_probability as tfp
from tensorflow_probability.python.layers import util as tfp_layers_util
import time
import math
import numpy as np

from pacoh_gp.abstract import RegressionModelMetaLearned
from pacoh_gp.util import _handle_input_dimensionality

class NNModel(keras.Model):
    def __init__(self, layer_sizes, output_dim=1):
        super().__init__()
        self.n_layers = len(layer_sizes)
        for i, size in enumerate(layer_sizes):
            setattr(self, 'hidden%i'%(i+1), keras.layers.Dense(size))
        self.out = keras.layers.Dense(output_dim)

    def call(self, x):
        for i in range(1, self.n_layers+1):
            x = keras.activations.relu(getattr(self, 'hidden%i'%i)(x))
        x = self.out(x)
        return x

class Encoder(keras.Model):
    def __init__(self, dim_w=5, name='encoder', **kwargs):
      super().__init__(name = name)
      self.dense_proj = layers.Dense(80, activation='relu')
      self.dense_mu = layers.Dense(dim_w)
      self.dense_sigma_w = layers.Dense(dim_w)

    def call(self, inputs):
      h = self.dense_proj(inputs)
      mu_w = self.dense_mu(h)
      sigma_w = self.dense_sigma_w(h)
      sigma_w = tf.nn.softplus(sigma_w)
      ws = mu_w + tf.random.normal(tf.shape(mu_w)) * sigma_w
      return ws, mu_w, sigma_w

def kl_qp_gaussian(mu_q, sigma_q, mu_p, sigma_p):
    """Kullback-Leibler KL(N(mu_q), Diag(sigma_q^2) || N(mu_p), Diag(sigma_p^2))"""
    sigma2_q = tf.square(sigma_q) + 1e-16
    sigma2_p = tf.square(sigma_p) + 1e-16
    temp = tf.math.log(sigma2_p) - tf.math.log(sigma2_q) - 1.0 + \
           sigma2_q / sigma2_p + tf.square(mu_q - mu_p) / sigma2_p  # n_target * d_w
    kl = 0.5 * tf.reduce_mean(temp, axis=1)
    return tf.reduce_mean(kl)

def np_to_tensor(list_of_numpy_objs):
    return [tf.convert_to_tensor(obj, dtype=tf.float32) for obj in list_of_numpy_objs]


class MAML_MR(RegressionModelMetaLearned):

    def __init__(self, meta_train_data, mode='MR-MAML-W', layer_sizes=(32, 32, 32, 32), num_iter_fit=20000,
                 lr_inner=0.01, lr_meta=1e-3, dim_w=5, kl_factor=5e-5, normalize_data=True, random_seed=None):

        super().__init__(normalize_data, random_seed)
        tf.random.set_seed(random_seed)

        assert mode in ['MAML', 'MR-MAML-W', 'MR-MAML-A']
        self.mode = mode

        # prepare data
        self._check_meta_data_shapes(meta_train_data)
        self._compute_normalization_stats(meta_train_data)
        self.meta_train_data = [self._normalize_data(x, y) for x, y in meta_train_data]

        # hyperparams
        self.lr_inner = lr_inner
        self.num_iter_fit = num_iter_fit
        self.dim_w = dim_w
        self.kl_factor = kl_factor

        # build NN model
        self.layer_sizes = layer_sizes
        self.model = NNModel(self.layer_sizes, self.output_dim)
        if self.mode == 'MAML':
            self.model.build((None, self.input_dim))
            self.encoder = None
            self.dim_w = None
        elif self.mode == 'MR-MAML-W':
            self.model.build((None, self.dim_w))

            self.kernel_posterior_fn = tfp_layers_util.default_mean_field_normal_fn(
                untransformed_scale_initializer=tf.compat.v1.initializers.random_normal(mean=-50., stddev=0.1))

            self.encoder = tf.keras.Sequential([
              tfp.layers.DenseReparameterization(100, activation=tf.nn.relu, kernel_posterior_fn=self.kernel_posterior_fn, input_shape=(self.input_dim,)),
              tfp.layers.DenseReparameterization(dim_w, kernel_posterior_fn=self.kernel_posterior_fn),
            ])
        elif self.mode == 'MR-MAML-A':
            self.model.build((None, self.dim_w))
            self.encoder = Encoder(dim_w = dim_w)
            self.encoder.build((None, self.input_dim))

        self.optimizer = keras.optimizers.Adam(lr_meta)

    def meta_fit(self, valid_tuples=None, verbose=True, log_period=500, n_iter=None):
        """
        fits the initial params via MAML

        Args:
            valid_tuples: list of valid tuples, i.e. [(test_context_x_1, test_context_t_1, test_x_1, test_t_1), ...]
            verbose: (boolean) whether to print training progress
            log_period: (int) number of steps after which to print stats
            n_iter: (int) number of gradient descent iterations
        """

        assert (valid_tuples is None) or (all([len(valid_tuple) == 4 for valid_tuple in valid_tuples]))

        t = time.time()

        if n_iter is None:
            n_iter = self.num_iter_fit

        cum_loss = 0.0

        for itr in range(1, n_iter + 1):
            # sample task and split data into train / val batch
            x, y = self.meta_train_data[np.random.choice(len(self.meta_train_data))]
            x_train, y_train, x_val, y_val = self._shuffle_split_convert(x, y)

            if self.mode == 'MAML':
                loss = self._meta_step(x_train, y_train, x_val, y_val)
            elif self.mode == 'MR-MAML-W':
                loss = self._meta_step_mr_w(x_train, y_train, x_val, y_val)
            elif self.mode == 'MR-MAML-A':
                loss = self._meta_step_mr_a(x_train, y_train, x_val, y_val)
            else:
                raise NotImplementedError

            cum_loss += loss

            # print training stats stats
            if itr == 1 or itr % log_period == 0:
                duration = time.time() - t
                avg_loss = cum_loss / (log_period if itr > 1 else 1.0)
                cum_loss = 0.0
                t = time.time()

                message = 'Iter %d/%d  -  Loss: %.6f  -  Time %.2f sec' % (
                itr, self.num_iter_fit, avg_loss.numpy(), duration)

                # if validation data is provided  -> compute the valid log-likelihood
                if valid_tuples is not None:
                    results_dict = self.eval_datasets(valid_tuples)
                    for metric, value in results_dict.items():
                        message += '  -  %s: %.4f ' %(metric, value)

                if verbose:
                    self.logger.info(message)

        self.fitted = True

        return loss.numpy()

    def _meta_step(self, x_train, y_train, x_val, y_val):
        with tf.GradientTape(watch_accessed_variables=False) as test_tape:
            test_tape.watch(self.model.trainable_variables)
            with tf.GradientTape() as train_tape:
                train_loss, _ = self.compute_loss(self.model, x_train, y_train)

            gradients_inner = train_tape.gradient(train_loss, self.model.trainable_variables)  # \nabla_{\theta}
            model_copy = self._copy_model(self.model, x_train)
            k = 0
            for j in range(len(model_copy.layers)):
                model_copy.layers[j].kernel = tf.subtract(self.model.layers[j].kernel,  # \phi_t = T(\theta, \nabla_{\theta})
                                                          tf.multiply(self.lr_inner, gradients_inner[k]))
                model_copy.layers[j].bias = tf.subtract(self.model.layers[j].bias,
                                                        tf.multiply(self.lr_inner, gradients_inner[k + 1]))
                k += 2
            test_loss, y_hat = self.compute_loss(model_copy, x_val, y_val)  # test_loss
        gradients_outer = test_tape.gradient(test_loss, self.model.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients_outer, self.model.trainable_variables))
        return test_loss

    def _meta_step_mr_w(self, x_train, y_train, x_val, y_val):
        all_var = self.encoder.trainable_variables + self.model.trainable_variables
        with tf.GradientTape(watch_accessed_variables=False) as test_tape:
            test_tape.watch(all_var)
            with tf.GradientTape() as train_tape:
                w_train = self.encoder(x_train)
                y_hat_train = self.model.call(w_train)
                train_loss = keras_backend.mean(keras.losses.mean_squared_error(y_train, y_hat_train))
            gradients_inner = train_tape.gradient(train_loss, self.model.trainable_variables)  # \nabla_{\theta}
            model_copy = self._copy_model(self.model, input_shape=self.dim_w)

            k = 0
            for j in range(len(model_copy.layers)):
                model_copy.layers[j].kernel = tf.subtract(self.model.layers[j].kernel,
                                                          # \phi_t = T(\theta, \nabla_{\theta})
                                                          tf.multiply(self.lr_inner, gradients_inner[k]))
                model_copy.layers[j].bias = tf.subtract(self.model.layers[j].bias,
                                                        tf.multiply(self.lr_inner, gradients_inner[k + 1]))
                k += 2

            w_val = self.encoder(x_val)
            y_hat_val = model_copy.call(w_val)
            mse_loss = keras_backend.mean(keras.losses.mean_squared_error(y_val, y_hat_val))
            kl_loss = self.kl_factor * sum(self.encoder.losses)
            val_loss = mse_loss + kl_loss

        gradients_outer = test_tape.gradient(val_loss, all_var)
        self.optimizer.apply_gradients(zip(gradients_outer, all_var))
        return val_loss

    def _meta_step_mr_a(self, x_train, y_train, x_val, y_val):
        with tf.GradientTape(watch_accessed_variables=False) as test_tape, tf.GradientTape(
                watch_accessed_variables=False) as encoder_test_tape:
            test_tape.watch(self.model.trainable_variables)
            encoder_test_tape.watch(self.encoder.trainable_variables)
            with tf.GradientTape() as train_tape:
                w_train, _, _ = self.encoder(x_train)
                y_hat = self.model.call(w_train)
                train_loss = keras_backend.mean(keras.losses.mean_squared_error(y_train, y_hat))

            # inner steps
            gradients_inner = train_tape.gradient(train_loss, self.model.trainable_variables)  # \nabla_{\theta}
            model_copy = model_copy = self._copy_model(self.model, w_train)
            k = 0
            for j in range(len(model_copy.layers)):
                model_copy.layers[j].kernel = tf.subtract(self.model.layers[j].kernel,  # \phi_t = T(\theta, \nabla_{\theta})
                                                          tf.multiply(self.lr_inner, gradients_inner[k]))
                model_copy.layers[j].bias = tf.subtract(self.model.layers[j].bias,
                                                        tf.multiply(self.lr_inner, gradients_inner[k + 1]))
                k += 2

            # compute valid/test loss
            w_validation, w_mu_validation, w_sigma_validation = self.encoder(x_val)
            test_mse, _ = self.compute_loss(model_copy, w_validation, y_val)
            kl_ib = kl_qp_gaussian(w_mu_validation, w_sigma_validation,
                                   tf.zeros(tf.shape(w_mu_validation)), tf.ones(tf.shape(w_sigma_validation)))
            test_loss = test_mse + self.kl_factor * kl_ib

        # outer grad steps
        gradients_outer = test_tape.gradient(test_loss, self.model.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients_outer, self.model.trainable_variables))

        gradients = encoder_test_tape.gradient(test_loss, self.encoder.trainable_variables)
        keras.optimizers.Adam(learning_rate=0.001).apply_gradients(zip(gradients, self.encoder.trainable_variables))

        return test_loss

    def eval(self, x_context, y_context, x_test, y_test, num_steps=(1, 5)):
        x_context, y_context = _handle_input_dimensionality(x_context, y_context)
        x_test, y_test = _handle_input_dimensionality(x_test, y_test)
        x_context, y_context = np_to_tensor(self._normalize_data(x_context, y_context))
        x_test = tf.convert_to_tensor(self._normalize_data(x_test), dtype=tf.float32)

        # copy model so we can use the same model multiple times
        if self.mode == 'MAML':
            copied_model = self._copy_model(self.model, x=x_context)
        else:
            copied_model = self._copy_model(self.model, input_shape=self.dim_w)

        optimizer = keras.optimizers.SGD(learning_rate=self.lr_inner)

        rmse_record = []
        if self.mode == 'MAML':
          w_target = x_test
        elif self.mode == 'MR-MAML-W':
          w_target = self.encoder(x_test)
        elif self.mode == 'MR-MAML-A':
          _, w_target, _ = self.encoder(x_test)
        else:
            raise NotImplementedError

        for step in range(0, np.max(num_steps) + 1):
            if step in num_steps:
                # make prediction & unnormalize
                y_hat_normalized = copied_model.call(w_target).numpy()
                y_hat = y_hat_normalized * self.y_std[None, :] + self.y_mean[None, :]
                # compute rmse
                rmse = np.sqrt(np.mean(np.sum((y_hat - y_test)**2, axis=-1)))
                rmse_record.append((step, rmse))
            loss = self._grad_step(x_context, y_context, copied_model, optimizer)
        return rmse_record

    def eval_datasets(self, test_tuples, **kwargs):
        assert (all([len(valid_tuple) == 4 for valid_tuple in test_tuples]))
        rmse_list = [self.eval(*test_data_tuple, **kwargs) for test_data_tuple in test_tuples]

        results_dict = {}
        for rmse_list_n_steps in zip(*rmse_list):
            n_steps = rmse_list_n_steps[0][0]
            avg_rmse = np.mean([rmse for _, rmse in rmse_list_n_steps])
            results_dict['rmse_%i_steps' % n_steps] = avg_rmse

        return results_dict

    def _grad_step(self, x, y, model, optimizer):
        if self.mode == 'MAML':
            with tf.GradientTape() as tape:
                loss, _ = self.compute_loss(model, x, y)
        if self.mode == 'MR-MAML-W':
            w = self.encoder(x)
            with tf.GradientTape() as tape:
                y_hat = model.call(w)
                loss = keras_backend.mean(keras.losses.mean_squared_error(y, y_hat))
        if self.mode == 'MR-MAML-A':
            _, w, _ = self.encoder(x)
            with tf.GradientTape() as tape:
                y_hat = model.call(w)
                loss = keras_backend.mean(keras.losses.mean_squared_error(y, y_hat))
        gradients = tape.gradient(loss, model.trainable_variables)
        optimizer.apply_gradients(zip(gradients, model.trainable_variables))
        return loss

    @staticmethod
    def _shuffle_split_convert(x, y, shuffle=True):
        if shuffle:
            # shuffle points
            indices = np.arange(x.shape[0])
            np.random.shuffle(indices)
            x, y = x[indices], y[indices]

        # split
        idx_split = math.ceil(x.shape[0] / 2.0)
        x_train, y_train = x[:idx_split], y[:idx_split]
        x_val, y_val = x[idx_split:], y[idx_split:]

        x_train, y_train = np_to_tensor([x_train, y_train])
        x_val, y_val = np_to_tensor([x_val, y_val])
        return x_train, y_train, x_val, y_val

    @staticmethod
    def compute_loss(model, x, y):
        y_hat = model.call(x)
        loss = keras_backend.mean(keras.losses.mean_squared_error(y, y_hat))
        return loss, y_hat

    def _copy_model(self, model, x=None, input_shape=None):
        '''
          Copy model weights to a new model.
          Args:
              model: model to be copied.
              x: An input example.
        '''
        copied_model = NNModel(self.layer_sizes, self.output_dim)
        if x is not None:
            copied_model.call(tf.convert_to_tensor(x))
        if input_shape is not None:
            copied_model.build(tf.TensorShape([None, input_shape]))
        copied_model.set_weights(model.get_weights())
        return copied_model


if __name__ == '__main__':
    # get meta data
    from pacoh.datasets.regression_datasets import SinusoidDataset, provide_data
    # ds_generator = SinusoidDataset()
    # meta_train_data = ds_generator.generate_meta_train_data(1000, n_samples=10)
    # meta_test_data = ds_generator.generate_meta_test_data(n_tasks=200, n_samples_context=10, n_samples_test=100)

    meta_train_data, _, meta_test_data = provide_data(dataset='swissfel', seed=28)

    # meta_algo = MAML_MR(meta_train_data, mode='MAML', layer_sizes=(32, 32, 32, 32), kl_factor=5e-5,
    #                     normalize_data=True, num_iter_fit=20000, random_seed=22)
    # meta_algo.meta_fit(valid_tuples=meta_test_data, log_period=500)

    #setup & train MAML-MR model
    meta_algo = MAML_MR(meta_train_data, mode='MR-MAML-W', layer_sizes=(32, 32, 32, 32),
                        normalize_data=True, kl_factor=5e-5, num_iter_fit=20000, dim_w=24)
    meta_algo.meta_fit(valid_tuples=meta_test_data, log_period=500)

    # meta_algo = MAML_MR(meta_train_data, mode='MR-MAML-A', layer_sizes=(32, 32, 32, 32), kl_factor=0.1,
    #                     num_iter_fit=20000, random_seed=22)
    # meta_algo.meta_fit(valid_tuples=meta_test_data[:20], log_period=500, )