import unittest
import tensorflow as tf


from pacoh.modules.batched_module import TFModuleBatched

class TestTFModuleBatched(unittest.TestCase):

    def setUp(self):

        def function(x, params):
            b = params[0]
            w = params[1:]
            return tf.multiply(x ** 2, w ** 2) + b

        class CustomModel(TFModuleBatched):
            def __init__(self, name=None):
                super().__init__(name=name)
                self.w = tf.Variable(initial_value=2 * tf.ones((2,)), trainable=True)
                self.b = tf.Variable(initial_value=3 * tf.ones((1,)), trainable=True)

            def __call__(self, x):
                params = tf.concat([self.b, self.w], axis=-1)
                return function(x, params)

        self.function = function
        self.model = CustomModel()

        self.loss_fn = lambda y: tf.reduce_sum(y ** 2)

    def test_call_forward_consistency(self):
        x = tf.random.normal((2,))
        y1 = self.model(x)
        vars_vectorized = self.model.get_variables_vectorized()
        y2 = self.model.forward_parametrized(x, vars_vectorized)
        tf.assert_equal(y1, y2)

    def test_first_order_grads(self):
        params = 1.5 * tf.random.normal((3,))
        x = tf.random.normal((2,))

        with tf.GradientTape() as tape:
            tape.watch(params)
            loss = self.loss_fn(self.function(x, params))
        grad1 = tape.gradient(loss, params)

        with tf.GradientTape() as tape:
            tape.watch(params)
            loss = self.loss_fn(self.model.forward_parametrized(x, params))
        grad2 = tape.gradient(loss, params)

        print(grad1, grad2)
        tf.assert_equal(grad1, grad2)

    def test_maml_grads(self):

        n_inner_steps = 2
        loss_fn = self.loss_fn
        function = self.function

        x = tf.random.normal((2,))

        @tf.function
        def maml_step_grad1(w):
            with tf.GradientTape() as tape_outer:
                tape_outer.watch(w)
                for i in range(n_inner_steps):
                    with tf.GradientTape() as tape_inner:
                        tape_inner.watch(w)
                        l_inner = loss_fn(function(x, w))
                    inner_grad = tape_inner.gradient(l_inner, w)
                    w = w - 0.01 * inner_grad
                l_outer = loss_fn(function(x, w))
            return tape_outer.gradient(l_outer, w)

        @tf.function
        def maml_step_grad2(w):
            with tf.GradientTape() as tape_outer:
                tape_outer.watch(w)
                for i in range(n_inner_steps):
                    with tf.GradientTape() as tape_inner:
                        tape_inner.watch(w)
                        l_inner = loss_fn(self.model.forward_parametrized(x, w))
                    inner_grad = tape_inner.gradient(l_inner, w)
                    w = w - 0.01 * inner_grad
                l_outer = loss_fn(self.model.forward_parametrized(x, w))
            return tape_outer.gradient(l_outer, w)

        w = tf.random.normal((3,))
        grad1 = maml_step_grad1(w)
        grad2 = maml_step_grad2(w)
        print(grad1, grad2)
        tf.assert_equal(grad1, grad2)