import unittest
import numpy as np
import copy

from pacoh.datasets import get_regression_datasets, MetaDatasetSampler, get_classification_datasets

class TestDataSamplerRegression1(unittest.TestCase):

    def setUp(self):
        self.data_config = {'dataset': 'sin_7',
                       'num_train_tasks': 3,
                       'num_test_valid_tasks': 4,
                       'num_train_samples': 5,
                       'num_context_samples': 5,
                       'num_test_valid_samples': 20,
                       'standardization': True,
                       'tiled': True,
                       'batch_size': 2,
                       'posterior_inference_batch_size': -1,
                       'meta_batch_size': -1
                       }



        self.n_batched_models = 12

    def test_seed_reproducability(self):
        config = copy.deepcopy(self.data_config)
        for tiled in [True, False]:
            config['tiled'] = tiled

            train1, val1 = get_regression_datasets(config, n_batched_models_train=self.n_batched_models,
                                                   n_batched_models_test=self.n_batched_models)

            train2, val2 = get_regression_datasets(config, n_batched_models_train=self.n_batched_models,
                                                   n_batched_models_test=self.n_batched_models)

            for i in range(5):
                # train
                x1, y1, n1, m1, _ = train1.get_meta_batch()
                x2, y2, n2, m2, _ = train2.get_meta_batch()

                assert np.array_equal(x1.numpy(), x2.numpy())
                assert np.array_equal(y1.numpy(), y2.numpy())
                assert np.array_equal(n1.numpy(), n2.numpy())
                assert np.array_equal(m1.numpy(), m2.numpy())

                # val
                x1, y1, n1, m1, _ = val1.get_meta_batch()
                x2, y2, n2, m2, _ = val2.get_meta_batch()

                assert np.array_equal(x1.numpy(), x2.numpy())
                assert np.array_equal(y1.numpy(), y2.numpy())
                assert np.array_equal(n1.numpy(), n2.numpy())
                assert np.array_equal(m1.numpy(), m2.numpy())

    def test_tiling(self):

        train, val = get_regression_datasets(self.data_config, n_batched_models_train=self.n_batched_models,
                                                   n_batched_models_test=self.n_batched_models)

        for i in range(5):

            # --- train ----
            x, y, n, m, _ = train.get_meta_batch()
            x, y, n = x.numpy(), y.numpy(), n.numpy()

            assert x.ndim == y.ndim == 4

            # check first (tiled) dimension
            assert x.shape[1] == self.n_batched_models
            assert y.shape[1] == self.n_batched_models

            # check train batch size
            batch_size = self.data_config['batch_size'] if self.data_config['batch_size'] > 0 \
                else self.data_config['num_train_samples']
            assert x.shape[2] == batch_size
            assert y.shape[2] == batch_size

            # check that tiled elements are equal
            assert x.shape[1] == y.shape[1]
            for i in range(x.shape[0]):
                for j in range(x.shape[1] - 1):
                    assert np.array_equal(x[i, j], x[i, j+1])
                    assert np.array_equal(y[i, j], y[i, j+1])

            # check that meta-batch size is correct
            assert x.shape[0] == y.shape[0] == 7
            assert m.shape[0] == n.shape[0] == 7

            # --- valid ----

            x, y, n, m, _ = val.get_meta_batch()
            x, y, n = x.numpy(), y.numpy(), n.numpy()

            assert x.ndim == y.ndim == 4

            # check first (tiled) dimension
            assert x.shape[1] == self.n_batched_models
            assert y.shape[1] == self.n_batched_models

            # check context batch size
            batch_size = self.data_config['posterior_inference_batch_size'] if self.data_config['posterior_inference_batch_size'] > 0 \
                else self.data_config['num_context_samples']
            assert x.shape[2] == batch_size
            assert y.shape[2] == batch_size

            # check that tiled elements are equal
            for i in range(x.shape[0]):
                for j in range(x.shape[1] - 1):
                    assert np.array_equal(x[i, j], x[i, j+1])
                    assert np.array_equal(y[i, j], y[i, j+1])

            # check that meta-batch size is correct
            assert x.shape[0] == y.shape[0] == 4
            assert m.shape[0] == n.shape[0] == 4

    def test_no_tiling(self):
        config = copy.deepcopy(self.data_config)
        config['tiled'] = False
        config['batch_size'] = -1
        config['posterior_inference_batch_size'] = 4

        train, val = get_regression_datasets(config, n_batched_models_train=self.n_batched_models,
                                                   n_batched_models_test=self.n_batched_models)

        for i in range(7):
            # train
            x, y, n, m, _ = train.get_meta_batch()
            x, y, n = x.numpy(), y.numpy(), n.numpy()

            assert x.ndim == y.ndim == 3

            # check train batch size
            batch_size = config['batch_size'] if config['batch_size'] > 0 \
                else config['num_train_samples']
            assert x.shape[1] == batch_size
            assert y.shape[1] == batch_size

            # valid
            x, y, n, m, _ = val.get_meta_batch()
            x, y, n = x.numpy(), y.numpy(), n.numpy()

            assert x.ndim == y.ndim == 3

            # check context batch size
            batch_size = config['posterior_inference_batch_size'] if config['posterior_inference_batch_size'] > 0 \
                else config['num_context_samples']
            assert x.shape[1] == batch_size
            assert y.shape[1] == batch_size

class TestDataSamplerRegression2(unittest.TestCase):

    def setUp(self):
        data_train = []
        for i in range(10):
            n_samples = i + 2
            value = float(3 * i) + 1
            x = value * np.ones((n_samples, 3)) + np.random.normal(scale=0.1, size=(n_samples, 3))
            y = value * np.ones((n_samples, 2)) + np.random.normal(scale=0.1, size=(n_samples, 2))
            data_train.append((x, y))
        self.data_train = data_train


    def test_task_sampler_tiled(self):
        n_batched_models = 7
        train_meta_ds = MetaDatasetSampler(self.data_train, per_task_batch_size=5, meta_batch_size=3,
                                           n_batched_models=n_batched_models, tiled=True, standardization=False)

        values = []
        for iter in range(3):
            x, y, n, _, _ = train_meta_ds.get_meta_batch()
            x, y, n = x.numpy(), y.numpy(), n.numpy()

            assert x.shape[1] == y.shape[1] == n_batched_models
            assert x.shape[2] == y.shape[2] == train_meta_ds.per_task_batch_size
            values.extend([int(np.mean(x[i])) for i in range(3)])

        assert len(set(values)) == 9

    def test_task_sampler(self):
        train_meta_ds = MetaDatasetSampler(self.data_train, per_task_batch_size=5, meta_batch_size=3,
                                           n_batched_models=None, tiled=False, standardization=False)
        values = []
        for iter in range(3):
            x, y, n, _, _ = train_meta_ds.get_meta_batch()
            x, y, n = x.numpy(), y.numpy(), n.numpy()

            assert x.shape[1] == y.shape[1] == train_meta_ds.per_task_batch_size
            values.extend([int(np.mean(x[i])) for i in range(3)])

        assert len(set(values)) == 9

    def test_standardization(self):
        train_meta_ds = MetaDatasetSampler(self.data_train, per_task_batch_size=5, meta_batch_size=3,
                                           n_batched_models=1, tiled=True,
                                           standardization=True)


        assert train_meta_ds.x_mean.numpy().mean() >= 5.0
        assert train_meta_ds.y_std.numpy().mean() > 2.0

        train_meta_ds = MetaDatasetSampler(self.data_train, per_task_batch_size=5, meta_batch_size=3,
                                           n_batched_models=1, tiled=True,
                                           standardization=False)

        assert np.abs(train_meta_ds.x_mean.numpy().mean() - 0.0) < 0.001
        assert np.abs(train_meta_ds.y_mean.numpy().mean() - 0.0) < 0.001
        assert np.abs(train_meta_ds.x_std.numpy().mean() - 1.0) < 0.001
        assert np.abs(train_meta_ds.y_std.numpy().mean() - 1.0) < 0.001

class TestDataSamplerClassification1(unittest.TestCase):

    def setUp(self):
        self.data_config = {'dataset': 'omniglot',
                            'n_classes': 5,
                            'n_tasks_train': 10,
                            'n_tasks_valid': 5,
                           'n_samples_train': 5,
                           'n_samples_context': 5,
                           'n_samples_test': 10,
                           'tiled': True,
                           'batch_size': 2,
                           'posterior_inference_batch_size': -1,
                           'meta_batch_size': -1
                           }

        self.n_batched_models = 12


    def test_seed_reproducability(self):
        config = copy.deepcopy(self.data_config)
        for tiled in [True, False]:
            config['tiled'] = tiled


            train1, val1 = get_classification_datasets(config, n_batched_models_train=self.n_batched_models,
                                                   n_batched_models_test=self.n_batched_models)

            train2, val2 = get_classification_datasets(config, n_batched_models_train=self.n_batched_models,
                                                   n_batched_models_test=self.n_batched_models)

            for i in range(5):
                # train
                x1, y1, n1, m1, _ = train1.get_meta_batch()
                x2, y2, n2, m2, _ = train2.get_meta_batch()

                assert np.array_equal(x1.numpy(), x2.numpy())
                assert np.array_equal(y1.numpy(), y2.numpy())
                assert np.array_equal(n1.numpy(), n2.numpy())
                assert np.array_equal(m1.numpy(), m2.numpy())

                # val
                x1, y1, n1, m1, _ = val1.get_meta_batch()
                x2, y2, n2, m2, _ = val2.get_meta_batch()

                assert np.array_equal(x1.numpy(), x2.numpy())
                assert np.array_equal(y1.numpy(), y2.numpy())
                assert np.array_equal(n1.numpy(), n2.numpy())
                assert np.array_equal(m1.numpy(), m2.numpy())

class TestDataProvidedOmniglot(unittest.TestCase):

    def setUp(self):
        self.data_config = {'dataset': 'omniglot',
                       'type': 'class',
                       'num_train_tasks': None,
                       'num_test_valid_tasks': None,
                       'num_train_samples': None,
                       'num_context_samples': None,
                       'num_test_valid_samples': None,
                       'standardization': False,
                       'tiled': True,
                       'batch_size': -1,
                       'meta_batch_size': 5,
                       'posterior_inference_batch_size': -1,
                       }

    def test_n_shot_1(self):
        self.data_config['dataset'] = 'omniglot_8shot_full'
        train_sampler, test_sampler = get_classification_datasets(self.data_config,
                                                                  n_batched_models_train=7,
                                                                  n_batched_models_test=7)
        n_classes = 5

        for i in range(5):
            x, y, n, m, _ = train_sampler.get_meta_batch()
            assert x.shape == (5, 7, n_classes * 20, 28, 28, 1)
            assert y.shape == (5, 7, n_classes * 20, n_classes)

        for i in range(3):
            x, y, n, m, _ = test_sampler.get_meta_batch()
            assert x.shape == (20, 7, n_classes*8, 28, 28, 1)
            assert y.shape == (20, 7, n_classes*8, n_classes)


    def test_n_shot_2(self):
        self.data_config['dataset'] = 'omniglot_6shot_small'
        train_sampler, test_sampler = get_classification_datasets(self.data_config,
                                                                  n_batched_models_train=7,
                                                                  n_batched_models_test=7)
        n_classes = 5

        for i in range(5):
            x, y, n, m, _ = train_sampler.get_meta_batch()
            assert x.shape == (5, 7, n_classes*6, 28, 28, 1)
            assert y.shape == (5, 7, n_classes*6, n_classes)

        for i in range(3):
            x, y, n, m, _ = test_sampler.get_meta_batch()
            assert x.shape == (20, 7, n_classes*6, 28, 28, 1)
            assert y.shape == (20, 7, n_classes*6, n_classes)
