import unittest
import numpy as np

from pacoh.algorithms import BayesianNeuralNetworkSVGD

class TestMetaRegressionSVGD(unittest.TestCase):


    def setUp(self) -> None:
        NUM_TRAIN_TASKS = 3
        NUM_TEST_VALID_TASKS = 2
        NUM_TRAIN_SAMPLES = 5
        NUM_CONTEXT_SAMPLES = 5
        NUM_TEST_VALID_SAMPLES = 5
        DATASET = 'sin'

        likelihood_config = {'class': {'learn_likelihood_variables': False},
                             'reg': {'log_var': 2 * np.log(0.01),
                                     'learn_likelihood_variables': True},
                             'type': 'reg'
                             }

        LAYER_SIZE = 8
        N_LAYERS = 2
        ACTIVATION = 'relu'

        model_config = {'fc': {'input_dim': None,
                               'output_dim': 1,
                               'layer_units': [LAYER_SIZE for _ in range(N_LAYERS)] + [1],
                               'activations': [ACTIVATION for _ in range(N_LAYERS)] + [None]
                               },
                        'cnn': {},
                        'type': 'fc'}

        data_config = {'dataset': DATASET,
                       'num_train_tasks': NUM_TRAIN_TASKS,
                       'num_test_valid_tasks': NUM_TEST_VALID_TASKS,
                       'num_train_samples': NUM_TRAIN_SAMPLES,
                       'num_context_samples': NUM_CONTEXT_SAMPLES,
                       'num_test_valid_samples': NUM_TEST_VALID_SAMPLES,
                       'standardization': True,
                       'tiled': True,
                       'batch_size': -1,
                       'posterior_inference_batch_size': -1,
                       'meta_batch_size': 10
                       }

        model_gaussian_prior = {'mean_mu': 0.0,
                                'bias_log_var_mu': -5,
                                'weight_log_var_mu': -5  # 2 * np.log(0.5)
                                }

        hyper_prior_config = {'mean_mean': 0.0,  # mean of the hyper-prior for the means
                              'bias_mean_std': 3,  # std of the hyper-prior for the mean of the biases
                              'kernel_mean_std': 1,  # std of the hyper-prior for the mean of the kernels
                              'log_var_mean': -1,
                              'bias_log_var_std': 1,  # std of the hyper-prior for the log-variances of the prior
                              'kernel_log_var_std': 1,  # std of the hyper-prior for the log-variances of the prior
                              'likelihood_log_var_mean_mean': -4,
                              'likelihood_log_var_mean_std': 1,
                              'likelihood_log_var_log_var_mean': -2,
                              'likelihood_log_var_log_var_std': 0.1
                              }

        likelihood_gaussian_prior = {'log_var_mu': 2 * np.log(0.01),
                                     'log_var_log_var': 2 * np.log(2.0)
                                     }

        prior_config = {'gaussian': {'model': model_gaussian_prior,
                                     'likelihood': likelihood_gaussian_prior,
                                     'init_std': 0.1,
                                     'add_noise': True,
                                     'init_from_models': True,
                                     'n_samples': 4,  # number of samples from prior for estimating the mll
                                     },
                        'type': 'gaussian'
                        }

        hyper_posterior_config = {'particle': {'n_samples': 2,
                                               },
                                  'type': 'particle'
                                  }

        posterior_config = {'particle': {'n_samples': 3,
                                         'use_init_std': True
                                         },
                            'type': 'particle'
                            }

        training_config = {'epochs': 2,
                           'n_train_loss_logs': 0,
                           'n_eval_runs': 0,
                           'optimizer': 'adam',
                           'learning_rate': 0.002,
                           'plot_functions': False,
                           'plot_pred_lines': True,
                           'plot_pred_std': True,
                           'prior_prob_weight': 0.0001,
                           'bandwidth': 1.0
                           }

        inference_training_config = {'epochs': 3,
                                     'n_train_loss_logs': 0,
                                     'n_eval_runs': 0,
                                     'optimizer': 'adam',
                                     'learning_rate': 0.01,
                                     'plot_functions': False,
                                     'plot_pred_lines': True,
                                     'plot_pred_std': True,
                                     'weight_decay': 1e-5,
                                     'prior_prob_weight': 0.1,
                                     'bandwidth': 1.0
                                     }

        self.prior_config = prior_config[prior_config['type']]
        self.posterior_config = posterior_config[posterior_config['type']]
        self.hyper_posterior_config = hyper_posterior_config[hyper_posterior_config['type']]

        self.model_config = model_config
        self.data_config = data_config
        self.hyper_prior_config = hyper_prior_config
        self.training_config = training_config
        self.inference_training_config = inference_training_config
        self.likelihood_config = likelihood_config

    def test_seed_consistency(self):
        import pacoh.datasets as datasets
        import pacoh.algorithms as algorithms
        import tensorflow as tf

        x_query = tf.expand_dims(tf.range(-5, 5, delta=0.1, dtype=tf.float32), axis=-1)

        # setup data set and corresponding samplers
        n_batched_models_train = self.hyper_posterior_config['n_samples'] * self.prior_config['n_samples']
        n_batched_models_test = self.hyper_posterior_config['n_samples'] * self.posterior_config['n_samples']


        # model 1
        train_ds1, val_ds1 = datasets.get_regression_datasets(self.data_config,
                                                            n_batched_models_train=n_batched_models_train,
                                                            n_batched_models_test=n_batched_models_test)

        meta_reg_model1 = algorithms.MetaBayesianNeuralNetwork_SVGD(train_ds1, val_ds1, self.model_config,
                                                                   self.likelihood_config,
                                                                   self.prior_config, self.posterior_config,
                                                                   self.hyper_prior_config, self.hyper_posterior_config,
                                                                   self.training_config, self.inference_training_config,
                                                                   random_seed=234)

        meta_reg_model1.meta_fit()

        y_hat1, pred_dist_1 = meta_reg_model1.predict(x_query, sample_functions=True, sample_from_prior=True)
        pred_std_1 = pred_dist_1.stddev()

        eval_results1, _ = meta_reg_model1.meta_eval()

        # model 2
        train_ds2, val_ds2 = datasets.get_regression_datasets(self.data_config,
                                                              n_batched_models_train=n_batched_models_train,
                                                              n_batched_models_test=n_batched_models_test)

        meta_reg_model2 = algorithms.MetaBayesianNeuralNetwork_SVGD(train_ds2, val_ds2, self.model_config,
                                                                    self.likelihood_config,
                                                                    self.prior_config, self.posterior_config,
                                                                    self.hyper_prior_config,
                                                                    self.hyper_posterior_config,
                                                                    self.training_config,
                                                                    self.inference_training_config,
                                                                    random_seed=234)

        meta_reg_model2.meta_fit()

        y_hat2, pred_dist_2 = meta_reg_model2.predict(x_query, sample_functions=True, sample_from_prior=True)
        pred_std_2 = pred_dist_2.stddev()

        eval_results2, _ = meta_reg_model2.meta_eval()

        assert np.all(tf.math.equal(y_hat1, y_hat2).numpy())
        assert np.all(tf.math.equal(pred_std_1, pred_std_2).numpy())
        assert all([eval_results1[key] == eval_results1[key] for key in eval_results2.keys()])


