import unittest
import numpy as np

from pacoh.algorithms import BayesianNeuralNetworkSVGD

class TestBNNRegression(unittest.TestCase):

    def setUp(self):

        likelihood_config = {'class': {'learn_likelihood_variables': False},
                             'reg': {'log_var': 2 * np.log(0.1),
                                     'learn_likelihood_variables': True},
                             'type': 'reg'
                             }

        model_config = {'fc': {'input_dim': None,
                               'output_dim': 1,
                               'layer_units': [40, 40, 40, 1],
                               'activations': ['tanh', 'tanh', 'tanh', None]
                               },
                        'type': 'fc'}

        model_gaussian_prior = {'mean_mu': 0.0,
                                'bias_log_var_mu': 2 * np.log(1.0),
                                'weight_log_var_mu': 2 * np.log(0.5)
                                }

        likelihood_gaussian_prior = {'log_var_mu': 2 * np.log(0.1),
                                     'log_var_log_var': 2 * np.log(2.0)
                                     }

        prior_config = {'gaussian': {'model': model_gaussian_prior,
                                     'likelihood': likelihood_gaussian_prior,
                                     'init_std': 0.1,
                                     'n_batched_priors': 1,
                                     'add_noise': False
                                     },
                        'type': 'gaussian'
                        }

        posterior_config = {'particle': {'n_samples': 10,
                                         'use_init_std': False
                                         },
                            'type': 'particle'
                            }

        training_config = {'epochs': 100,
                           'n_train_loss_logs': 0,
                           'n_eval_runs': 0,
                           'optimizer': 'adam',
                           'learning_rate': 0.01,
                           'plot_functions': False,
                           'plot_pred_lines': False,
                           'plot_pred_std': False,
                           'prior_prob_weight': 0.2,
                           'bandwidth': 10.0
                           }


        prior_config = prior_config[prior_config['type']]
        posterior_config = posterior_config[posterior_config['type']]

        self.likelihood_config = likelihood_config
        self.model_config = model_config
        self.prior_config = prior_config
        self.posterior_config = posterior_config
        self.training_config = training_config

        # create datasets and corresponding samplers
        self.n_batched_models = posterior_config['n_samples']

        # setup data
        # Data setup
        from pacoh.datasets.regression_datasets import SinusoidDataset
        from pacoh.datasets.data_sampler import MetaDatasetSampler

        rds = np.random.RandomState(12)
        meta_data_sampler = SinusoidDataset(random_state=rds)
        self.meta_data = meta_data_sampler.generate_meta_test_data(n_tasks=1, n_samples_context=20, n_samples_test=100)
        meta_test_sampler = MetaDatasetSampler(self.meta_data, per_task_batch_size=5, n_batched_models=self.n_batched_models,
                                               tiled=True)

        self.dataset_sampler = meta_test_sampler.tasks[0]

    def test_BNN_regression_fit(self):
        # BNN model
        import tensorflow as tf
        reg_model = BayesianNeuralNetworkSVGD(self.dataset_sampler, self.model_config,
                                                self.likelihood_config,
                                                self.prior_config, self.posterior_config,
                                                self.training_config, random_seed=345)
        reg_model.fit()


        context_x, context_y, test_x, test_y = self.meta_data[0]

        # if test fails set plot = True for diagnostics
        plot = False
        if plot:
            x_plot = tf.expand_dims(tf.range(-8, 8, delta=0.1, dtype=tf.float32), axis=-1)
            _, pred_dist = reg_model.predict(x_plot)
            pred_mean, pred_stddev = pred_dist.mean().numpy().flatten(), pred_dist.stddev().numpy().flatten()
            x_plot = x_plot.numpy().flatten()

            from matplotlib import pyplot as plt
            plt.plot(x_plot, pred_mean)
            plt.scatter(test_x.flatten(), test_y.flatten())
            plt.fill_between(x_plot, pred_mean - pred_stddev, pred_mean + pred_stddev, alpha=0.2)
            plt.show()

        # check that that context data is fit well
        context_x, context_y = tf.cast(context_x, dtype=tf.float32), tf.cast(context_y, dtype=tf.float32)
        pred_mean, pred_dist = reg_model.predict(context_x)
        rmse = tf.math.sqrt(tf.reduce_mean((pred_mean - context_y)**2))

        assert rmse < 0.5

        # compute avg std in tails (outside training support)
        x_min, x_max = tf.reduce_min(test_x, axis=0)[0], tf.reduce_max(test_x, axis=0)[0]
        x_tails = tf.concat([tf.range(x_min-4, x_min, delta=0.1), tf.range(x_min+4, x_max, delta=0.1)], axis=0)
        x_tails = tf.cast(tf.expand_dims(x_tails, axis=-1), dtype=tf.float32)
        _, pred_dist = reg_model.predict(x_tails)
        pred_mean, pred_stddev = pred_dist.stddev(), pred_dist.stddev()
        avg_std_tails = tf.reduce_mean(pred_stddev)

        # compute std tails (in training data support)
        test_x = tf.cast(test_x, dtype=tf.float32)
        _, pred_dist = reg_model.predict(test_x)
        pred_mean, pred_stddev = pred_dist.stddev(), pred_dist.stddev()
        avg_std_support = tf.reduce_mean(pred_stddev)

        # check that uncertainty in tails is higher than in data support
        assert avg_std_tails > avg_std_support

    def test_BNN_regression_seed_consistency(self):
        import copy
        import tensorflow as tf
        training_config = copy.deepcopy(self.training_config)
        training_config['epochs'] = 2

        context_x, context_y, test_x, test_y = self.meta_data[0]
        test_x = tf.cast(test_x, dtype=tf.float32)

        from pacoh.datasets.data_sampler import MetaDatasetSampler

        # RUN 1
        meta_data = copy.deepcopy(self.meta_data)
        meta_test_sampler = MetaDatasetSampler(meta_data, per_task_batch_size=5,
                                               n_batched_models=self.n_batched_models,
                                               tiled=True)

        dataset_sampler = meta_test_sampler.tasks[0]

        reg_model1 = BayesianNeuralNetworkSVGD(dataset_sampler, self.model_config,
                                              self.likelihood_config,
                                              self.prior_config, self.posterior_config,
                                              training_config, random_seed=123)
        reg_model1.fit()
        pred_mean1, pred_dist1 = reg_model1.predict(test_x)
        pred_std_1 = pred_dist1.stddev()

        # RUN 2
        meta_data = copy.deepcopy(self.meta_data)
        meta_test_sampler = MetaDatasetSampler(meta_data, per_task_batch_size=5,
                                               n_batched_models=self.n_batched_models,
                                               tiled=True)

        dataset_sampler = meta_test_sampler.tasks[0]

        reg_model2 = BayesianNeuralNetworkSVGD(dataset_sampler, self.model_config,
                                              self.likelihood_config,
                                              self.prior_config, self.posterior_config,
                                              training_config, random_seed=123)
        reg_model2.fit()
        pred_mean2, pred_dist2 = reg_model2.predict(test_x)
        pred_std_2 = pred_dist2.stddev()

        assert np.all(tf.math.equal(pred_mean1, pred_mean2).numpy())
        assert np.all(tf.math.equal(pred_std_1, pred_std_2).numpy())